/* subroutines for insn-output.c for motorola dsp561XX.
   copyright ( c ) 1988 free software foundation, inc.

this file is part of gnu cc.

gnu cc is distributed in the hope that it will be useful,
but without any warranty.  no author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  refer to the gnu cc general public
license for full details.

everyone is granted permission to copy, modify and redistribute
gnu cc, but only under the conditions described in the
gnu cc general public license.	 a copy of this license is
supposed to have been given to you along with gnu cc so you
can know your rights and responsibilities.  it should be in a
file named copying.  among other things, the copyright notice
and this notice must be preserved on all copies.  */

#define SC "\n\t"

#ifndef file
#include <stdio.h>
#endif

#include <string.h>
#if defined ( _MSDOS )
extern void *malloc( );
#endif

#include <ctype.h>

#include "flags.h"
#include "tree.h"
#if ! defined( _MSDOS )
#include "hard-reg-set.h"
#else
#include "hardrset.h"
#endif

static const char mem_space = 'x';

enum reg_class regno_reg_class ( );

func_attributes current_func_info = FUNC_NO_INFO;

int FirstSave;

/* the dsp56100 uses the IEEE single precision format. this routine assumes
   that the host has 32 bit integers and that float is ieee sp. */
unsigned int
get_float_format ( number )
    double number;
{
    float ieee_sp = (float) number;
    unsigned int *peek = (unsigned int*) & ieee_sp;
    
#if defined( _MSDOS ) || defined( MIPSEL )
    return ( 0x0000ffff & ((*peek) >> 16 )) | ( 0xffff0000 & ((*peek) << 16 ));
#else
    return *peek;
#endif
}


#define N_REG_COUNT 4
#define FIRST_N_REG 6

static rtx current_n_reg_value[N_REG_COUNT];

void
clear_n_reg_values ( )
{
    int i = N_REG_COUNT;
    
    while ( i -- )
    {
	current_n_reg_value[i] = NULL;
    }
}

int
load_n_reg_p ( reg, new_value )
    rtx reg, new_value;
{
    if ( current_n_reg_value[REGNO ( reg ) - FIRST_N_REG] &&
	rtx_equal_p ( current_n_reg_value[REGNO ( reg ) - FIRST_N_REG], 
		      new_value ) &&
	CONSTANT_P ( new_value ))
    {
	return 0;
    }
    else
    {
	current_n_reg_value[REGNO ( reg ) - FIRST_N_REG] =  new_value;

	return 1;
    }
}

char*
move_singleword ( operands )
    rtx *operands;
{
    if (( rtx_equal_p ( operands[0], operands[1] )) &&
	(( REG != GET_CODE ( operands[0] )) ||
	 ( ! MEM_IN_STRUCT_P ( operands[0] ))) &&
	(( REG != GET_CODE ( operands[1] )) ||
	 ( ! MEM_IN_STRUCT_P ( operands[1] ))))
    {
	/* if we're doing a nop move, return a null string. note that
	   an insn could look like a nop move isasmuchas rtx_equal_p returns
	   true; however the MEM_IN_STRUCT_P kludge used by alter_subreg ( )
	   (final.c) may indicate a move from the LSW of a long to an int
	   in the same accumulator. */

	return "";
    }
    
    switch ( which_alternative )
    {
    case 0:
	if (( ! MEM_IN_STRUCT_P ( operands[1] )) &&
	    ( DST_REGS == REGNO_REG_CLASS ( REGNO ( operands[0] )) &&
	     ( DST_REGS == REGNO_REG_CLASS ( REGNO ( operands[1] )) ||
	      IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))))
	{
	    return "tfr	%1,%0";
	}
	else
	{
	    return "move	%e1,%0";
	}

    case 1:
	if ( PLUS == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[0], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    }
	    
	    switch ( INTVAL ( operands[3] ))
	    {
	    case 0:
		return "move	%e1,x:(%2)";
		    
	    case 1:
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	%e1,x:(%2+%q3)";
		}
		else if ( REGNO ( operands[2] ) != REGNO ( operands[1] ))
		{
		    return "lea	(%2)+,%2" SC
			   "move	%e1,x:(%2)-";
		}
		    
	    case -1:
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	%e1,x:(%2+%q3)";
		}
		else if ( REGNO ( operands[2] ) != REGNO ( operands[1] ))
		{
		    return "move	(%2)-" SC
			   "move	%e1,x:(%2)+";
		}
		
	    default:
		/* r2 + constant is limited to a single signed byte, otherwise
		   conventional (for the 56k) addressing must be used. */
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ) &&
		    128 > INTVAL ( operands[3] ) &&
		    -129 < INTVAL ( operands[3] ))
		{
		    return "move	%e1,x:(%2+%q3)";
		}
		else if ( load_n_reg_p ( operands[2], operands[3] ))
		{
		    return "move	#%q3,%j2" SC
			   "move	%e1,x:(%2+%j2)";
		}
		else
		{
		    return "move	%e1,x:(%2+%j2)";
		}
	    }
	}
	else
	{
	    return "move	%e1,%f0";
	}
	
    case 2:
	if ( PLUS == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[1], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    }
	    
	    switch ( INTVAL ( operands[3] ))
	    {
	    case 0:
		return "move	x:(%2),%0";
		    
	    case 1:
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	x:(%2+%q3),%0";
		}
		else if ( REGNO ( operands[0] ) != REGNO ( operands[2] ))
		{
		    return "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%0";
		}
		    
	    case -1:
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	x:(%2+%q3),%0";
		}
		else if ( REGNO ( operands[0] ) != REGNO ( operands[2] ))
		{
		    return "move	(%2)-" SC
			   "move	x:(%2)+,%0";
		}
		
	    default:
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ) &&
		    128 > INTVAL ( operands[3] ) &&
		    -129 < INTVAL ( operands[3] ))
		{
		    return "move	x:(%2+%q3),%0";
		}
		else if ( load_n_reg_p ( operands[2], operands[3] ))
		{
		    return "move	#%q3,%j2" SC
			   "move	x:(%2+%j2),%0";
		}
		else
		{
		    return "move	x:(%2+%j2),%0";
		}
	    }
	}
	else
	{
	    return "move	%f1,%0";
	}
	
    case 3:
	if ( 0 == INTVAL ( operands[1] ))
	{
	    if ( DST_REGS == REGNO_REG_CLASS ( REGNO ( operands[0] )))
	    {
		return "clr	%0";
	    }
	    else
	    {
		return "move	#%q1,%0";
	    }
	}
	else
	{
	    return "move	#>%q1,%0";
	}
    }
}

char*
move_doubleword ( is_float_p, operands )
    int is_float_p;
    rtx *operands;
{
    extern union real_extract adjusted_dval;

    if ( rtx_equal_p ( operands[0], operands[1] ))
    {
	return "";
    }
    
    switch ( which_alternative )
    {
    case 0:
	if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	%g1,%g0" SC
		       "move	%1,%0";
	    }
	    else
	    {
		return "tfr	%i1,%0";
	    }
	}
	else
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	%e1,%g0" SC
		       "move	%h1,%0";
	    }
	    else
	    {
		return "tfr	%1,%0";
	    }
	}

    case 1:
	if ( PLUS == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[0], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    }
	    
	    if ( ! INTVAL ( operands[3] ))
	    {
		if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
		{
		    return "move	%1,x:(%2)+" SC
			   "move	%g1,x:(%2)-";
		}
		else
		{
		    return "move	%h1,x:(%2)+" SC
			   "move	%e1,x:(%2)-";
		}
	    }
	    else
	    {
		/* r2 + constant is limited to a single signed byte, otherwise
		   conventional (for the 56k) addressing must be used. */

		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    if ( 127 > INTVAL ( operands[3] ) &&
			-128 < INTVAL ( operands[3] ))
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
			{
			    return "move	%g1,x:(%2+%q3+1)" SC
				   "move	%1,x:(%2+%q3)";
			}
			else
			{
			    return "move	%e1,x:(%2+%q3+1)" SC
				   "move	%h1,x:(%2+%q3)";
			}
		    }
		    else
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2" SC
				       "lea	(%2)+,%2" SC
				       "move	%g1,x:(%2+%j2)" SC
				       "move	(%2)-" SC
				       "move	%1,x:(%2+%j2)";
			    }
			    else
			    {
				return "lea	(%2)+,%2" SC
				       "move	%g1,x:(%2+%j2)" SC
				       "move	(%2)-" SC
				       "move	%1,x:(%2+%j2)";
			    }
			}
			else
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2" SC
				       "lea	(%2)+,%2" SC
				       "move	%e1,x:(%2+%j2)" SC
				       "move	(%2)-" SC
				       "move	%h1,x:(%2+%j2)";
			    }
			    else
			    {
				return "lea	(%2)+,%2" SC
				       "move	%e1,x:(%2+%j2)" SC
				       "move	(%2)-" SC
				       "move	%h1,x:(%2+%j2)";
			    }
			}
		    }
		}
		else
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2" SC
				   "lea	(%2)+,%2" SC
				   "move	%g1,x:(%2+%j2)" SC
				   "move	(%2)-" SC
				   "move	%1,x:(%2+%j2)";
			}
			else
			{
			    return "lea	(%2)+,%2" SC
				   "move	%g1,x:(%2+%j2)" SC
				   "move	(%2)-" SC
				   "move	%1,x:(%2+%j2)";
			}
		    }
		    else
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2" SC
				   "lea	(%2)+,%2" SC
				   "move	%e1,x:(%2+%j2)" SC
				   "move	(%2)-" SC
				   "move	%h1,x:(%2+%j2)";
			}
			else
			{
			    return "lea	(%2)+,%2" SC
				   "move	%e1,x:(%2+%j2)" SC
				   "move	(%2)-" SC
				   "move	%h1,x:(%2+%j2)";
			}
		    }
		}
	    }
	}
	else if ( POST_INC == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,x:(%2)+" SC
		       "move	%g1,x:(%2)+";
	    }
	    else
	    {
		return "move	%h1,x:(%2)+" SC
		       "move	%e1,x:(%2)+";
	    }
	}
	else if ( POST_DEC == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,x:(%2)+" SC
		       "move	%g1,x:(%2)-" SC
		       "move	(%2)-" SC
		       "move	(%2)-";
	    }
	    else
	    {
		return "move	%h1,x:(%2)+" SC
		       "move	%e1,x:(%2)-" SC
		       "move	(%2)-" SC
		       "move	(%2)-";
	    }
	}
	else if ( REG == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    operands[2] = XEXP ( operands[0], 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,x:(%2)+" SC
		       "move	%g1,x:(%2)-";
	    }
	    else
	    {
		return "move	%h1,x:(%2)+" SC
		       "move	%e1,x:(%2)-";
	    }
	}
	else
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,%f0" SC
		       "move	%g1,%f0+1";
	    }
	    else
	    {
		return "move	%h1,%f0" SC
		       "move	%e1,%f0+1";
	    }
	}
     
    case 2:
	if ( PLUS == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[1], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    }
	    
	    if ( ! INTVAL ( operands[3] ))
	    {
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		    {
			return "move	x:(%2+1),%g0" SC
			       "move	x:(%2),%0";
		    }
		    else
		    {
			return "move	x:(%2+1),%0" SC
			       "move	x:(%2),%h0";
		    }
		}
		else
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		    {
			return "move	x:(%2)+,%0" SC
			       "move	x:(%2)-,%g0";
		    }
		    else
		    {
			return "lea	(%2)+,%2" SC
			       "move	x:(%2)-,%0" SC
			       "move	x:(%2),%h0";
		    }
		}
	    }
	    else
	    {
		if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
		{
		    if ( 127 > INTVAL ( operands[3] ) &&
			-129 < INTVAL ( operands[3] ))
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
			{
			    return "move	x:(%2+%q3+1),%g0" SC
				   "move	x:(%2+%q3),%0";
			}
			else
			{
			    return "move	x:(%2+%q3+1),%0" SC
				   "move	x:(%2+%q3),%h0";
			}
		    }
		    else
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2" SC
				       "lea	(%2)+,%2" SC
				       "move	x:(%2+%j2),%g0" SC
				       "move	(%2)-" SC
				       "move	x:(%2+%j2),%0";
			    }
			    else
			    {
				return "lea	(%2)+,%2" SC
				       "move	x:(%2+%j2),%g0" SC
				       "move	(%2)-" SC
				       "move	x:(%2+%j2),%0";
			    }
			}
			else
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2" SC
				       "lea	(%2)+,%2" SC
				       "move	x:(%2+%j2),%0" SC
				       "move	(%2)-" SC
				       "move	x:(%2+%j2),%h0";
			    }
			    else
			    {
				return "lea	(%2)+,%2" SC
				       "move	x:(%2+%j2),%0" SC
				       "move	(%2)-" SC
				       "move	x:(%2+%j2),%h0";
			    }
			}
		    }
		}
		else
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2" SC
				   "lea	(%2)+,%2" SC
				   "move	x:(%2+%j2),%g0" SC
				   "move	(%2)-" SC
				   "move	x:(%2+%j2),%0";
			}
			else
			{
			    return "lea	(%2)+,%2" SC
				   "move	x:(%2+%j2),%g0" SC
				   "move	(%2)-" SC
				   "move	x:(%2+%j2),%0";
			}
		    }
		    else
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2" SC
				   "lea	(%2)+,%2" SC
				   "move	x:(%2+%j2),%0" SC
				   "move	(%2)-" SC
				   "move	x:(%2+%j2),%h0";
			}
			else
			{
			    return "lea	(%2)+,%2" SC
				   "move	x:(%2+%j2),%0" SC
				   "move	(%2)-" SC
				   "move	x:(%2+%j2),%h0";
			}
		    }
		}
	    }
	}
	else if ( POST_INC == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		if ( load_n_reg_p ( operands[2], 
				   gen_rtx ( CONST_INT, VOIDmode, 2 )))
		{
		    return "move	#2,%j2" SC
			   "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%g0" SC
			   "move	x:(%2)+%j2,%0";
		}
		else
		{
		    return "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%g0" SC
			   "move	x:(%2)+%j2,%0";
		}
	    }
	    else
	    {
		if ( load_n_reg_p ( operands[2], 
				   gen_rtx ( CONST_INT, VOIDmode, 2 )))
		{
		    return "move	#2,%j2" SC
			   "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%0" SC
			   "move	x:(%2)+%j2,%h0";
		}
		else
		{
		    return "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%0" SC
			   "move	x:(%2)+%j2,%h0";
		}
	    }
	}
	else if ( POST_DEC == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "lea	(%2)+,%2" SC
		       "move	x:(%2)-,%g0" SC
		       "move	x:(%2)-,%0" SC
		       "move	(%2)-";
	    }
	    else
	    {
		return "lea	(%2)+,%2" SC
		       "move	x:(%2)-,%0" SC
		       "move	x:(%2)-,%h0" SC
		       "move	(%2)-";
	    }
	}
	else if ( REG == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    operands[2] = XEXP ( operands[1], 0 );
	    
	    if ( DSP_R2_REGNUM == REGNO ( operands[2] ))
	    {
		if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		{
		    return "move	x:(%2+1),%g0" SC
			   "move	x:(%2),%0";
		}
		else
		{
		    return "move	x:(%2+1),%0" SC
			   "move	x:(%2),%h0";
		}
	    }
	    else
	    {
		if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		{
		    return "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%g0" SC
			   "move	x:(%2),%0";
		}
		else
		{
		    return "lea	(%2)+,%2" SC
			   "move	x:(%2)-,%0" SC
			   "move	x:(%2),%h0";
		}
	    }
	}
	else
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	%f1+1,%g0" SC
		       "move	%f1,%0";
	    }
	    else
	    {
		return "move	%f1+1,%0" SC
		       "move	%f1,%h0";
	    }
	}

    case 3:
	if ( is_float_p )
	{
	    unsigned int memory_value =
		get_float_format
		    (((union real_extract*) & 
		      CONST_DOUBLE_LOW ( operands[1] ))->d );
	
#if defined( __WATCOMC__ )
	    operands[2] =
		gen_rtx ( CONST_INT, VOIDmode, memory_value  & 0x0000ffff );
	
	    operands[3] = 
		gen_rtx ( CONST_INT, VOIDmode, 
			 (( memory_value >> 16 ) & 0x0000ffff ));
#else
	    operands[2] =
		gen_rtx ( CONST_INT, VOIDmode,
			 (( memory_value >> 16 ) & 0x0000ffff ));
	
	    operands[3] = 
		gen_rtx ( CONST_INT, VOIDmode, memory_value & 0x0000ffff );
#endif
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	#>%q2,%g0" SC
		       "move	#>%q3,%0";
	    }
	    else
	    {
		if ( 0 == INTVAL ( operands[3] ))
		{
		    if ( 0 == INTVAL ( operands[2] ))
		    {
			return "clr	%0";
		    }
		    else
		    {
			return "move	#>%q2,%0";
		    }
		}
		else
		{
		    return "move	#>%q2,%0" SC
			   "move	#>%q3,%h0";
		}
	    }
	}
	else /* not a floating constant. */
	{
	    if ( CONST_INT == GET_CODE ( operands[1] ))
	    {
		operands[2] =
		    gen_rtx ( CONST_INT, VOIDmode,
			     ( INTVAL ( operands[1] ) >> 16 ) & 0x0000ffff );

		operands[3] =
		    gen_rtx ( CONST_INT, VOIDmode, 
			     INTVAL ( operands[1] ) & 0x0000ffff );
	    }
	    else
	    {
		operands[2] =
		    gen_rtx ( CONST_INT, VOIDmode, 
			     ( CONST_DOUBLE_LOW ( operands[1] ) >> 16 )
			     & 0x0000ffff );
		operands[3] = 
		    gen_rtx ( CONST_INT, VOIDmode,
			     CONST_DOUBLE_LOW ( operands[1] ) & 0x0000ffff );
	    }
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		if ( rtx_equal_p ( operands[2], operands[3] ))
		{
		    return "move	#>%q2,%g0" SC
			   "move	%g0,%0";
		}
		else
		{
		    return "move	#>%q2,%g0" SC
			   "move	#>%q3,%0";
		}
	    }
	    else
	    {
		if ( 0 == INTVAL ( operands[3] ))
		{
		    if ( 0 == INTVAL ( operands[2] ))
		    {
			return "clr	%0";
		    }
		    else
		    {
			return "move	#>%q2,%0";
		    }
		}
		else
		{
		    if ( rtx_equal_p ( operands[2], operands[3] ))
		    {
			return "move	#>%q2,%0" SC
			       "move	%0,%h0";
		    }
		    else
		    {
			return "move	#>%q2,%0" SC
			       "move	#>%q3,%h0";
		    }
		}
	    }
	}
    }
}

void
local_output_double ( file, value )
    FILE* file;
    double value;
{
    unsigned int memory_value = get_float_format ( value );
    
#if defined( __WATCOMC__ )
    fprintf ( file, "\tdc\t$%04x\n\tdc\t$%04x\n",
	     (( memory_value >> 16 ) & 0xffff ),
	     ( memory_value & 0xffff ));
#else
    fprintf ( file, "\tdc\t$%04x\n\tdc\t$%04x\n",
	     ( memory_value & 0xffff ),
	     (( memory_value >> 16 ) & 0xffff ));
#endif
}

#include "config/dsp.c"
