/* subroutines for insn-output.c for motorola HAWK.
   copyright ( c ) 1988 free software foundation, inc.

this file is part of gnu cc.

gnu cc is distributed in the hope that it will be useful,
but without any warranty.  no author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  refer to the gnu cc general public
license for full details.

everyone is granted permission to copy, modify and redistribute
gnu cc, but only under the conditions described in the
gnu cc general public license.	 a copy of this license is
supposed to have been given to you along with gnu cc so you
can know your rights and responsibilities.  it should be in a
file named copying.  among other things, the copyright notice
and this notice must be preserved on all copies.  */


#ifndef file
#include <stdio.h>
#endif

#include <string.h>
#if defined ( _MSDOS )
extern void *malloc( );
#endif

#include <ctype.h>

#include "flags.h"
#include "tree.h"
#if ! defined( _MSDOS )
#include "hard-reg-set.h"
#else
#include "hardrset.h"
#endif

func_attributes current_func_info = FUNC_NO_INFO;

static char local_call_used_regs[] = CALL_USED_REGISTERS;
enum reg_class regno_reg_class ( );

int FirstSave;

void
PushRegs ( file )
    FILE* file;
{
    int reg_index;

    FirstSave = -1;
    
    for ( reg_index = 0; reg_index < FIRST_PSEUDO_REGISTER; ++ reg_index )
    {
	if ( regs_ever_live[reg_index] && 
	    ! local_call_used_regs[reg_index] )
	{
	    if ( 0 > FirstSave )
	    {
		FirstSave = reg_index;
	    }
	    
	    if ( DST_REGS == REGNO_REG_CLASS ( reg_index ))
	    {
		fprintf ( file, "\
	move	%s0,x:(r2)+\n\
	move	%s1,x:(r2)+\n", reg_names[reg_index], reg_names[reg_index] );
	    }
	    else
	    {
		fprintf ( file, "\
	move	%s,x:(r2)+\n", reg_names[reg_index] );
	    }
	    
	}
    }
}

void
PopRegs ( file )
    FILE* file;
{
    int reg_index = FIRST_PSEUDO_REGISTER;
    int first_save = 0;
    
#define SAVE_CHECK \
{\
    if ( ! first_save )\
    {\
	first_save = 1;\
	fprintf ( file, "\tmove	(r2)-\n" );\
    }\
}


    while ( -- reg_index >= 0 )
    {
	if ( regs_ever_live[reg_index] && 
	    ! local_call_used_regs[reg_index] )
	{
	    SAVE_CHECK;
	    
	    if ( DST_REGS == REGNO_REG_CLASS ( reg_index ))
	    {
		if ( reg_index == FirstSave )
		{
		    fprintf ( file, "\
	move	x:(r2)-,%s\n\
	move	x:(r2),%s0\n", reg_names[reg_index], reg_names[reg_index] );
		}
		else
		{
		    fprintf ( file, "\
	move	x:(r2)-,%s\n\
	move	x:(r2)-,%s0\n", reg_names[reg_index], reg_names[reg_index] );
		}
	    }
	    else
	    {
		if ( reg_index == FirstSave )
		{
		    fprintf ( file, "\
	move	x:(r2),%s\n", reg_names[reg_index] );
		}
		else
		{
		    fprintf ( file, "\
	move	x:(r2)-,%s\n", reg_names[reg_index] );
		}
	    }
	}
    }
}

int
go_if_legitimate_address ( mode, x, strict )
    int mode;
    rtx x;
    int strict;
{
    /* constant pointer */

    if ( CONSTANT_ADDRESS_P ( x ) )
    {
	return 1;
    }

    /* note that REG_OK_STRICT may NOT be defined here! */
    /* register pointer */

    if ( REG_P ( x ) && 
	(( strict ) ? 
	 ( REGNO_OK_FOR_INDEX_P ( REGNO ( x ))) :
	 ( REG_OK_FOR_INDEX_P ( x ))))
    {
	return 1;
    }
    
    /* register arithmetic */
    
    switch ( GET_CODE ( x ) )
    {
    case SUBREG:
	return go_if_legitimate_address ( mode, XEXP ( x, 0 ), strict );
					 
    case POST_DEC:
    case POST_INC:
    case PRE_INC:
    case PRE_DEC:
	if ( REG_P ( XEXP ( x, 0 )) && 
	    (( strict ) ? 
	     ( REGNO_OK_FOR_INDEX_P ( REGNO ( XEXP ( x, 0 )))) :
	     ( REG_OK_FOR_INDEX_P ( XEXP ( x, 0 )))))
	{
	    return 1;
	}
	break;
	
    case PLUS:
	if ( REG_P ( XEXP ( x, 0 )) && 
	    ( Pmode == GET_MODE ( XEXP ( x, 0 ))) &&
	    ( CONST_INT == GET_CODE ( XEXP ( x, 1 ))) &&
	    (( strict ) ? 
	     ( REGNO_OK_FOR_INDEX_P ( REGNO ( XEXP ( x, 0 )))) :
	     ( REG_OK_FOR_INDEX_P ( XEXP ( x, 0 )))))
	{
	    return 1;
	}

	if ( REG_P ( XEXP ( x, 1 )) && 
	    ( Pmode == GET_MODE ( XEXP ( x, 1 ))) &&
	    ( CONST_INT == GET_CODE ( XEXP ( x, 0 ))) &&
	    (( strict ) ? 
	     ( REGNO_OK_FOR_INDEX_P ( REGNO ( XEXP ( x, 1 )))) :
	     ( REG_OK_FOR_INDEX_P ( XEXP ( x, 1 )))))
	{
	    return 1;
	}
    }
    return 0;
}

void
print_operand_address ( file, addr )
    FILE* file;
    rtx addr;
{
    register rtx reg0;
    register rtx reg1;

    switch ( GET_CODE ( addr ) )
    {

    case REG:
	fprintf ( file, "(%s)", reg_names[REGNO ( addr )] );
	break;

    case POST_INC:
	fprintf ( file, "(%s)+", reg_names[REGNO ( XEXP ( addr, 0 ) )] );
	break;

    case POST_DEC:
	fprintf ( file, "(%s)-", reg_names[REGNO ( XEXP ( addr, 0 ) )] );
	break;

    case PLUS:
	fprintf ( file, "(%s+%d)", reg_names[REGNO ( XEXP ( addr, 0 ) )],
		 INTVAL ( XEXP ( addr, 1 )));
	break;
	
    default:
	output_addr_const ( file, addr );
    }
}

void
notice_update_cc ( exp, insn )
    rtx exp;
    rtx insn;
{
    if ( PARALLEL == GET_CODE ( exp ))
    {
	rtx call_insn = XVECEXP ( exp, 0, 0 );
	if ( GET_CODE ( call_insn ) == SET &&
	     GET_CODE ( XEXP ( call_insn, 1 )) == CALL )
	{
	    cc_status.flags = 0;
	    cc_status.value1 = cc0_rtx;
	    cc_status.value2 = gen_rtx ( REG, SImode, HAWK_A_REGNUM );
	    cc_status.mdep = CC_SIGNED;
	}
	else
	{
	    CC_STATUS_INIT;
	}
    }
    else if ( SET == GET_CODE ( exp ))
    {
	switch ( GET_CODE ( SET_DEST ( exp )))
	{
	case CC0:
	    cc_status.flags = 0;
	    cc_status.value1 = SET_DEST ( exp );
	    cc_status.value2 = SET_SRC ( exp );
	    break;
	    
	case REG:
	    CC_STATUS_INIT;
	}
    }
    else
    {
	CC_STATUS_INIT;
    }
}

rtx
next_cc_relevancy ( insn )
    rtx insn;
{
    rtx peek = NEXT_INSN ( insn );
    
    for ( ; NULL != peek; peek = NEXT_INSN ( peek ))
    {
	switch ( GET_CODE( peek ))
	{
	case JUMP_INSN:
	case CALL_INSN:
	case CODE_LABEL:
	case BARRIER:
	    return peek;
	    
	case INSN:
	    if (( PARALLEL == GET_CODE ( PATTERN ( peek ))) &&
		( SET == GET_CODE ( XVECEXP ( PATTERN ( peek ), 0, 0 ))) &&
		( 0 == strcmp ( GET_RTX_FORMAT ( GET_CODE ( XEXP ( XVECEXP ( PATTERN ( peek ), 0, 0 ), 1 ))), "ee" )) &&
		( cc0_rtx == XEXP ( XEXP ( XVECEXP ( PATTERN ( peek ), 0, 0 ), 1 ), 0 )))
	    {
		return peek;
	    }
	}
    }
    return insn;
}

enum mdep_cc_info
next_cc_use ( insn )
    rtx insn;
{
    enum rtx_code jump_code;
    rtx peek = NEXT_INSN ( insn );
    
    for ( ; NULL != peek; peek = NEXT_INSN ( peek ))
    {
	if ( JUMP_INSN == GET_CODE ( peek ))
	{
	    break;
	}
	if ( CALL_INSN == GET_CODE ( peek ) ||
	    CODE_LABEL == GET_CODE ( peek ) ||
	    BARRIER == GET_CODE ( peek ) )
	{
	    return CC_UNKNOWN;
	}
	if ( INSN == GET_CODE ( peek ) &&
	    PARALLEL == GET_CODE ( PATTERN ( peek )) &&
	    SET == GET_CODE ( XVECEXP ( PATTERN ( peek ), 0, 0 )) &&
	    0 == strcmp ( GET_RTX_FORMAT ( GET_CODE ( XEXP ( XVECEXP ( PATTERN ( peek ), 0, 0 ), 1 ))), "ee" ) &&
	    ( cc0_rtx == XEXP ( XEXP ( XVECEXP ( PATTERN ( peek ), 0, 0 ), 1 ), 0 ) ||
	     cc0_rtx == XEXP ( XEXP ( XVECEXP ( PATTERN ( peek ), 0, 0 ), 1 ), 1 )))
	{
	    jump_code = GET_CODE ( XEXP ( XVECEXP ( PATTERN ( peek ), 0, 0 ), 1 ));
	    
	    if ( GEU == jump_code || GTU == jump_code || LEU == jump_code
		|| LTU == jump_code )
	    {
		return CC_UNSIGNED;
	    }
	    else
	    {
		return CC_SIGNED;
	    }
	}
    }

    if (( NULL == peek ) || 
	( JUMP_INSN != GET_CODE ( peek )) ||
	( SET != GET_CODE ( PATTERN ( peek ))) ||
	( IF_THEN_ELSE != GET_CODE ( XEXP ( PATTERN ( peek ), 1 ))))
    {
	return CC_UNKNOWN;
    }
    jump_code = GET_CODE ( XEXP ( XEXP ( PATTERN ( peek ), 1 ), 0 ));
    
    if ( GEU == jump_code || GTU == jump_code || LEU == jump_code
	|| LTU == jump_code || EQU == jump_code || NEU == jump_code )
    {
	return CC_UNSIGNED;
    }
    else
    {
	return CC_SIGNED;
    }
}

/* hard_regno_mode_ok -
 * this global function implements the tm-m56k.h file macro of the same name.
 * a 1 is returned if regno can be used to hold a value of type mode.
 */

int
hard_regno_mode_ok ( regno, mode )
    int regno;
    int mode;
{
    switch ( mode )
    {
    case VOIDmode:
	return 1;
	
    case SFmode:
    case DFmode:
    case DImode:
	/* for doubles, floats, and longs, we need 32-bit registers. */
	if ( HAWK_X0_REGNUM == regno || HAWK_Y0_REGNUM == regno ||
	    HAWK_A_REGNUM == regno || HAWK_B_REGNUM == regno )
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
    /* for all other cases, we assume that a 16-bit reg is needed. */
    return 1;
}

/* modes_tieable_p -
 * this global function implements the tm-m56k.h file macro of the same name.
 * a 1 is returned if both modes can always use the same pseudo register.
 */

int
modes_tieable_p ( mode1, mode2 )
    int mode1;
    int mode2;
{
    if ( mode1 == mode2 )
    {
	return 1;
    }
    
    switch ( mode1 )
    {
    case SFmode:
    case DFmode:
    case DImode:
	if ( SFmode == mode2 || DFmode == mode2 || DImode == mode2 )
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
	break;
    }
    return 0;
}

/* regno_reg_class -
 * this global function implements the tm-m56k.h file macro of the same name.
 * the class number of the smallest class containing reg number regno is
 * returned.
 */

enum reg_class
regno_reg_class ( regno )
    int regno;
{
    switch ( regno )
    {
    case HAWK_X1_REGNUM:
	return MPY1_REGS;
	
    case HAWK_Y1_REGNUM:
	return MPY2_REGS;
	
    case HAWK_X0_REGNUM:
	return X_DBL_REGS;
	
    case HAWK_Y0_REGNUM:
	return Y_DBL_REGS;
	
    case HAWK_A_REGNUM:
    case HAWK_B_REGNUM:
	return DST_REGS;

    default:
	return ADDR_REGS;
    }
}


/* reg_class_from_letter -
 * this global function implements the tm-m56k.h file macro of the same name.
 * the class number of the given letter is returned.
 */

enum reg_class
reg_class_from_letter ( c )
    char c;
{
    switch ( c )
    {
    case 'q':
	return SRC1_REGS;
    case 'D':
	return DST_REGS;
    case 'A':
	return ADDR_REGS;
    case 't':
	return MPY1_REGS;
    case 'T':
	return MPY2_REGS;
    case 'Q':
	return SRC_REGS;
    case 'x':
	return X_DBL_REGS;
    case 'y':
	return Y_DBL_REGS;
    case 'R':
	return INC_REGS;
	
    default:
	return NO_REGS;
    }
}

/* reg_class_from_letter -
 * this global function implements the tm-m56k.h file macro of the same name.
 * the max number of consecutive registers needed per mode is returned.
 */

int
class_max_nregs ( class, mode )
    int class;
    int mode;
{
    if (( SFmode == mode || DFmode == mode || DImode == mode ) &&
	( SRC1_REGS == class || MPY1_REGS == class || SRC_REGS == class ||
	 MPY2_REGS == class || X_DBL_REGS == class || Y_DBL_REGS == class ))
    {
	return 2;
    }
    else
    {
	return 1;
    }
}

/* the hawk uses the IEEE single precision format. this routine assumes
   that the host has 32 bit integers and that float is ieee sp. */
unsigned int
get_float_format ( number )
    double number;
{
    float ieee_sp = (float) number;
    unsigned int *peek = (unsigned int*) & ieee_sp;
    
#if defined( _MSDOS ) || defined( MIPSEL )
    return ( 0x0000ffff & ((*peek) >> 16 )) | ( 0xffff0000 & ((*peek) << 16 ));
#else
    return *peek;
#endif
}

static rtx
ensure_signed_int_offsets ( x )
    rtx x;
{
    if ( CONST_INT == GET_CODE ( x ))
    {
 	INTVAL ( x ) =(( INTVAL ( x ) << ( HOST_BITS_PER_INT - BITS_PER_UNIT ))
		       >> ( HOST_BITS_PER_INT - BITS_PER_UNIT ));
    }
    else
    {
	char *p, *fmt = GET_RTX_FORMAT ( GET_CODE ( x ));
	
	p = fmt;
	
	while ( '\0' != ( *p ))
	{
	    if ( 'e' == ( *p ))
	    {
		XEXP ( x, p - fmt ) = 
		    ensure_signed_int_offsets ( XEXP ( x, p - fmt ));
	    }
	    ++ p;
	}
    }
    return x;
}

static int
find_symbol ( x )
    rtx x;
{
    int found_symbol = 0;
    char *p, *fmt = GET_RTX_FORMAT ( GET_CODE ( x ));

    if ( SYMBOL_REF == GET_CODE ( x ))
    {
	return 1;
    }
    
    p = fmt;
    
    while ( '\0' != ( *p ))
    {
	if ( 'e' == ( *p ))
	{
	    found_symbol |= find_symbol ( XEXP ( x, p - fmt ));
	}
	++ p;
    }
    return found_symbol;
}

rtx
ensure_integer_result_size ( x )
    rtx x;
{
    if ( find_symbol ( x ) || CONST_INT == GET_CODE ( x ))
    {
	x = ensure_signed_int_offsets ( x );
    }
    else
    {
	char *p, *fmt = GET_RTX_FORMAT ( GET_CODE ( x ));
	
	p = fmt;
	
	while ( '\0' != ( *p ))
	{
	    if ( 'e' == ( *p ))
	    {
		XEXP ( x, p - fmt ) = 
		    ensure_integer_result_size ( XEXP ( x, p - fmt ));
	    }
	    ++ p;
	}
    }
    return x;
}

void
print_operand ( file, x, code )
    FILE* file;
    rtx x;
    char code;
{
    if ( 'q' == code )
    {
	x = ensure_integer_result_size ( copy_rtx ( x ));

	output_addr_const ( file, x );
	return;
    }
    
    switch ( GET_CODE ( x ))
    {
    case REG:
	switch ( code )
	{
	case 'd':
	    fprintf ( file, "%s", reg_names[REGNO ( x )] );
	    break;
	    
	case 'e':
	    if ( DST_REGS == REGNO_REG_CLASS ( REGNO ( x )))
	    {
		if ( MEM_IN_STRUCT_P ( x ))
		{
		    /* see comment in final.c */
		    fprintf ( file, "%s0", reg_names[REGNO ( x )] );
		}
		else
		{
		    fprintf ( file, "%s1", reg_names[REGNO ( x )] );
		}
	    }
	    else
	    {
		fprintf ( file, "%s", reg_names[REGNO ( x )] );
	    }
	    break;
	    
	case 'g':
	    fprintf ( file, "%s", reg_names[REGNO ( x ) ^ 1] );
	    break;
	    
	case 'h':
	    fprintf ( file, "%s0", reg_names[REGNO ( x )] );
	    break;

	case 'i':
	    fprintf ( file, "%c", *( reg_names[REGNO ( x )] ));
	    break;

	case 'j':
	    fprintf ( file, "n%c", * ( reg_names[REGNO ( x )] + 1 ));
	    break;
	    
	case 'k':
	    fprintf ( file, "%s2", reg_names[REGNO ( x )] );
	    break;
	    
	case 'm':
	    fprintf ( file, "%c", (reg_names[REGNO ( x )])[0] );
	    break;
	    
	case 'o':
	    fprintf ( file, "%c10", (reg_names[REGNO ( x )])[0] );
	    break;
	    
	case 't':
	    fputc ( reg_names[REGNO ( x )][0], file );
	    fputc ( '1', file );
	    break;
	    
	default:
	    if ( MEM_IN_STRUCT_P ( x ))
	    {
		/* see comment in final.c */
		fprintf ( file, "%s0", reg_names[REGNO ( x )] );
	    }
	    else
	    {
		fprintf ( file, "%s", reg_names[REGNO ( x )] );
	    }
	    break;
	}
	break;
	
    case MEM:
	switch ( code )
	{
	case 'f':
	    /* print out a 'l' instead of an x or y when we need to */
	    fprintf ( file, "x:" );
	    output_address ( XEXP ( x, 0 ));
	    break;
	    
	default:
	    output_address ( XEXP ( x, 0 ));
	}
	return;

    case CONST_INT:
	if ( 0 == ( INTVAL ( x ) >> 8 ))
	{
	    fprintf ( file, "#>$%02x", INTVAL ( x ) );
	}
	else
	{
	    fprintf ( file, "#>$%04x", INTVAL ( x ) );
	}
	return;

    default:
#if 0
	fprintf ( file, "#" );
	if ( CONSTANT_ADDRESS_P ( x ))
	{
	    output_addr_const ( file, x );
	}
	else
	{
	    output_operand ( x, 'c' );
	}
#endif
	abort ( );
    }
}

void
simplify_cse_victims ( first )
    rtx first;
{
    rtx point, elem;
    
    while ( first )
    {
	/* is it a 3 element PARALLEL ? */
	if (( INSN == GET_CODE ( first )) &&
	    ( point = PATTERN ( first ), PARALLEL == GET_CODE ( point )) &&
	    ( 3 == XVECLEN ( point, 0 )) &&

	    /* is the 1st element a ( SET ( REG ) ( REG / CONST_INT )) ? */
	    ( elem = XVECEXP ( point, 0, 0 ), SET == GET_CODE ( elem )) &&
	    ( REG == GET_CODE ( SET_DEST ( elem ))) &&
	    (( REG == GET_CODE ( SET_SRC ( elem ))) ||
	     ( CONST == GET_CODE ( SET_SRC ( elem ))) ||
	     ( CONST_INT == GET_CODE ( SET_SRC ( elem )))) &&

	    /* is the 2nd element a ( CLOBBER ( REG )) ? */
	    ( elem = XVECEXP ( point, 0, 1 ), CLOBBER == GET_CODE ( elem )) &&
	    ( REG == GET_CODE ( XEXP ( elem, 0 ))) &&

	    /* is the 3rd element a ( CLOBBER ( REG )) ? */
	    ( elem = XVECEXP ( point, 0, 2 ), CLOBBER == GET_CODE ( elem )) &&
	    ( REG == GET_CODE ( XEXP ( elem, 0 ))))
	{
	    /* we've got a blown-out div template ! substitute with a simple
	       ( SET ( REG ) ( REG / CONST_INT )) (note: really the 1st 
	       element of the parallel). */

	    PATTERN ( first ) = XVECEXP ( point, 0, 0 );
	}
	first = NEXT_INSN ( first );
    }
}


static int
mpy_insn_p ( insn )
    rtx insn;
{
    return (( INSN == GET_CODE ( insn )) &&
	    ( SET == GET_CODE ( PATTERN ( insn ))) &&
	    ( REG == GET_CODE ( SET_DEST ( PATTERN ( insn )))) &&
	    ( MULT == GET_CODE ( SET_SRC ( PATTERN ( insn )))));
}

static int
mac_insn_p ( insn )
    rtx insn;
{
    /* mac +/- shape detector. */
    return (( INSN == GET_CODE ( insn )) &&
	    ( SET == GET_CODE ( PATTERN ( insn ))) &&
	    ( REG == GET_CODE ( SET_DEST ( PATTERN ( insn )))) &&
	    (( PLUS == GET_CODE ( SET_SRC ( PATTERN ( insn )))) ||
	     ( MINUS == GET_CODE ( SET_SRC ( PATTERN ( insn ))))) &&
	    (( MULT == GET_CODE ( XEXP ( SET_SRC ( PATTERN ( insn )), 0 ))) ||
	     ( MULT == GET_CODE ( XEXP ( SET_SRC ( PATTERN ( insn )), 1 )))));
}

static rtx
mac_addative_input ( mac )
    rtx mac;
{
    return ( MULT == GET_CODE ( XEXP ( SET_SRC ( PATTERN ( mac )), 0 ))) ?
	XEXP ( SET_SRC ( PATTERN ( mac )), 1 ) :
	    XEXP ( SET_SRC ( PATTERN ( mac )), 0 );
}

void
reduce_norm_freq ( first )
    rtx first;
{
    /* point to the last insn.
       search up the list for each mac insn:
           detemine the addative_input.
	   foreach insn in LOG_LINKS( point ):
	       if insn SETs the addative_input, and insn is a mac or mpy, and
	          there are no reload generated references to the operand
		  between these two instructions, then:
	           mark the pointed to mac as not needing input normalization
		   mark the mpy/mac insn as not needing output normalization.
       
       the templates for mpy/mac will check for these flags. */

    rtx last = first, point;
    
    /* find the last insn, search up from it */
    while ( NULL != NEXT_INSN ( last ))
    {
	last = NEXT_INSN ( last );
    }

    while ( last != first )
    {
	if ( mac_insn_p ( last ))
	{
	    rtx add_in = mac_addative_input ( last );
	    rtx insn_chain = LOG_LINKS ( last );
	    
	    while ( NULL != insn_chain )
	    {
		rtx this_insn = XEXP ( insn_chain, 0 );

		if ((( mac_insn_p ( this_insn )) ||
		     ( mpy_insn_p ( this_insn ))) &&
		    ( add_in == SET_DEST ( PATTERN ( this_insn ))) &&
		    ( ! reg_used_between_p ( add_in, this_insn, last )))
		{
		    /* kludge up: re-use existing flags, in_struct and used, 
		       to inhibit the emission of normalization code at the
		       front and back end of mac (and mpy) instructions. */

		    /* turn off last's pre-norm. */
		    MEM_IN_STRUCT_P ( last ) = 1;
		    
		    /* turn off this_insn's post-norm. */
		    RTX_UNCHANGING_P ( this_insn ) = 1;
		}
		insn_chain = XEXP ( insn_chain, 1 );
	    }
	}
	last = PREV_INSN ( last );
    }
}

/* this extraneous function is needed because the MetaWare C compiler
   can't handle the comma operator with pointer values. */

rtx
local_function_incoming_arg ( cum, mode, type, named )
    CUMULATIVE_ARGS cum;
    enum machine_mode mode;
    tree type;
    int named;
{
    current_func_info |= FUNC_HAS_STACK_PARMS;

    return FUNCTION_ARG (cum, mode, type, named);
}

#define N_REG_COUNT 4
#define FIRST_N_REG 6

static rtx current_n_reg_value[N_REG_COUNT];

void
clear_n_reg_values ( )
{
    int i = N_REG_COUNT;
    
    while ( i -- )
    {
	current_n_reg_value[i] = NULL;
    }
}

int
load_n_reg_p ( reg, new_value )
    rtx reg, new_value;
{
    if ( current_n_reg_value[REGNO ( reg ) - FIRST_N_REG] &&
	rtx_equal_p ( current_n_reg_value[REGNO ( reg ) - FIRST_N_REG], 
		      new_value ) &&
	CONSTANT_P ( new_value ))
    {
	return 0;
    }
    else
    {
	current_n_reg_value[REGNO ( reg ) - FIRST_N_REG] =  new_value;

	return 1;
    }
}
	
/* sort the compares into the signed and unsigned varieties. */
void
sort_compares ( insns )
    rtx insns;
{
    while ( insns )
    {
	if (( INSN == GET_CODE ( insns )) &&
	    ( SET == GET_CODE ( PATTERN ( insns ))) &&
	    ( cc0_rtx == SET_DEST ( PATTERN ( insns ))) &&
	    ( CC_UNSIGNED == next_cc_use ( insns )))
	{
	    UNSIGNED_COMPARE_P ( insns ) = 1;
	    INSN_CODE ( insns ) = -1;
	}
	insns = NEXT_INSN ( insns );
    }
}

char*
move_singleword ( operands )
    rtx *operands;
{
    if (( rtx_equal_p ( operands[0], operands[1] )) &&
	(( REG != GET_CODE ( operands[0] )) ||
	 ( ! MEM_IN_STRUCT_P ( operands[0] ))) &&
	(( REG != GET_CODE ( operands[1] )) ||
	 ( ! MEM_IN_STRUCT_P ( operands[1] ))))
    {
	/* if we're doing a nop move, return a null string. note that
	   an insn could look like a nop move isasmuchas rtx_equal_p returns
	   true; however the MEM_IN_STRUCT_P kludge used by alter_subreg ( )
	   (final.c) may indicate a move from the LSW of a long to an int
	   in the same accumulator. */

	return "";
    }
    
    switch ( which_alternative )
    {
    case 0:
	if (( ! MEM_IN_STRUCT_P ( operands[1] )) &&
	    ( DST_REGS == REGNO_REG_CLASS ( REGNO ( operands[0] ))) &&
	    ( DST_REGS == REGNO_REG_CLASS ( REGNO ( operands[1] )) ||
	     IS_SRC_OR_MPY_P ( REGNO ( operands[1] ))))
	{
	    return "tfr	%1,%0";
	}
	else
	{
	    return "move	%e1,%0";
	}

    case 1:
	if ( PLUS == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[0], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    }
	    
	    switch ( INTVAL ( operands[3] ))
	    {
	    case 0:
		return "move	%e1,x:(%2)";
		    
	    case 1:
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	%e1,x:(%2+%q3)";
		}
		else if ( REGNO ( operands[2] ) != REGNO ( operands[1] ))
		{
		    return "lea	(%2)+,%2\n\tmove	%e1,x:(%2)-";
		}
		    
	    case -1:
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	%e1,x:(%2+%q3)";
		}
		else if ( REGNO ( operands[2] ) != REGNO ( operands[1] ))
		{
		    return "move	(%2)-\n\tmove	%e1,x:(%2)+";
		}
		
	    default:
		/* r2 + constant is limited to a single signed byte, otherwise
		   conventional (for the 56k) addressing must be used. */
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ) &&
		    128 > INTVAL ( operands[3] ) &&
		    -129 < INTVAL ( operands[3] ))
		{
		    return "move	%e1,x:(%2+%q3)";
		}
		else if ( load_n_reg_p ( operands[2], operands[3] ))
		{
		    return "move	#%q3,%j2\n\tmove	%e1,x:(%2+%j2)";
		}
		else
		{
		    return "move	%e1,x:(%2+%j2)";
		}
	    }
	}
	else
	{
	    return "move	%e1,%f0";
	}
	
    case 2:
	if ( PLUS == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[1], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    }
	    
	    switch ( INTVAL ( operands[3] ))
	    {
	    case 0:
		return "move	x:(%2),%0";
		    
	    case 1:
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	x:(%2+%q3),%0";
		}
		else if ( REGNO ( operands[0] ) != REGNO ( operands[2] ))
		{
		    return "lea	(%2)+,%2\n\tmove	x:(%2)-,%0";
		}
		    
	    case -1:
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    return "move	x:(%2+%q3),%0";
		}
		else if ( REGNO ( operands[0] ) != REGNO ( operands[2] ))
		{
		    return "move	(%2)-\n\tmove	x:(%2)+,%0";
		}
		
	    default:
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ) &&
		    128 > INTVAL ( operands[3] ) &&
		    -129 < INTVAL ( operands[3] ))
		{
		    return "move	x:(%2+%q3),%0";
		}
		else if ( load_n_reg_p ( operands[2], operands[3] ))
		{
		    return "move	#%q3,%j2\n\tmove	x:(%2+%j2),%0";
		}
		else
		{
		    return "move	x:(%2+%j2),%0";
		}
	    }
	}
	else
	{
	    return "move	%f1,%0";
	}
	
    case 3:
	if ( 0 == INTVAL ( operands[1] ))
	{
	    if ( DST_REGS == REGNO_REG_CLASS ( REGNO ( operands[0] )))
	    {
		return "clr	%0";
	    }
	    else
	    {
		return "move	#%q1,%0";
	    }
	}
	else
	{
	    return "move	#>%q1,%0";
	}
    }
}

char*
move_doubleword ( is_float_p, operands )
    int is_float_p;
    rtx *operands;
{
    extern union real_extract adjusted_dval;

    if ( rtx_equal_p ( operands[0], operands[1] ))
    {
	return "";
    }
    
    switch ( which_alternative )
    {
    case 0:
	if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	%g1,%g0\n\tmove	%1,%0";
	    }
	    else
	    {
		return "tfr	%i1,%0";
	    }
	}
	else
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "tfr	%1,%i0";
	    }
	    else
	    {
		return "tfr	%1,%0";
	    }
	}

    case 1:
	if ( PLUS == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[0], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[0], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    }
	    
	    if ( ! INTVAL ( operands[3] ))
	    {
		if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
		{
		    return "move	%1,x:(%2)+\n\tmove	%g1,x:(%2)-";
		}
		else
		{
		    return "move	%h1,x:(%2)+\n\tmove	%e1,x:(%2)-";
		}
	    }
	    else
	    {
		/* r2 + constant is limited to a single signed byte, otherwise
		   conventional (for the 56k) addressing must be used. */

		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    if ( 127 > INTVAL ( operands[3] ) &&
			-128 < INTVAL ( operands[3] ))
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
			{
			    return "move	%g1,x:(%2+%q3+1)\n\tmove	%1,x:(%2+%q3)";
			}
			else
			{
			    return "move	%e1,x:(%2+%q3+1)\n\tmove	%h1,x:(%2+%q3)";
			}
		    }
		    else
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	%g1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%1,x:(%2+%j2)";
			    }
			    else
			    {
				return "lea	(%2)+,%2\n\tmove	%g1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%1,x:(%2+%j2)";
			    }
			}
			else
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	%e1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%h1,x:(%2+%j2)";
			    }
			    else
			    {
				return "lea	(%2)+,%2\n\tmove	%e1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%h1,x:(%2+%j2)";
			    }
			}
		    }
		}
		else
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	%g1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%1,x:(%2+%j2)";
			}
			else
			{
			    return "lea	(%2)+,%2\n\tmove	%g1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%1,x:(%2+%j2)";
			}
		    }
		    else
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	%e1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%h1,x:(%2+%j2)";
			}
			else
			{
			    return "lea	(%2)+,%2\n\tmove	%e1,x:(%2+%j2)\n\tmove	(%2)-\n\tmove	%h1,x:(%2+%j2)";
			}
		    }
		}
	    }
	}
	else if ( POST_INC == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,x:(%2)+\n\tmove	%g1,x:(%2)+";
	    }
	    else
	    {
		return "move	%h1,x:(%2)+\n\tmove	%e1,x:(%2)+";
	    }
	}
	else if ( POST_DEC == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[0], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,x:(%2)+\n\tmove	%g1,x:(%2)-\n\tmove	(%2)-\n\tmove	(%2)-";
	    }
	    else
	    {
		return "move	%h1,x:(%2)+\n\tmove	%e1,x:(%2)-\n\tmove	(%2)-\n\tmove	(%2)-";
	    }
	}
	else if ( REG == GET_CODE ( XEXP ( operands[0], 0 )))
	{
	    operands[2] = XEXP ( operands[0], 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,x:(%2)+\n\tmove	%g1,x:(%2)-";
	    }
	    else
	    {
		return "move	%h1,x:(%2)+\n\tmove	%e1,x:(%2)-";
	    }
	}
	else
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[1] )))
	    {
		return "move	%1,%f0\n\tmove	%g1,%f0+1";
	    }
	    else
	    {
		return "move	%h1,%f0\n\tmove	%e1,%f0+1";
	    }
	}
     
    case 2:
	if ( PLUS == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    if ( REG == GET_CODE ( XEXP ( XEXP ( operands[1], 0 ), 0 )))
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 1 );
	    }
	    else
	    {
		operands[2] = XEXP ( XEXP ( operands[1], 0 ), 1 );
		operands[3] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    }
	    
	    if ( ! INTVAL ( operands[3] ))
	    {
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		    {
			return "move	x:(%2+1),%g0\n\tmove	x:(%2),%0";
		    }
		    else
		    {
			return "move	x:(%2+1),%0\n\tmove	x:(%2),%h0";
		    }
		}
		else
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		    {
			return "move	x:(%2)+,%0\n\tmove	x:(%2)-,%g0";
		    }
		    else
		    {
			return "lea	(%2)+,%2\n\tmove	x:(%2)-,%0\n\tmove	x:(%2),%h0";
		    }
		}
	    }
	    else
	    {
		if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
		{
		    if ( 127 > INTVAL ( operands[3] ) &&
			-129 < INTVAL ( operands[3] ))
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
			{
			    return "move	x:(%2+%q3+1),%g0\n\tmove	x:(%2+%q3),%0";
			}
			else
			{
			    return "move	x:(%2+%q3+1),%0\n\tmove	x:(%2+%q3),%h0";
			}
		    }
		    else
		    {
			if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	x:(%2+%j2),%g0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%0";
			    }
			    else
			    {
				return "lea	(%2)+,%2\n\tmove	x:(%2+%j2),%g0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%0";
			    }
			}
			else
			{
			    if ( load_n_reg_p ( operands[2], operands[3] ))
			    {
				return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	x:(%2+%j2),%0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%h0";
			    }
			    else
			    {
				return "lea	(%2)+,%2\n\tmove	x:(%2+%j2),%0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%h0";
			    }
			}
		    }
		}
		else
		{
		    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	x:(%2+%j2),%g0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%0";
			}
			else
			{
			    return "lea	(%2)+,%2\n\tmove	x:(%2+%j2),%g0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%0";
			}
		    }
		    else
		    {
			if ( load_n_reg_p ( operands[2], operands[3] ))
			{
			    return "move	#%q3,%j2\n\tlea	(%2)+,%2\n\tmove	x:(%2+%j2),%0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%h0";
			}
			else
			{
			    return "lea	(%2)+,%2\n\tmove	x:(%2+%j2),%0\n\tmove	(%2)-\n\tmove	x:(%2+%j2),%h0";
			}
		    }
		}
	    }
	}
	else if ( POST_INC == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		if ( load_n_reg_p ( operands[2], 
				   gen_rtx ( CONST_INT, VOIDmode, 2 )))
		{
		    return "move	#2,%j2\n\tlea	(%2)+,%2\n\tmove	x:(%2)-,%g0\n\tmove	x:(%2)+%j2,%0";
		}
		else
		{
		    return "lea	(%2)+,%2\n\tmove	x:(%2)-,%g0\n\tmove	x:(%2)+%j2,%0";
		}
	    }
	    else
	    {
		if ( load_n_reg_p ( operands[2], 
				   gen_rtx ( CONST_INT, VOIDmode, 2 )))
		{
		    return "move	#2,%j2\n\tlea	(%2)+,%2\n\tmove	x:(%2)-,%0\n\tmove	x:(%2)+%j2,%h0";
		}
		else
		{
		    return "lea	(%2)+,%2\n\tmove	x:(%2)-,%0\n\tmove	x:(%2)+%j2,%h0";
		}
	    }
	}
	else if ( POST_DEC == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    operands[2] = XEXP ( XEXP ( operands[1], 0 ), 0 );
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "lea	(%2)+,%2\n\tmove	x:(%2)-,%g0\n\tmove	x:(%2)-,%0\n\tmove	(%2)-";
	    }
	    else
	    {
		return "lea	(%2)+,%2\n\tmove	x:(%2)-,%0\n\tmove	x:(%2)-,%h0\n\tmove	(%2)-";
	    }
	}
	else if ( REG == GET_CODE ( XEXP ( operands[1], 0 )))
	{
	    operands[2] = XEXP ( operands[1], 0 );
	    
	    if ( HAWK_R2_REGNUM == REGNO ( operands[2] ))
	    {
		if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		{
		    return "move	x:(%2+1),%g0\n\tmove	x:(%2),%0";
		}
		else
		{
		    return "move	x:(%2+1),%0\n\tmove	x:(%2),%h0";
		}
	    }
	    else
	    {
		if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
		{
		    return "lea	(%2)+,%2\n\tmove	x:(%2)-,%g0\n\tmove	x:(%2),%0";
		}
		else
		{
		    return "lea	(%2)+,%2\n\tmove	x:(%2)-,%0\n\tmove	x:(%2),%h0";
		}
	    }
	}
	else
	{
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	%f1+1,%g0\n\tmove	%f1,%0";
	    }
	    else
	    {
		return "move	%f1+1,%0\n\tmove	%f1,%h0";
	    }
	}

    case 3:
	if ( is_float_p )
	{
	    unsigned int memory_value =
		get_float_format
		    (((union real_extract*) & 
		      CONST_DOUBLE_LOW ( operands[1] ))->d );
	
#if defined( __WATCOMC__ )
	    operands[2] =
		gen_rtx ( CONST_INT, VOIDmode, memory_value  & 0x0000ffff );
	
	    operands[3] = 
		gen_rtx ( CONST_INT, VOIDmode, 
			 (( memory_value >> 16 ) & 0x0000ffff ));
#else
	    operands[2] =
		gen_rtx ( CONST_INT, VOIDmode,
			 (( memory_value >> 16 ) & 0x0000ffff ));
	
	    operands[3] = 
		gen_rtx ( CONST_INT, VOIDmode, memory_value & 0x0000ffff );
#endif
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		return "move	#>%q2,%g0\n\tmove	#>%q3,%0";
	    }
	    else
	    {
		if ( 0 == INTVAL ( operands[3] ))
		{
		    if ( 0 == INTVAL ( operands[2] ))
		    {
			return "clr	%0";
		    }
		    else
		    {
			return "move	#>%q2,%0";
		    }
		}
		else
		{
		    return "move	#>%q2,%0\n\tmove	#>%q3,%h0";
		}
	    }
	}
	else /* not a floating constant. */
	{
	    if ( CONST_INT == GET_CODE ( operands[1] ))
	    {
		operands[2] =
		    gen_rtx ( CONST_INT, VOIDmode,
			     ( INTVAL ( operands[1] ) >> 16 ) & 0x0000ffff );

		operands[3] =
		    gen_rtx ( CONST_INT, VOIDmode, 
			     INTVAL ( operands[1] ) & 0x0000ffff );
	    }
	    else
	    {
		operands[2] =
		    gen_rtx ( CONST_INT, VOIDmode, 
			     ( CONST_DOUBLE_LOW ( operands[1] ) >> 16 )
			     & 0x0000ffff );
		operands[3] = 
		    gen_rtx ( CONST_INT, VOIDmode,
			     CONST_DOUBLE_LOW ( operands[1] ) & 0x0000ffff );
	    }
	    
	    if ( IS_SRC_OR_MPY_P ( REGNO ( operands[0] )))
	    {
		if ( rtx_equal_p ( operands[2], operands[3] ))
		{
		    return "move	#>%q2,%g0\n\tmove	%g0,%0";
		}
		else
		{
		    return "move	#>%q2,%g0\n\tmove	#>%q3,%0";
		}
	    }
	    else
	    {
		if ( 0 == INTVAL ( operands[3] ))
		{
		    if ( 0 == INTVAL ( operands[2] ))
		    {
			return "clr	%0";
		    }
		    else
		    {
			return "move	#>%q2,%0";
		    }
		}
		else
		{
		    if ( rtx_equal_p ( operands[2], operands[3] ))
		    {
			return "move	#>%q2,%0\n\tmove	%0,%h0";
		    }
		    else
		    {
			return "move	#>%q2,%0\n\tmove	#>%q3,%h0";
		    }
		}
	    }
	}
    }
}

void
local_output_double ( file, value )
    FILE* file;
    double value;
{
    unsigned int memory_value = get_float_format ( value );
    
#if defined( __WATCOMC__ )
    fprintf ( file, "\tdc\t$%04x\n\tdc\t$%04x\n",
	     (( memory_value >> 16 ) & 0xffff ),
	     ( memory_value & 0xffff ));
#else
    fprintf ( file, "\tdc\t$%04x\n\tdc\t$%04x\n",
	     ( memory_value & 0xffff ),
	     (( memory_value >> 16 ) & 0xffff ));
#endif
}

int
local_fix_frame_pointer_address ( addr_ptr, depth )
    rtx *addr_ptr;
    int depth;
{
    int frame_offset, stack_offset, i;
    
    if ( PLUS == GET_CODE ( *addr_ptr ))
    {
	if ( REG == GET_CODE ( XEXP ( *addr_ptr, 0 )) &&
	    FRAME_POINTER_REGNUM == REGNO ( XEXP ( *addr_ptr, 0 )))
	{
	    frame_offset = INTVAL ( XEXP ( *addr_ptr, 1 ));
	}
	else if ( REG == GET_CODE ( XEXP ( *addr_ptr, 1 )) &&
		 FRAME_POINTER_REGNUM == REGNO ( XEXP ( *addr_ptr, 1 )))
	{
	    frame_offset = INTVAL ( XEXP ( *addr_ptr, 0 ));
	}
	else
	{
	    return 0;
	}
    }
    else if ( REG == GET_CODE ( *addr_ptr ) &&
	     FRAME_POINTER_REGNUM == REGNO ( *addr_ptr ))
    {
	frame_offset = 0;
    }
    else
    {
	return 0;
    }

    /* parameters may be one unit further away from the stack pointer due
       to a saved ssh. paramter references always have negative offsets
       from the frame pointer. */

    stack_offset = frame_offset - depth - get_frame_size ( );
    
    if (( current_func_info & FUNC_ISNT_LEAF ) || 
	! TARGET_CALL_OVERHEAD_REDUCTION )
    {
	-- stack_offset;
    }

    /* we must also account for the registers that have been saved, since
       they're between the stack pointer and all local data and paramters. */

    for ( i = 0; i < FIRST_PSEUDO_REGISTER; ++ i )
    {
	if ( HAWK_FAKE_REGNUM == i )
	{
	    continue;
	}
	if ( regs_ever_live[i] && ! local_call_used_regs[i] )
	{
	    -- stack_offset;
	    
	    /* it takes two words to save an accumulator. */
	    if ( DST_REGS == REGNO_REG_CLASS ( i ))
	    {
		-- stack_offset;
	    }
	}
    }
	    
    if ( ! stack_offset )
    {
	*addr_ptr = stack_pointer_rtx;
    }
    else
    {
	*addr_ptr = gen_rtx ( PLUS, Pmode, stack_pointer_rtx, 
			     gen_rtx ( CONST_INT, VOIDmode, stack_offset ));
    }
    return stack_offset;
}

int
multiply_operator ( op, mode )
    rtx op;
    enum machine_mode mode;
{
    return ( SImode == mode && 
	    ( MULT == GET_CODE ( op ) || UMULT == GET_CODE ( op )));
}

#include "config/dsp.c"
