/* Definitions of target machine for GNU compiler, for the Motorola dsp96000
   digital signal processing chip.
   Copyright (C) 1988 Free Software Foundation, Inc.

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  */


/* Note that some other tm- files include this one and then override
   many of the definitions that relate to assembler syntax.  */

#define DSP96000 1
#define TARGET_PREDEF "DSP96K"
#define TARGET_NAME "DSP96002"

#undef POINTER_SMALLER_THAN_INT
#undef INT_SMALLER_THAN_LONG
#undef LONG_SMALLER_THAN_LONG_LONG
#define HAVE_UNSIGNED_BRANCH_CODES
#undef HAVE_FRACTIONAL_ACCUMULATORS
#define HAVE_L_MEMORY
#define HAVE_FRAME_POINTER

#include "config/dsp.h"

/* Macro to define tables used to set the flags.
   This is a list in braces of pairs in braces,
   each pair being { "NAME", VALUE }
   where VALUE is the bits to set or minus the bits to clear.
   An empty string NAME is used to identify the default VALUE.  */

#define TARGET_DSP				( target_flags & ( 3 + 96 ))
#define TARGET_LINV_PLUS_BIV_PROMOTION		( target_flags & 1 )
#define TARGET_DO_LOOP_GENERATION		( target_flags & 2 )
#define TARGET_X_MEMORY	 			( target_flags & 4 )
#define TARGET_Y_MEMORY				( target_flags & 8 )
#define TARGET_L_MEMORY				( target_flags & 16 )
#define TARGET_CALL_OVERHEAD_REDUCTION		( target_flags & 32 )
#define TARGET_RELOAD_CLEANUP			( target_flags & 64 )
#define TARGET_STACK_CHECK			( target_flags & 4096 )
#define TARGET_P_MEM_SWITCHTABLE		( target_flags & 8192 )
#define TARGET_CONSERVE_P_MEM			( target_flags & 16384 )
#define TARGET_REP				( target_flags & 32768 )

#if !defined( MSIL_DISABLE )
#define TARGET_AUTO_INIT			( target_flags & (1<<16) &0)
#endif

#define TARGET_SWITCHES  			\
  {{ "dsp", ( 3 + 96 ) },			\
   { "no-dsp", -( 3 + 96 ) },	 		\
   { "linv-plus-biv-promotion", 1 },		\
   { "no-linv-plus-biv-promotion", -1 },	\
   { "do-loop-generation", 2 },			\
   { "no-do-loop-generation", -2 },		\
   { "x-memory", 4 },				\
   { "y-memory", 8 },				\
   { "l-memory", 16 },				\
   { "call-overhead-reduction", 32 },           \
   { "no-call-overhead-reduction", -32 },       \
   { "reload-cleanup", 64 },			\
   { "no-reload-cleanup", -64 },		\
   { "stack-check", 4096 },                     \
   { "no-stack-check", 4096 },                  \
   { "p-mem-switchtable", 8192 },		\
   { "no-p-mem-switchtable", -8192 },		\
   { "conserve-p-mem", 16384 },		\
   { "no-conserve-p-mem", -16384 },		\
   { "rep", 32768 },		\
   { "no-rep", -32768 },		\
   { "", TARGET_DEFAULT}}

#define TARGET_DEFAULT ( 3 + 16 + 96 )

#define MAX_FIXED_MODE_SIZE 32


/* Standard register usage.  */

/*
  Number of actual hardware registers.
  The hardware registers are assigned numbers for the compiler
  from 0 to just below FIRST_PSEUDO_REGISTER.
  All registers that the compiler knows about must be given numbers,
  even those that are not normally considered general registers.

  40 registers for the m96k ( 24 data + 8 address + 8 index )
*/
#define FIRST_ADDRESS_REGISTER 24
#define FIRST_PSEUDO_REGISTER 40

#define DSP_D0L_REGNUM 0
#define DSP_D0M_REGNUM 1
#define DSP_D0H_REGNUM 2
#define DSP_D1L_REGNUM 3
#define DSP_D1M_REGNUM 4
#define DSP_D1H_REGNUM 5
#define DSP_D2L_REGNUM 6
#define DSP_D2M_REGNUM 7
#define DSP_D2H_REGNUM 8
#define DSP_D3L_REGNUM 9
#define DSP_D3M_REGNUM 10
#define DSP_D3H_REGNUM 11
#define DSP_D4L_REGNUM 12
#define DSP_D4M_REGNUM 13
#define DSP_D4H_REGNUM 14
#define DSP_D5L_REGNUM 15
#define DSP_D5M_REGNUM 16
#define DSP_D5H_REGNUM 17
#define DSP_D6L_REGNUM 18
#define DSP_D6M_REGNUM 19
#define DSP_D6H_REGNUM 20
#define DSP_D7L_REGNUM 21
#define DSP_D7M_REGNUM 22
#define DSP_D7H_REGNUM 23

#define DSP_R0_REGNUM 24
#define DSP_R1_REGNUM 25
#define DSP_R2_REGNUM 26
#define DSP_R3_REGNUM 27
#define DSP_R4_REGNUM 28
#define DSP_R5_REGNUM 29
#define DSP_R6_REGNUM 30
#define DSP_R7_REGNUM 31

#define DSP_N0_REGNUM 32
#define DSP_N1_REGNUM 33
#define DSP_N2_REGNUM 34
#define DSP_N3_REGNUM 35
#define DSP_N4_REGNUM 36
#define DSP_N5_REGNUM 37
#define DSP_N6_REGNUM 38
#define DSP_N7_REGNUM 39
#define LAST_PARAMETER_REGNUM DSP_D1H_REGNUM


/* 1 for registers that have pervasive standard uses
   and are not available for the register allocator.

   for the m96k:

   0..23   = (d0.l,d0.m,d0.h)..(d7.l,d7.m,d7.h) general data registers
   24..29  = r0..r5 address registers
   30	   = r6 frame pointer
   31      = r7 stack pointer
   32..37  = n0..n5 index registers
   38	   = n6 index register associted in hardware to frame pointer
   39	   = n7 index register associted in hardware to stack pointer
*/
#define FIXED_REGISTERS   \
 {0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  1, 0, 0, 0, 0, 0, 1, 0, \
  1, 0, 0, 0, 0, 0, 1, 0}

/* Like FIXED_REGISTERS but has a 1 for each register that is clobbered
   ( in general ) by function calls. This macro therefore identifies the
   registers that are not available for general allocation of values that
   live across function calls.
   If a register has 0 in CALL_USED_REGISTERS the compiler automatically
   saves it on function entry if the register is used within that function.

   These must include the FIXED_REGISTERS and also any
   registers that can be used without being saved.
   The latter must include the registers where values are returned
   and the register where structure-value addresses are passed.
   Aside from that, you can include as many other registers as you like.  */
#define CALL_USED_REGISTERS  \
 {1, 1, 1, 1, 1, 1, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  1, 0, 0, 0, 0, 0, 1, 0, \
  1, 0, 0, 0, 0, 0, 1, 0}

/* Return number of consecutive hard regs needed starting at reg REGNO
   to hold something of mode MODE.
   This is ordinarily the length in words of a value of mode MODE
   but can be less for certain modes in special long registers.

   m96k puts floats and doubles into 96 bit registers, double integers
   into 64 bit registers and everything else into 32 bit registers */

#define HARD_REGNO_NREGS(REGNO, MODE) \
    ( ( MODE == DFmode || MODE == SFmode ) ? 3 : ( MODE == DImode ) ? 2 : 1 )

/* Register to use for pushing function arguments.  */
#define STACK_POINTER_REGNUM 30

/* Base register for access to local variables of the function.  */
#define FRAME_POINTER_REGNUM 24

/* Register in which static-chain is passed to a function.  */
/* This is currently not used but this may change for pascal or ada or ... */
#define STATIC_CHAIN_REGNUM 15

/* Register in which address to store a structure value
   is passed to a function.  */
/* pcc for the m96k returns pointer types in r0 and all others in d0,
   however there does not seem to be a way to get gcc to do the same thing.
   */
#define STRUCT_VALUE_REGNUM 31


/* Define the classes of registers for register constraints in the
   machine description.  Also define ranges of constants.

   One of the classes must always be named ALL_REGS and include all hard regs.
   If there is more than one class, another class must be named NO_REGS
   and contain no registers.

   The name GENERAL_REGS must be the name of a class (or an alias for
   another name such as ALL_REGS).  This is the class of registers
   that is allowed by "g" or "r" in a register constraint.
   Also, registers outside this class are allocated only when
   instructions express preferences for them.

   The classes must be numbered in nondecreasing order; that is,
   a larger-numbered class must never be contained completely
   in a smaller-numbered class.

   For any two classes, it is very desirable that there be another
   class that represents their union.  */
   
/* dsp96k register set looks like this:

   d0..d7 are general purpose date registers each of which can be
   used as either a set of three 32 bit registers or as a single 96
   bit register. In general, only the low 32 bits of each register
   can be used for integer type instructions. The only execeptions
   are shift instructions which use the high order 32 bits as a shift
   count.
   
   d8..d9 are 96 bit registers identical in type to d0..d7. However d8..d9
   cannot be used by most opcodes.

   r0..r7 are 32 bit address registers.

   n0..n7 are 32 bit index registers.

   For the compiler the register set will be used as follows:

	   data type			register class

	   integer 			d0..d7

	   floating point		d0..d7

	   pointer			r0..r7

   n0..n7 will be used as index registers and also for pointer,
   integer operations such a multiplication. They will have to be
   allocated based on the name of the address register being used
   i.e for r0, n0 will have to be allocated. */

enum reg_class
{
    NO_REGS,
    LO_DATA_REGS,
    HI_DATA_REGS,
    DI_REGS,
    FP_REGS,
    ADDR_REGS,
    DATA_OR_ADDR_REGS,
    INDX_REGS,
    ALL_REGS,
    LIM_REG_CLASSES
};

/* Since GENERAL_REGS is the same class as ALL_REGS,
   don't give it a different class number; just make it an alias.  */

#undef GENERAL_REGS
#define GENERAL_REGS ALL_REGS

/* Give names of register classes as strings for dump file.   */

#define REG_CLASS_NAMES {					\
	"NO_REGS", "LO_DATA_REGS", "HI_DATA_REGS",		\
	"DI_REGS", "FP_REGS", "ADDR_REGS", "DATA_OR_ADDR_REGS",	\
	"INDX_REGS", "ALL_REGS"					\
								}

/* Define which registers fit in which classes.
   This is an initializer for a vector of HARD_REG_SET
   of length N_REG_CLASSES.  */

#define REG_CLASS_CONTENTS {					\
	000000000000, 000000000000,	/* NO_REGS */		\
	000011111111, 000000000000,	/* LO_DATA_REGS .l */	\
	000044444444, 000000000000,	/* HI_DATA_REGS .h */	\
	000033333333, 000000000000,     /* DI_REGS .ml */       \
	000077777777, 000000000000,	/* FP_REGS */		\
	037700000000, 000000000000,	/* ADDR_REGS */		\
	037711111111, 000000000000,	/* DATA_OR_ADDR_REGS */	\
	000000000000, 000000000377,	/* INDX_REGS */		\
	037777777777, 000000000377 	/* ALL_REGS */		};	

/* The same information, inverted:
   Return the class number of the smallest class containing
   reg number REGNO.  This could be a conditional expression
   or could index an array.  */

#define INDEX_REG_CLASS INDX_REGS
#define BASE_REG_CLASS ADDR_REGS

/* The letters I, J, K, L and M in a register constraint string
   can be used to stand for particular ranges of immediate operands.
   This macro defines what the ranges are.
   C is the letter, and VALUE is a constant value.
   Return 1 if VALUE is in the range specified by C.
*/
#define CONST_OK_FOR_LETTER_P(VALUE, C)  			\
    ( ( 'I' == ( C ) && 0 == ( VALUE ))				\
     || ( 'J' == ( C ) && ( 1 ==  ( VALUE )))			\
     || ( 'K' == ( C ) && ( -2 <= ( VALUE ) && 2 >= ( VALUE )))	\
     || ( 'L' == ( C ) && ( 0 == ( 0xfff80000 & ( VALUE )))))

/* Similar, but for floating constants, and defining letters G and H.
   Here VALUE is the CONST_DOUBLE rtx itself.  */

#define CONST_DOUBLE_OK_FOR_LETTER_P(VALUE, C) 0


/* Offset of first parameter from the argument pointer register value.  */
#define FIRST_PARM_OFFSET(FNDECL) \
    local_first_parm_offset ( FNDECL )

/* Define how to find the value returned by a function.
   VALTYPE is the data type of the value (as a tree).
   If the precise function being called is known, FUNC is its FUNCTION_DECL;
   otherwise, FUNC is 0.  */

#define LIBCALL_VALUE(MODE)  gen_rtx (REG, MODE, 0)

#define FUNCTION_VALUE(VALTYPE, FUNC)  LIBCALL_VALUE ( TYPE_MODE (VALTYPE) )

/* Define how to find the value returned by a library function
   assuming the value has mode MODE.  */

/* 1 if N is a possible register number for a function value
   as seen by the caller.*/

#define FUNCTION_VALUE_REGNO_P(N) ( 0 == ( N ) )

/* 1 if N is a possible register number for function argument passing. */

#define FUNCTION_ARG_REGNO_P(N) ( 0 == ( N ) || 3 == ( N ) )

/* Define a data type for recording info about an argument list
   during the scan of that argument list.  This data type should
   hold all necessary information about the function itself
   and about the args processed so far, enough to enable macros
   such as FUNCTION_ARG to determine where the next arg should go.
*/
#define CUMULATIVE_ARGS int

/* Initialize a variable CUM of type CUMULATIVE_ARGS
   for a call to a function whose data type is FNTYPE.
   For a library call, FNTYPE is 0.
*/
#define INIT_CUMULATIVE_ARGS(CUM,FNTYPE) \
 ((CUM) = 0 )

/* Update the data in CUM to advance over an argument
   of mode MODE and data type TYPE.
   (TYPE is null for libcalls where that information may not be available.)  */

#define FUNCTION_ARG_ADVANCE(CUM, MODE, TYPE, NAMED) ((CUM) ++)

/* Determine where to put an argument to a function.
   Value is zero to push the argument on the stack,
   or a hard register in which to store the argument.

   MODE is the argument's machine mode.
   TYPE is the data type of the argument (as a tree).
    This is null for libcalls where that information may
    not be available.
   CUM is a variable of type CUMULATIVE_ARGS which gives info about
    the preceding args and about the function being called.
   NAMED is nonzero if this argument is a named parameter
    (otherwise it is an extra parameter matching an ellipsis).  */

#define FUNCTION_ARG(CUM, MODE, TYPE, NAMED)			\
(( 2 <= CUM || BLKmode == MODE || 0 == NAMED )			\
 ? 0 : gen_rtx ( REG, ( MODE ), 3 * CUM ))

/* Define where a function finds its arguments.
   This would be different from FUNCTION_ARG if we had register windows.  */

#define FUNCTION_INCOMING_ARG(CUM, MODE, TYPE, NAMED)\
    local_function_incoming_arg ( CUM, MODE, TYPE, NAMED )

/* For an arg passed partly in registers and partly in memory,
   this is the number of registers used.
   For args passed entirely in registers or entirely in memory, zero.  */

#define FUNCTION_ARG_PARTIAL_NREGS(CUM, MODE, TYPE, NAMED) 0

#define FIX_FRAME_POINTER_ADDRESS(ADDR,DEPTH) 

/* Addressing modes, and classification of registers for them.  */

#define HAVE_POST_INCREMENT
#define HAVE_POST_DECREMENT

#undef HAVE_PRE_DECREMENT
#undef HAVE_PRE_INCREMENT 

/* Macros to check register numbers against specific register classes.  */

/* These assume that REGNO is a hard or pseudo reg number.
   They give nonzero only if REGNO is a hard reg of the suitable class
   or a pseudo reg currently allocated to a suitable hard reg.
   Since they use reg_renumber, they are safe only once reg_renumber
   has been allocated, which happens in local-alloc.c.  */

#define REGNO_OK_FOR_BASE_P(REGNO)  				\
    ( ( REGNO < 32 && REGNO > 23 ) || 				\
    ( ( ( unsigned ) reg_renumber[REGNO] < 32 ) &&		\
    ( ( unsigned ) reg_renumber[REGNO] > 23 ) ) )

#define REGNO_OK_FOR_INDEX_P(REGNO) 				\
    ( ( REGNO < 40 && REGNO > 31 ) || 				\
    ( ( ( unsigned ) reg_renumber[REGNO] < 40 ) &&		\
    ( ( unsigned ) reg_renumber[REGNO] > 31 ) ) )

/* Now macros that check whether X is a register and also,
   strictly, whether it is in a specified class.

   These macros are specific to the the m96000, and may be used only
   in code for printing assembler insns and in conditions for
   define_optimization.  */

/* Maximum number of registers that can appear in a valid memory address.  */

#define MAX_REGS_PER_ADDRESS 1

/* Recognize any constant value that is a valid address.  */

#define CONSTANT_ADDRESS_P(X)  CONSTANT_P (X)

/* Nonzero if the constant value X is a legitimate general operand.
   It is given that X satisfies CONSTANT_P or is a CONST_DOUBLE.  */

#define LEGITIMATE_CONSTANT_P(X) (CONSTANT_P(X))

/* The macros REG_OK_FOR..._P assume that the arg is a REG rtx
   and check its validity for a certain class.
   We have two alternate definitions for each of them.
   The usual definition accepts all pseudo regs; the other rejects
   them unless they have been allocated suitable hard regs.
   The symbol REG_OK_STRICT causes the latter definition to be used.

   Most source files want to accept pseudo regs in the hope that
   they will get allocated to the class that the insn wants them to be in.
   Source files for reload pass need to be strict.
   After reload, it makes no difference, since pseudo regs have
   been eliminated by then.  */

# ifndef REG_OK_STRICT

/*
 * Nonzero if X is a hard reg that can be used as an index
 * or if it is a pseudo reg.
 */
#define REG_OK_FOR_INDEX_P(X) 						\
    ( REG_P ( X ) &&                                                    \
     (( REGNO ( X ) >= FIRST_PSEUDO_REGISTER ) ||                       \
      ( REGNO ( X ) >= DSP_N0_REGNUM &&                                 \
       REGNO ( X ) <= DSP_N7_REGNUM )))

/*
 * Nonzero if X is a hard reg that can be used as a base reg
 * or if it is a pseudo reg.
 */
#define REG_OK_FOR_BASE_P(X) 						\
    ( REG_P ( X ) &&                                                    \
     (( REGNO ( X ) >= FIRST_PSEUDO_REGISTER ) ||                       \
      ( REGNO ( X ) >= DSP_R0_REGNUM &&                                 \
       REGNO ( X ) <= DSP_R7_REGNUM )))

# else

/* Nonzero if X is a hard reg that can be used as an index.  */
# define REG_OK_FOR_INDEX_P(X) REGNO_OK_FOR_INDEX_P (REGNO (X))

/* Nonzero if X is a hard reg that can be used as a base reg.  */
# define REG_OK_FOR_BASE_P(X) REGNO_OK_FOR_BASE_P (REGNO (X))

# endif

/* GO_IF_LEGITIMATE_ADDRESS recognizes an RTL expression
   that is a valid memory address for an instruction.
   The MODE argument is the machine mode for the MEM expression
   that wants to use this address.

   On the m96000, the actual legitimate addresses must be REG+REG or
   REG+SMALLINT. But we can treat a SYMBOL_REF as legitimate if it is
   part of this function's constant-pool, because such addresses can
   actually be output as REG+SMALLINT.  */

# ifdef REG_OK_STRICT
# define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR) 		\
    if ( go_if_legitimate_address ( MODE, X, 1 ) ) goto ADDR;
# else
# define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR) 		\
    if ( go_if_legitimate_address ( MODE, X, 0 ) ) goto ADDR;
# endif

/* Try machine-dependent ways of modifying an illegitimate address
   to be legitimate.  If we find one, return the new, valid address.
   This macro is used in only one place: `memory_address' in explow.c.

   OLDX is the address as it was before break_out_memory_refs was called.
   In some cases it is useful to look at this to decide what needs to be done.

   MODE and WIN are passed so that this macro can use
   GO_IF_LEGITIMATE_ADDRESS.

   It is always safe for this macro to do nothing.  It exists to recognize
   opportunities to optimize the output.  */

#define LEGITIMIZE_ADDRESS(X,OLDX,MODE,WIN)

/* Go to LABEL if ADDR (a legitimate address expression)
   has an effect that depends on the machine mode it is used for.
   The m96000 is word addressed with 64 bits at each address so
   this is never true.  */

# define GO_IF_MODE_DEPENDENT_ADDRESS(ADDR,LABEL)


/* Specify the machine mode that this machine uses
   for the index in the tablejump instruction.  */
#define CASE_VECTOR_MODE SImode

/* Define this if a raw index is all that is needed for a
   `tablejump' insn.  */
#undef CASE_TAKES_INDEX_RAW

/* Define this if the tablejump instruction expects the table
   to contain offsets from the address of the table.
   Do not define this if the table should contain absolute addresses.  */
#undef CASE_VECTOR_PC_RELATIVE

/* Specify the tree operation to be used to convert reals to integers.  */
#define IMPLICIT_FIX_EXPR FIX_ROUND_EXPR

/* This is the kind of divide that is easiest to do in the general case.  */
#define EASY_DIV_EXPR TRUNC_DIV_EXPR

/* Define this as 1 if `char' should by default be signed; else as 0.  */
#define DEFAULT_SIGNED_CHAR 1

/* Max number of bytes we can move from memory to memory
   in one reasonably fast instruction.  */
#define MOVE_MAX 1

/* We can't use move by pieces due to l space problems. */
#define MOVE_RATIO 0
#define MOVE_RATIO_96 5

/* when is it better not to use bitset/bitclr in setting a bitfield ? 
   i.e. how many bits ? */
#define BIT_THRESHOLD 4

/* Value is 1 if truncating an integer of INPREC bits to OUTPREC bits
   is done just by pretending it is already truncated.  */
#define TRULY_NOOP_TRUNCATION(OUTPREC, INPREC) \
    ( 32 == OUTPREC && 64 == INPREC )

/* Specify the machine mode that pointers have.
   After generation of rtl, the compiler makes no further distinction
   between pointers and any other objects of this machine mode.  */
#define Pmode SImode

/* A function address in a call instruction
   is a byte address (for indexing purposes)
   so give the MEM rtx a byte's mode.  */
#define FUNCTION_MODE SImode

/* Compute the cost of computing a constant rtl expression RTX
   whose rtx-code is CODE.  The body of this macro is a portion
   of a switch statement.  If the code is computed here,
   return it with a return statement.  Otherwise, break from the switch.  */

#define CONST_COSTS(RTX,CODE) \
  case CONST_INT:						\
  case CONST:							\
  case LABEL_REF:						\
  case SYMBOL_REF:						\
  case CONST_DOUBLE:						\
    return 2;

/* How to refer to registers in assembler output.
   This sequence is indexed by compiler's hard-register-number (see above).  */

#define REGISTER_NAMES 				\
{ "d0", "d0", "d0", "d1", "d1", "d1",		\
  "d2", "d2", "d2", "d3", "d3", "d3",		\
  "d4", "d4", "d4", "d5", "d5", "d5",		\
  "d6", "d6", "d6", "d7", "d7", "d7",		\
  "r0", "r1", "r2", "r3", "r4", "r5",		\
  "r6", "r7",					\
  "n0", "n1", "n2", "n3", "n4", "n5",		\
  "n6", "n7" }

#define ASM_OUTPUT_CASE_LABEL(STREAM, PREFIX, NUM, TABLE ) 	\
    (( TARGET_P_MEM_SWITCHTABLE ) ? 				\
     ( fprintf ( STREAM, "%s\n", TEXT_SECTION_ASM_OP )) :	\
     ( fprintf (STREAM, "%s\n", DATA_SECTION_ASM_OP )),		\
     ASM_OUTPUT_INTERNAL_LABEL( STREAM, PREFIX, NUM ))

#define ASM_OUTPUT_CASE_END(STREAM, NUM, TABLE ) 		\
  fprintf (STREAM, "%s\n", TEXT_SECTION_ASM_OP )		\

/* This is how to output an element of a case-vector that is relative.  */

#define ASM_OUTPUT_ADDR_DIFF_ELT(FILE, VALUE, REL)  \
  fprintf (FILE, "\tdc L%d-L%d\n", VALUE, REL)

/* This is how to output an element of a case-vector that is absolute.  */

#define ASM_OUTPUT_ADDR_VEC_ELT(FILE, VALUE)  \
  fprintf (FILE, "\tdc L%d\n", VALUE)

extern double ldexp ( );

/* global memory space indicator - either 'x' or 'y'. */
extern char mem_space;
extern char memory_model;

/* this enum is used as the mdep portion of cc_... structs. branches, conds and
   jumps on the 96k need to know whether the ccs were set by a floating point
   instruction or not. */

typedef enum
{
    INTEGER_CCS, /* regular cmp or integer op. */
    FLOAT_CCS    /* fcmp or floating op. */
} compare_type;

#define CC_STATUS_MDEP_INIT INTEGER_CCS
#define CC_STATUS_MDEP compare_type

/* these force a min number of operands allowed. */
#define MIN_RECOG_OPERANDS 16
#define MIN_DUP_OPERANDS 16

/* we undef abort so that we can print a proper warning message and not dump
   core. (code in rtl.c) */
#undef abort

/* allow certain pragmas. */
int dsp_process_pragma_line ( );

#define TARGET_PROCESS_PRAGMA_LINE( f ) \
dsp_process_pragma_line ( f )
