/* Definitions of target machine for GNU compiler, for the Motorola DSP56300
   digital signal processing chip.
   Copyright (C) 1988 Free Software Foundation, Inc.

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  */


/* Note that some other tm- files include this one and then override
   many of the definitions that relate to assembler syntax.  */

#define MCU56300 1
#define TARGET_PREDEF "MCU563C"
#define TARGET_NAME "MCU5630X"

#undef FRACTIONAL_SUPPORT
#undef POINTER_SMALLER_THAN_INT
#define POINTER_SMALLER_THAN_LONG
#define POINTER_LARGER_THAN_INT
#define INT_SMALLER_THAN_LONG
#undef LONG_SMALLER_THAN_LONG_LONG
#undef HAVE_UNSIGNED_BRANCH_CODES
#define HAVE_FRACTIONAL_ACCUMULATORS
#undef HAVE_L_MEMORY
#undef HAVE_FRAME_POINTER
#undef HAVE_INTEGER_MULT
#define HAVE_IFCCS

#include "config/dsp.h"

/* Macro to define tables used to set the flags.
   This is a list in braces of pairs in braces,
   each pair being { "NAME", VALUE }
   where VALUE is the bits to set or minus the bits to clear.
   An empty string NAME is used to identify the default VALUE.  */

#define DSP_ALL_OPTS				483
#define TARGET_DSP				( target_flags & DSP_ALL_OPTS )
#define TARGET_LINV_PLUS_BIV_PROMOTION		( target_flags & 1 )
#define TARGET_DO_LOOP_GENERATION		( target_flags & (1<<1))
#define TARGET_CALL_OVERHEAD_REDUCTION		( target_flags & (1<<5))
#define TARGET_REP				( target_flags & (1<<6))
#define TARGET_RELOAD_CLEANUP			( target_flags & (1<<7))
#define TARGET_NORMALIZATION_REDUCTION          ( target_flags & (1<<8))
#define TARGET_STACK_CHECK			( target_flags & (1<<9))
#define TARGET_P_MEM_SWITCHTABLE		( target_flags & (1<<10))
#define TARGET_CONSERVE_P_MEM			( target_flags & (1<<11))
#define TARGET_SWITCHES		\
  {{ "dsp", DSP_ALL_OPTS },	       	        \
   { "no-dsp", -DSP_ALL_OPTS },	 	        \
   { "linv-plus-biv-promotion", 1 },		\
   { "no-linv-plus-biv-promotion", -1 },	\
   { "do-loop-generation", (1<<1) },		\
   { "no-do-loop-generation", -(1<<1) },       	\
   { "call-overhead-reduction",(1<<5) },        \
   { "no-call-overhead-reduction", -(1<<5) },   \
   { "rep", (1<<6) },				\
   { "no-rep", -(1<<6) },      			\
   { "reload-cleanup", (1<<7) },		\
   { "no-reload-cleanup", -(1<<7) },		\
   { "normalization-reduction", (1<<8) },       \
   { "no-normalization-reduction", -(1<<8) },   \
   { "stack-check", (1<<9) },                   \
   { "no-stack-check", -(1<<9) },               \
   { "p-mem-switchtable", (1<<10) },		\
   { "no-p-mem-switchtable", -(1<<10) },	\
   { "conserve-p-mem", (1<<11) },		\
   { "no-conserve-p-mem", -(1<<11) },		\
   { "c-for-dsp", (1<<15) },			\
   { "no-c-for-dsp", -(1<<15) },		\
   { "", TARGET_DEFAULT}}

#define TARGET_DEFAULT ( DSP_ALL_OPTS + (1<<11))

/* Standard register usage.  */


/*
  Number of actual hardware registers.
  The hardware registers are assigned numbers for the compiler
  from 0 to just below FIRST_PSEUDO_REGISTER.
  All registers that the compiler knows about must be given numbers,
  even those that are not normally considered general registers.

  for the DSP56300, we have x1,x0,y1,y0,a,b, r0-r7 and n0-n7
*/
#define FIRST_PSEUDO_REGISTER 15

/* 1 for registers that have pervasive standard uses
   and are not available for the register allocator.

   for the DSP56300:

   0..5    = (a, b, x0, x1, y0, y1)
   6..11   = (r0, r1, r2, r3, r4, r5)
   12      = (r6) stack pointer.
   13      = (r7)
   14..21  = (n0 ... n7)
   22      = (fake frame pointer)
*/
#define DSP_A_REGNUM 0
#define DSP_B_REGNUM 1
#define DSP_X0_REGNUM 2
#define DSP_X1_REGNUM 3
#define DSP_Y0_REGNUM 4
#define DSP_Y1_REGNUM 5
#define DSP_R0_REGNUM 6
#define DSP_R1_REGNUM 7
#define DSP_R2_REGNUM 8
#define DSP_R3_REGNUM 9
#define DSP_R4_REGNUM 10
#define DSP_R5_REGNUM 11
#define DSP_R6_REGNUM 12
#define DSP_R7_REGNUM 13
#define DSP_FAKE_REGNUM 14
#define FIRST_ADDRESS_REGISTER DSP_R0_REGNUM
#define LAST_ADDRESS_REGISTER DSP_R7_REGNUM

#define CALLER_SAVES \
    DSP_X0_REGNUM, DSP_X1_REGNUM,		\
    DSP_R0_REGNUM, DSP_R4_REGNUM,		\
    DSP_R1_REGNUM, DSP_R5_REGNUM,		\
    DSP_A_REGNUM, DSP_B_REGNUM,

#define CALLEE_SAVES \
    DSP_Y0_REGNUM, DSP_Y1_REGNUM,		\
    DSP_R2_REGNUM, DSP_R7_REGNUM, DSP_R3_REGNUM,\

#define CANT_USE \
    DSP_R6_REGNUM, DSP_FAKE_REGNUM

#define LEAF_REG_ALLOC_ORDER { CALLER_SAVES CALLEE_SAVES CANT_USE }

#define REG_ALLOC_ORDER { CALLEE_SAVES CALLER_SAVES CANT_USE }

/* 1 for registers that have pervasive standard uses
   and are not available for the register allocator. */
#define FIXED_REGISTERS   \
 {0, 0, 0, 0, 0, 0,       \
  0, 0, 0, 0, 0, 0, 1, 0, \
  1} 

/* Like FIXED_REGISTERS but has a 1 for each register that is clobbered
   ( in general ) by function calls. This macro therefore identifies the
   registers that are not available for general allocation of values that
   live across function calls.
   If a register has 0 in CALL_USED_REGISTERS the compiler automatically
   saves it on function entry if the register is used within that function.

   These must include the FIXED_REGISTERS and also any
   registers that can be used without being saved.
   The latter must include the registers where values are returned
   and the register where structure-value addresses are passed.
   Aside from that, you can include as many other registers as you like.  */
/* with the DSP56300 arguments will be returned in a. */
#define CALL_USED_REGISTERS  \
 {1, 1, 1, 1, 0, 0,          \
  1, 1, 0, 0, 1, 1, 1, 0,    \
  1} 

/* Return number of consecutive hard regs needed starting at reg REGNO
   to hold something of mode MODE.
   This is ordinarily the length in words of a value of mode MODE
   but can be less for certain modes in special long registers.

   DSP56300 puts floats/doubles into 48 bit regs, longs into 48 bit regs. */

#define HARD_REGNO_NREGS( REGNO, MODE ) \
    ((( DFmode == (MODE) || SFmode == (MODE) || DImode == (MODE) ) && \
      ( (REGNO) != DSP_A_REGNUM ) && ( (REGNO) != DSP_B_REGNUM )) ? 2 : \
     (( (REGNO) > DSP_B_REGNUM && PDImode == (MODE) && \
       (REGNO) < FIRST_ADDRESS_REGISTER ) ? 2 : 1 ))
  
/* Register to use for pushing function arguments.  */
#define STACK_POINTER_REGNUM DSP_R6_REGNUM

/* Base register for access to local variables of the function.  */
#define FRAME_POINTER_REGNUM DSP_FAKE_REGNUM

/* Base register for access to all statically allocated data when
   -mpid is used. */
#define DATA_POINTER_REGNUM DSP_R1_REGNUM

/* Register in which static-chain is passed to a function.  */
/* This is currently not used but this may change for pascal or ada or ... */
#define STATIC_CHAIN_REGNUM DSP_A_REGNUM

/* Register in which address to store a structure value
   is passed to a function.  */
#define STRUCT_VALUE_REGNUM DSP_R7_REGNUM


/* Define the classes of registers for register constraints in the
   machine description.  Also define ranges of constants.

   One of the classes must always be named ALL_REGS and include all hard regs.
   If there is more than one class, another class must be named NO_REGS
   and contain no registers.

   The name GENERAL_REGS must be the name of a class (or an alias for
   another name such as ALL_REGS).  This is the class of registers
   that is allowed by "g" or "r" in a register constraint.
   Also, registers outside this class are allocated only when
   instructions express preferences for them.

   The classes must be numbered in nondecreasing order; that is,
   a larger-numbered class must never be contained completely
   in a smaller-numbered class.

   For any two classes, it is very desirable that there be another
   class that represents their union.  */
   
enum reg_class
{
    NO_REGS,
    ADDR_REGS,    /* SImode registers: r0-r7 */
    MPY_REGS,	  /* source regs valid for mpy/mac multiplier x0,y0. */
    DST_REGS,     /* All-mode destination registers: a, b */
    SRC_REGS,     /* All-mode source registers: x1, x0, y1, y0 */
    ALL_REGS,
    LIM_REG_CLASSES
};

/* Give names of register classes as strings for dump file.   */

#define REG_CLASS_NAMES { \
	"NO_REGS", "ADDR_REGS", "MPY_REGS", \
        "DST_REGS", "SRC_REGS", "ALL_REGS" }

/* Define which registers fit in which classes.
   This is an initializer for a vector of HARD_REG_SET
   of length N_REG_CLASSES.  */

#define REG_CLASS_CONTENTS {			\
	000000000000, /* NO_REGS */		\
	000000037700, /* ADDR_REGS */		\
	000000000024, /* MPY_REGS */		\
	000000000003, /* DST_REGS */		\
	000000000074, /* SRC_REGS */		\
	000000077777  /* ALL_REGS */	}

/* Since GENERAL_REGS is the same class as NO_REGS,
   don't give it a different class number; just make it an alias.  */
#define GENERAL_REGS ALL_REGS

/* The same information, inverted:
   Return the class number of the smallest class containing
   reg number REGNO.  This could be a conditional expression
   or could index an array.  */

#define IS_SRC_OR_MPY_P(REGNO)    \
    ( DSP_X0_REGNUM == (REGNO) || \
      DSP_X1_REGNUM == (REGNO) || \
      DSP_Y0_REGNUM == (REGNO) || \
      DSP_Y1_REGNUM == (REGNO))

/* The class value for index registers, and the one for base regs.  */
#define INDEX_REG_CLASS ADDR_REGS
#define BASE_REG_CLASS ADDR_REGS

/* The letters I, J, K, L and M in a register constraint string
   can be used to stand for particular ranges of immediate operands.
   This macro defines what the ranges are.
   C is the letter, and VALUE is a constant value.
   Return 1 if VALUE is in the range specified by C.
*/
#define CONST_OK_FOR_LETTER_P(VALUE, C) 			\
    (( 'I' == ( C ) && ((unsigned) VALUE ) < 0x00800000 ) ||	\
     ( 'J' == ( C ) && ((unsigned) VALUE ) < 0x00001000 ) ||	\
     ( 'i' == ( C )))

/* Similar, but for floating constants, and defining letters G and H.
   Here VALUE is the CONST_DOUBLE rtx itself.  */

#define CONST_DOUBLE_OK_FOR_LETTER_P(VALUE, C) 1


/* Offset of first parameter from the argument pointer register value.  */
#define FIRST_PARM_OFFSET(FNDECL) -1

/* Define how to find the value returned by a function.
   VALTYPE is the data type of the value (as a tree).
   If the precise function being called is known, FUNC is its FUNCTION_DECL;
   otherwise, FUNC is 0.  */
#define LIBCALL_VALUE(MODE)  (gen_rtx ( REG, MODE, DSP_A_REGNUM ))

#define FUNCTION_VALUE(VALTYPE, FUNC)  LIBCALL_VALUE ( TYPE_MODE (VALTYPE) )

/* Define how to find the value returned by a library function
   assuming the value has mode MODE.  */

/* 1 if N is a possible register number for a function value
   as seen by the caller.*/

#define FUNCTION_VALUE_REGNO_P(N) ( DSP_A_REGNUM == ( N ))

/* 1 if N is a possible register number for function argument passing. */

#define FUNCTION_ARG_REGNO_P(N) ( 0 == (N) || 1 == (N))

/* Define a data type for recording info about an argument list
   during the scan of that argument list.  This data type should
   hold all necessary information about the function itself
   and about the args processed so far, enough to enable macros
   such as FUNCTION_ARG to determine where the next arg should go.
*/
#define CUMULATIVE_ARGS int

/* Initialize a variable CUM of type CUMULATIVE_ARGS
   for a call to a function whose data type is FNTYPE.
   For a library call, FNTYPE is 0.
*/
#define INIT_CUMULATIVE_ARGS(CUM,FNTYPE) ((CUM) = 0)

/* Update the data in CUM to advance over an argument
   of mode MODE and data type TYPE.
   (TYPE is null for libcalls where that information may not be available.)  */

#define FUNCTION_ARG_ADVANCE(CUM, MODE, TYPE, NAMED) ((CUM) ++)

/* Determine where to put an argument to a function.
   Value is zero to push the argument on the stack,
   or a hard register in which to store the argument.

   MODE is the argument's machine mode.
   TYPE is the data type of the argument (as a tree).
    This is null for libcalls where that information may
    not be available.
   CUM is a variable of type CUMULATIVE_ARGS which gives info about
    the preceding args and about the function being called.
   NAMED is nonzero if this argument is a named parameter
    (otherwise it is an extra parameter matching an ellipsis).  */

#define FUNCTION_ARG(CUM, MODE, TYPE, NAMED) \
(( 2 <= (CUM) || BLKmode == (MODE) || 0 == (NAMED)) ? \
 0 : gen_rtx ( REG, (MODE), (CUM)))

/* Define where a function finds its arguments.
   This would be different from FUNCTION_ARG if we had register windows.  */

#define FUNCTION_INCOMING_ARG(CUM, MODE, TYPE, NAMED)	\
    local_function_incoming_arg ( CUM, MODE, TYPE, NAMED )

/* For an arg passed partly in registers and partly in memory,
   this is the number of registers used.
   For args passed entirely in registers or entirely in memory, zero.  */

#define FUNCTION_ARG_PADDING(A,B) none
#define FUNCTION_ARG_PARTIAL_NREGS(CUM, MODE, TYPE, NAMED) 0
#undef FUNCTION_ARG_PARTIAL_NREGS

/* If the memory address ADDR is relative to the frame pointer,
   correct it to be relative to the stack pointer instead.
   This is for when we don't use a frame pointer.
   ADDR should be a variable name.  */

#define FIX_FRAME_POINTER_ADDRESS(ADDR,DEPTH) \
local_fix_frame_pointer_address ( &(ADDR),(DEPTH))

/* Addressing modes, and classification of registers for them.  */

#define HAVE_POST_INCREMENT
#define HAVE_POST_DECREMENT

#undef HAVE_PRE_DECREMENT
#undef HAVE_PRE_INCREMENT 

/* Macros to check register numbers against specific register classes.  */

/* These assume that REGNO is a hard or pseudo reg number.
   They give nonzero only if REGNO is a hard reg of the suitable class
   or a pseudo reg currently allocated to a suitable hard reg.
   Since they use reg_renumber, they are safe only once reg_renumber
   has been allocated, which happens in local-alloc.c.  */

#define REGNO_OK_FOR_BASE_P(REGNO)  				\
	(( DSP_R0_REGNUM <= REGNO && DSP_R7_REGNUM >= REGNO ) ||\
	 ( DSP_R0_REGNUM <= reg_renumber[REGNO] &&              \
	   DSP_R7_REGNUM >= reg_renumber[REGNO] ) ||            \
	   DSP_FAKE_REGNUM == REGNO || DSP_FAKE_REGNUM == reg_renumber[REGNO] )

#define REGNO_OK_FOR_INDEX_P(REGNO) REGNO_OK_FOR_BASE_P(REGNO)

/* Now macros that check whether X is a register and also,
   strictly, whether it is in a specified class. */

/* Maximum number of registers that can appear in a valid memory address.  */

#define MAX_REGS_PER_ADDRESS 1

/* Recognize any constant value that is a valid address.  */

#define CONSTANT_ADDRESS_P(X)  CONSTANT_P(X)

/* Nonzero if the constant value X is a legitimate general operand.
   It is given that X satisfies CONSTANT_P or is a CONST_DOUBLE.  */

#define LEGITIMATE_CONSTANT_P(X) (CONSTANT_P(X)||CONST_DOUBLE==GET_CODE(X))

/* The macros REG_OK_FOR..._P assume that the arg is a REG rtx
   and check its validity for a certain class.
   We have two alternate definitions for each of them.
   The usual definition accepts all pseudo regs; the other rejects
   them unless they have been allocated suitable hard regs.
   The symbol REG_OK_STRICT causes the latter definition to be used.

   Most source files want to accept pseudo regs in the hope that
   they will get allocated to the class that the insn wants them to be in.
   Source files for reload pass need to be strict.
   After reload, it makes no difference, since pseudo regs have
   been eliminated by then.  */

#ifndef REG_OK_STRICT

/*
 * Nonzero if X is a hard reg that can be used as an index
 * or if it is a pseudo reg.
 */
#define REG_OK_FOR_INDEX_P(X) 						\
    ( REG_P ( X ) &&                                                    \
     (( REGNO ( X ) >= FIRST_PSEUDO_REGISTER ) ||                       \
      ( REGNO ( X ) >= DSP_R0_REGNUM &&                                 \
        REGNO ( X ) <= DSP_R7_REGNUM ) ||                               \
      ( REGNO ( X ) == DSP_FAKE_REGNUM )))
/*
 * Nonzero if X is a hard reg that can be used as a base reg
 * or if it is a pseudo reg.
 */
#define REG_OK_FOR_BASE_P(X) REG_OK_FOR_INDEX_P(X)

#else

/* Nonzero if X is a hard reg that can be used as an index.  */
#define REG_OK_FOR_INDEX_P(X) REGNO_OK_FOR_INDEX_P (REGNO (X))

/* Nonzero if X is a hard reg that can be used as a base reg.  */
#define REG_OK_FOR_BASE_P(X) REGNO_OK_FOR_BASE_P (REGNO (X))

#endif

/* GO_IF_LEGITIMATE_ADDRESS recognizes an RTL expression
   that is a valid memory address for an instruction.
   The MODE argument is the machine mode for the MEM expression
   that wants to use this address.
*/

#ifdef REG_OK_STRICT
#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR) 		\
    if ( go_if_legitimate_address ( MODE, X, 1 ) ) goto ADDR;
#else
#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR) 		\
    if ( go_if_legitimate_address ( MODE, X, 0 ) ) goto ADDR;
#endif

/* Try machine-dependent ways of modifying an illegitimate address
   to be legitimate.  If we find one, return the new, valid address.
   This macro is used in only one place: `memory_address' in explow.c.

   OLDX is the address as it was before break_out_memory_refs was called.
   In some cases it is useful to look at this to decide what needs to be done.

   MODE and WIN are passed so that this macro can use
   GO_IF_LEGITIMATE_ADDRESS.

   It is always safe for this macro to do nothing.  It exists to recognize
   opportunities to optimize the output.  */

#define LEGITIMIZE_ADDRESS(X,OLDX,MODE,WIN)

/* Go to LABEL if ADDR (a legitimate address expression)
   has an effect that depends on the machine mode it is used for. */

#define GO_IF_MODE_DEPENDENT_ADDRESS(ADDR,LABEL)


/* Specify the machine mode that this machine uses
   for the index in the tablejump instruction.  */
#define CASE_VECTOR_MODE SImode

/* Define this if a raw index is all that is needed for a
   `tablejump' insn.  */
#undef CASE_TAKES_INDEX_RAW

/* Define this if the tablejump instruction expects the table
   to contain offsets from the address of the table.
   Do not define this if the table should contain absolute addresses.  */
#undef CASE_VECTOR_PC_RELATIVE

/* Specify the tree operation to be used to convert reals to integers.  */
#define IMPLICIT_FIX_EXPR FIX_ROUND_EXPR

/* This is the kind of divide that is easiest to do in the general case.  */
#define EASY_DIV_EXPR TRUNC_DIV_EXPR

/* Define this as 1 if `char' should by default be signed; else as 0.  */
#define DEFAULT_SIGNED_CHAR 1

/* Max number of bytes we can move from memory to memory
   in one reasonably fast instruction.  */
#define MOVE_MAX 1

/* We can't use move by pieces due to l space problems. */
#define MOVE_RATIO 0
#define MOVE_RATIO_DSP56300 5

/* when is it better not to use bitset/bitclr in setting a bitfield ? 
   i.e. how many bits ? */
#define BIT_THRESHOLD 4

/* Value is 1 if truncating an integer of INPREC bits to OUTPREC bits
   is done just by pretending it is already truncated.  */
#define TRULY_NOOP_TRUNCATION(OUTPREC, INPREC) \
(((INPREC) == (OUTPREC)) || \
 ( LONG_TYPE_SIZE != (INPREC) && \
  POINTER_SIZE != (INPREC) && POINTER_SIZE != (OUTPREC)))

/* Specify the machine mode that pointers have.
   After generation of rtl, the compiler makes no further distinction
   between pointers and any other objects of this machine mode.  */
#define Pmode PDImode

/* A function address in a call instruction
   is a byte address (for indexing purposes)
   so give the MEM rtx a byte's mode.  */
#define FUNCTION_MODE SImode

/* Compute the cost of computing a constant rtl expression RTX
   whose rtx-code is CODE.  The body of this macro is a portion
   of a switch statement.  If the code is computed here,
   return it with a return statement.  Otherwise, break from the switch.  */

#define CONST_COSTS(RTX,CODE) \
  case LABEL_REF:						\
  case SYMBOL_REF:						\
  case CONST_DOUBLE:						\
    return 2; 							\
  default:							\
    return 0;


/* How to refer to registers in assembler output.
   This sequence is indexed by compiler's hard-register-number (see above).  */

#define REGISTER_NAMES 					\
{ "a", "b", "x0", "x1", "y0", "y1",			\
  "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7",       \
  "fake" } 

/* note that -mp-mem-switchtable isn't supported with -mpic. */
#define ASM_OUTPUT_CASE_LABEL(STREAM, PREFIX, NUM, TABLE )	\
     ((( TARGET_P_MEM_SWITCHTABLE ) ? 				\
      ( fprintf ( STREAM, "%s\n", TEXT_SECTION_ASM_OP )) :	\
      ( fprintf (STREAM, "%s\n", DATA_SECTION_ASM_OP ))), 	\
     ( ASM_OUTPUT_INTERNAL_LABEL( STREAM, PREFIX, NUM )))

#define ASM_OUTPUT_CASE_END(STREAM, NUM, TABLE ) 		\
  fprintf (STREAM, "%s\n", TEXT_SECTION_ASM_OP )

/* This is how to output an element of a case-vector that is relative.  */

#define ASM_OUTPUT_ADDR_DIFF_ELT(FILE, VALUE, REL)  \
  fprintf (FILE, "\tdcl\t1+L%d-L%dc\n", VALUE, REL)

/* This is how to output an element of a case-vector that is absolute.  */

#define ASM_OUTPUT_ADDR_VEC_ELT(FILE, VALUE)  \
  fprintf (FILE, "\tdcl\tL%d\n", VALUE)

