/* subroutines and data common to all motorola dsp gcc retargets. */
/* Copyright (C) 1991,1992,1993,1994,1995 Motorola, Inc. */

/* run time and load time counter setups. */

char *x_load_default = NULL,
*x_run_default = "x:";
char *x_load = NULL,
*x_run = "x:";

#if defined( HAVE_L_MEMORY )
char *y_load_default = NULL,
*y_run_default = "y:";
char *y_load = NULL,
*y_run = "y:";
char *l_load_default = NULL,
*l_run_default = "l:";
char *l_load = NULL,
*l_run = "l:";

#endif
char *p_load_default = NULL,
*p_run_default = "p:";
char *p_load = NULL,
*p_run = "p:";

#if defined( TARGET_PID )
char *rti_load_default = NULL,
*rti_run_default = "@(7):";
char *rti_load = NULL,
*rti_run = "@(7):";

#endif

#if defined( _MSDOS )
#include "bblock.h"
#include "cpt.h"
#else
#include "basic-block.h"
#include "c-parse.tab.h"
#endif
#include "c-parse.h"

void warning (char *,...);
rtx get_next_nonnote_insn ();

/* reload pass cleanup.

   The gcc register allocation scheme, as it exists in 1.37.1 leaves much
   to be desired. The reload section in particular, creates excessive and
   unnecessary loads and stores, because it does not take (the lack of) 
   global variable interference into account. 

   Until the register allocation phase ( local-alloc.c, global-alloc.c, 
   reload.c, and reload1.c ) is replaced, we will bolt on this post allocation
   "reload cleaner". It will scrub each basic block as well as can be done 
   without considering global data flow information.

   The Algorithm is as follows:

   foreach basic block b in the function,

   compress_lives ( b )
   kill_loads ( b )
   kill_stores ( b )

   end

   ; compress_lives - if a reload can be averted by moving a register
   ; transfer, do so and let kill_loads/stores eliminate the reload.

   ; kill_loads - attempt to eliminate or alter gratuitous reload created 
   ; load instructions. try to eliminate loads by getting the value from 
   ; a source cheaper than a stack slot.

   ; kill_stores - attempt to eliminate gratuitous reload created store 
   ; instructions.

 */

rtx clean_subregs ();
static int safe_strcmp ();
static void compliv_mark_touched ();
static void compress_lives ();
static int get_sf_offset ();
static void global_kill_stores ();
static void kill_loads ();
static void kill_stores ();
static void kill_prior_contents ();
static void kloads_mark_touched ();
static void kstores_mark_touched ();
static int modes_compatable_p ();
static int reg_possibly_live ();
static int reg_possibly_live_between_p ();
static int rtx_same_p ();
static int trailing_affected_slots ();

/* this is used as a conservative guesstimate for the largest fp/sp offset. */
static int slot_bias;

#if defined( HAVE_L_MEMORY ) && defined( TARGET_PACK_L_MEMORY )
/* this is used as a base number to index into x memory slots. */
static int x_slot_base;

#endif

/* this flag is used to indicate that the changes haven't settled yet. */
static int something_changed_p;

/**************************************************************************/
/* this is the main routine for cleaning up after the reload pass is run. */
/* is is called by toplev.c.                                              */
/**************************************************************************/

void
cleanup_reloads (insns)
rtx insns;
{
    int block_num = n_basic_blocks;

#if defined( FRAME_GROWS_DOWNWARD )
    slot_bias = get_frame_size () * 2 - STARTING_FRAME_OFFSET;
#else
    slot_bias = get_frame_size () * 2 + STARTING_FRAME_OFFSET;
#endif

#if defined( HAVE_L_MEMORY ) && defined( TARGET_PACK_L_MEMORY )
    /* define the X memory portion as the "top half". */
    x_slot_base = slot_bias + 1;
#endif

    while (block_num--)
    {
	int max_iters = 5;

	do
	{
	    something_changed_p = FALSE;

	    compress_lives (block_num);
	    kill_loads (block_num);
	    kill_stores (block_num);
	}
	while (TRUE == something_changed_p && --max_iters);
    }
    global_kill_stores (insns);
}

/* contains the last value loaded into a register. used by kill_loads. */
static rtx reg_value[FIRST_PSEUDO_REGISTER];

/* the mode associated with the value in a register. used by kill_loads. */
static enum machine_mode reg_value_mode[FIRST_PSEUDO_REGISTER];

/* contains the last value loaded into a stack slot. the array is 
   allocated by either kill_loads or compress_lives. compress_lives uses
   slot_value as a mapping from slot offsets to the last insn to set that
   slot. */
static rtx *slot_value;

/* the mode associated with the value in a stack slot. the memory for
   the array is allocated by kill_loads or compress_lives. */
static enum machine_mode *slot_value_mode;

/* a marker for subsequent memory (or register) locations affected
   by a wider than single unit move. */
rtx look_to_previous_elem;

static void
kill_loads (block_num)
int block_num;
{
#if defined( HAVE_L_MEMORY ) && defined( TARGET_PACK_L_MEMORY )
    /* define the X memory portion as the "top half". */
    int slot_count = (slot_bias + 1) * 2;

#else
    /* the maximum stack slot offset in this function. */
    int slot_count = slot_bias + 1;

#endif

    /* this is the generic last location used, which may be aliased but
       not volatile. it allows us to eliminate senseless loads from one
       non-reload generated address. */
    rtx non_reload_address = NULL;

    /* this is the value associated with that address. */
    rtx nra_value;

    /* this is the mode associated with that value. */
    enum machine_mode nra_mode;

    /* allocate memory for the grouping marker. */
    look_to_previous_elem = (rtx) alloca (sizeof (struct rtx_def));

    /* allocate memory for the slot_value and mode arrays. */
    slot_value = (rtx *) alloca (sizeof (rtx) * slot_count);
    slot_value_mode = (enum machine_mode *)
	alloca (sizeof (enum machine_mode) * slot_count);

    /*******************************/
    /* initialize data structures. */
    /*******************************/
    {
	/* in_defs are the unknown values resident in storage locations upon
	   entry to the basic block. */

	rtx in_defs = (rtx)
	alloca ((slot_count + FIRST_PSEUDO_REGISTER) *
		sizeof (struct rtx_def));

	int i = slot_count;

	while (i--)
	{
	    /* each slot location has a unique unknown value upon block 
	       entry. */
	    PUT_CODE (in_defs, UNKNOWN);
	    slot_value[i] = in_defs++;
	    slot_value_mode[i] = VOIDmode;
	}

	i = FIRST_PSEUDO_REGISTER;

	while (i--)
	{
	    PUT_CODE (in_defs, UNKNOWN);
	    reg_value[i] = in_defs++;
	    reg_value_mode[i] = VOIDmode;
	}
    }

    /* scan through the basic block once, in execution order. */
    {
	rtx insn = basic_block_head[block_num];
	rtx tail = basic_block_end[block_num];

	while (insn != tail)
	{
	    rtx body;

	    if (CALL_INSN == GET_CODE (insn))
	    {
		/* for a call, make sure that all of the call clobbered
		   registers are marked as undefined. */

		int i = FIRST_PSEUDO_REGISTER;

		while (i--)
		{
		    if ((TEST_HARD_REG_BIT (call_used_reg_set, i)) ||
			(TEST_HARD_REG_BIT (fixed_reg_set, i)))
		    {
			reg_value[i] = rtx_alloc (UNKNOWN);
			reg_value_mode[i] = VOIDmode;
		    }
		}
		insn = NEXT_INSN (insn);

		continue;
	    }

	    if (INSN != GET_CODE (insn))
	    {
		insn = NEXT_INSN (insn);

		continue;
	    }

	    body = PATTERN (insn);

	    /******************************************/
	    /* is this a register load of some kind ? */
	    /******************************************/

	    if ((SET == GET_CODE (body)) &&
		((REG == GET_CODE (SET_DEST (body))) ||
		 ((SUBREG == GET_CODE (SET_DEST (body))) &&
		  (REG == GET_CODE (XEXP (SET_DEST (body), 0))))))
	    {
		enum rtx_code src_code = GET_CODE (SET_SRC (body));

		rtx set_dest = ((REG == GET_CODE (SET_DEST (body))) ?
				SET_DEST (body) :
				XEXP (SET_DEST (body), 0));

		int regno = REGNO (set_dest);

		/* if the register being set is part of the address being
		   tracked as the "non_reload_address", then that address
		   must be invalidated. */

		if (reg_mentioned_p (set_dest, non_reload_address))
		{
		    non_reload_address = NULL;
		    nra_value = NULL;
		}

		/***************************************************/
		/* is the register being set to a constant value ? */
		/***************************************************/

		if ((CONST_INT == src_code) ||
		    (CONST_DOUBLE == src_code) ||
		    (CONST == src_code))
		{
		    /* does the register already have this value ? */

		    if (rtx_same_p (reg_value[regno], SET_SRC (body)) &&
			modes_compatable_p (reg_value_mode[regno],
					    GET_MODE (SET_DEST (body))))
		    {
			/* yes, eliminate a needless constant load. */

			something_changed_p = TRUE;
			PUT_CODE (insn, NOTE);
			NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
			NOTE_SOURCE_FILE (insn) = 0;
		    }
		    else
		    {
			/* note the new register contents. */

			kill_prior_contents (reg_value, regno);

			reg_value[regno] = SET_SRC (body);
			reg_value_mode[regno] =
			    GET_MODE (SET_DEST (body));

			/* note any subsequent registers clobbered by
			   the load. */
			{
			    int i = HARD_REGNO_NREGS (regno,
						      reg_value_mode[regno]);

			    while (--i)
			    {
				reg_value[regno + i] = look_to_previous_elem;
			    }
			}
		    }
		}

		/*****************************************/
		/* is this a register to register copy ? */
		/*****************************************/

		else if ((REG == src_code) ||
			 ((SUBREG == src_code) &&
			  (REG == GET_CODE (XEXP (SET_SRC (body), 0)))))
		{
		    int src_regno = ((REG == src_code) ?
				     REGNO (SET_SRC (body)) :
				     REGNO (XEXP (SET_SRC (body), 0)));

		    /* if the register already has this value, then ... */

		    if (rtx_same_p (reg_value[src_regno],
				    reg_value[regno]) &&
			modes_compatable_p (reg_value_mode[regno],
					    GET_MODE (SET_DEST (body))))
		    {

			/* ... eliminate a needless register copy. */

			something_changed_p = TRUE;
			PUT_CODE (insn, NOTE);
			NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
			NOTE_SOURCE_FILE (insn) = 0;

			/* note any slight change in mode. */

			reg_value_mode[regno] =
			    GET_MODE (SET_DEST (body));
		    }
		    else
		    {
			/* take note of the new value. */

			kill_prior_contents (reg_value, regno);

			reg_value[regno] = reg_value[src_regno];
			reg_value_mode[regno] = GET_MODE (SET_DEST (body));

			/* note any subsequently affected regs. */
			{
			    int i = HARD_REGNO_NREGS (regno,
						      reg_value_mode[regno]);

			    while (--i)
			    {
				reg_value[regno + i] = look_to_previous_elem;
			    }
			}
		    }
		}

		/***************************************/
		/* is this a memory to register load ? */
		/***************************************/

		else if ((MEM == src_code) ||
			 ((SUBREG == src_code) &&
			  (MEM == GET_CODE (XEXP (SET_SRC (body), 0)))))
		{
		    /*********************************************************/
		    /* is this a reload generated load from the stack ?      */
		    /* do we know what value resides in this stack slot ?    */
		    /* do we already have it in this register, or can we get */
		    /* it cheaper from somewhere else ?                      */
		    /*********************************************************/

		    if ((MEM == src_code) ?
			RTX_RELOAD_GENERATED_P (XEXP (SET_SRC (body), 0)) :
			RTX_RELOAD_GENERATED_P (XEXP (XEXP (
						    SET_SRC (body), 0), 0)))
		    {
			int offset =
			((MEM == src_code) ?
			 get_sf_offset (SET_SRC (body)) :
			 get_sf_offset (XEXP (SET_SRC (body), 0)));

			if (rtx_same_p (reg_value[regno],
					slot_value[offset]) &&
			    modes_compatable_p (slot_value_mode[offset],
						reg_value_mode[regno]) &&
			    modes_compatable_p (slot_value_mode[offset],
						GET_MODE (SET_DEST (body))))
			{
			    /* eliminate needless register load. */

			    something_changed_p = TRUE;
			    PUT_CODE (insn, NOTE);
			    NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
			    NOTE_SOURCE_FILE (insn) = 0;
			}
			else
			{
			    /* take note of the new value. */

			    kill_prior_contents (reg_value, regno);

			    reg_value[regno] = slot_value[offset];
			    reg_value_mode[regno] =
				GET_MODE (SET_DEST (body));

			    /* note any subsequently affected regs. */
			    {
				int i = HARD_REGNO_NREGS (regno,
						     reg_value_mode[regno]);

				while (--i)
				{
				    reg_value[regno + i] =
					look_to_previous_elem;
				}
			    }

			    /* can we get that value cheaper ? i.e from
			       a constant load or another register ? */

			    {
				/* a register transfer would be the cheapest
				   way, otherwise take a constant load if 
				   possible. */

				rtx orig_src = SET_SRC (body);
				rtx reg_val = reg_value[regno];

				int other_regno = FIRST_PSEUDO_REGISTER;

				while (other_regno--)
				{
				    if ((regno != other_regno) &&
					rtx_same_p (reg_val,
						    reg_value[other_regno])
					&& (reg_value_mode[regno] ==
					    reg_value_mode[other_regno]))
				    {
					something_changed_p = TRUE;

					SET_SRC (body) =
					    gen_rtx (REG,
						     reg_value_mode[regno],
						     other_regno);

					/* this may mean that we need to
					   "re-recog" this insn. */

					INSN_CODE (insn) = -1;

					/* a NULL reg_val means that we found
					   a suitable register source. */

					reg_val = NULL;

					break;
				    }
				}

				/* we didn't get a register, try for a
				   constant load. */

				if ((NULL != reg_val) &&
				    (orig_src == SET_SRC (body)) &&
				    ((CONSTANT_P (reg_val)) ||
				     (CONST_DOUBLE == GET_CODE (reg_val)))
				    &&
				    modes_compatable_p (GET_MODE (reg_val),
						GET_MODE (SET_DEST (body))))
				{
				    something_changed_p = TRUE;

				    SET_SRC (body) = reg_val;

				    /* this may mean that we need to
				       "re-recog" this insn. */

				    INSN_CODE (insn) = -1;
				}
			    }
			}
		    }

		    /*****************************************************/
		    /* otherwise is this a load from the last non-reload */
		    /* address stored to ? (remember we track only one   */
		    /* potentially aliased location at a time.           */
		    /*****************************************************/

		    else if ((MEM == src_code) ?
			     rtx_same_p (non_reload_address,
					 XEXP (SET_SRC (body), 0)) :
			     rtx_same_p (non_reload_address,
					 XEXP (XEXP (SET_SRC (body), 0), 0)))
		    {
			/* is the same value resident in this register 
			   already ? */

			if ((MEM == src_code) ?
			    MEM_VOLATILE_P (SET_SRC (body)) :
			    MEM_VOLATILE_P (XEXP (SET_SRC (body), 0)))
			{
			    non_reload_address = NULL;
			}
			else if (rtx_same_p (nra_value, reg_value[regno])
				 && nra_mode == reg_value_mode[regno])
			{
			    /* the value is already here. kill the needless
			       load instruction. */

			    something_changed_p = TRUE;
			    PUT_CODE (insn, NOTE);
			    NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
			    NOTE_SOURCE_FILE (insn) = 0;
			}
			else
			    /* note that we performed the load. */
			{
			    kill_prior_contents (reg_value, regno);

			    reg_value[regno] = nra_value;
			    reg_value_mode[regno] =
				GET_MODE (SET_DEST (body));

			    /* note any subsequently affected regs. */
			    {
				int i = HARD_REGNO_NREGS (regno,
						     reg_value_mode[regno]);

				while (--i)
				{
				    reg_value[regno + i] =
					look_to_previous_elem;
				}
			    }
			}
		    }

		    /****************************************************/
		    /* it's a load from a non-reload generated mem src. */
		    /****************************************************/

		    else
		    {
			enum machine_mode dest_mode =
			GET_MODE (SET_DEST (body));

			int i = HARD_REGNO_NREGS (regno, dest_mode);

			kill_prior_contents (reg_value, regno);

			reg_value[regno] = rtx_alloc (UNKNOWN);
			reg_value_mode[regno] = dest_mode;

			while (--i)
			{
			    reg_value[regno + i] = look_to_previous_elem;
			}
		    }
		}

		/*******************************************************/
		/* the register must be getting set by some sort of    */
		/* arithmetic op. create a new undefined value for it. */
		/*******************************************************/

		else
		{
		    /* mark the register value as completely undefined. */

		    reg_value[regno] = rtx_alloc (UNKNOWN);
		    reg_value_mode[regno] = GET_MODE (SET_DEST (body));

		    {
			int i = HARD_REGNO_NREGS (regno,
						  reg_value_mode[regno]);

			while (--i)
			{
			    reg_value[i + regno] = look_to_previous_elem;
			}
		    }
		}
	    }

	    /****************************************************************/
	    /* for stack slot stores:                                       */
	    /* if the source is a register, then copy the source register's */
	    /* value as the new stack slot's value.                         */
	    /* otherwise, mark the stack slot as undefined with a           */
	    /* a new UNKNOWN.                                               */
	    /****************************************************************/

	    else if ((SET == GET_CODE (body)) &&
		     (((MEM == GET_CODE (SET_DEST (body))) &&
		       (RTX_RELOAD_GENERATED_P (XEXP (SET_DEST (body),
						      0)))) ||
		      ((SUBREG == GET_CODE (SET_DEST (body))) &&
		       (MEM == GET_CODE (XEXP (SET_DEST (body), 0))) &&
		       (RTX_RELOAD_GENERATED_P (XEXP (
					  XEXP (SET_DEST (body), 0), 0))))))
	    {
		int offset = ((MEM == GET_CODE (SET_DEST (body))) ?
			      get_sf_offset (SET_DEST (body)) :
			      get_sf_offset (XEXP (SET_DEST (body), 0)));

		if (REG == GET_CODE (SET_SRC (body)))
		{
		    int regno = REGNO (SET_SRC (body));

		    kill_prior_contents (slot_value, offset);

		    slot_value[offset] = reg_value[regno];
		    slot_value_mode[offset] = GET_MODE (SET_DEST (body));

		    /* handle the case where more than one stack slot 
		       location is affected by the store. */
		    {
			int i =
			trailing_affected_slots (slot_value_mode[offset]);

			while (i)
			{
			    slot_value[offset + (i--)] =
				look_to_previous_elem;
			}
		    }
		}
		else
		    /* slot value is considered unknown. */
		{
		    kill_prior_contents (slot_value, offset);

		    slot_value[offset] = rtx_alloc (UNKNOWN);
		    slot_value_mode[offset] = GET_MODE (SET_DEST (body));

		    /* handle the case where more than one stack slot 
		       location is affected by the store. */
		    {
			int i =
			trailing_affected_slots (GET_MODE (SET_DEST (body)));

			while (i)
			{
			    slot_value[offset + (i--)] =
				look_to_previous_elem;
			}
		    }
		}
	    }

	    /**************************************************************/
	    /* otherwise if we have a non-volatile memory store, we track */
	    /* the value and the address until another store of this type */
	    /* is found, or until an instruction with non-obvious         */
	    /* side-effects is encountered.                               */
	    /**************************************************************/

	    else if ((SET == GET_CODE (body)) &&
		     (REG == GET_CODE (SET_SRC (body))) &&
		     (((MEM == GET_CODE (SET_DEST (body))) &&
		       (!MEM_VOLATILE_P (SET_DEST (body)))) ||
		      ((SUBREG == GET_CODE (SET_DEST (body))) &&
		       (MEM == GET_CODE (XEXP (SET_DEST (body), 0))) &&
		       (!MEM_VOLATILE_P (XEXP (SET_DEST (body), 0))))))
	    {
		non_reload_address = (MEM == GET_CODE (SET_DEST (body))) ?
		    XEXP (SET_DEST (body), 0) :
		    XEXP (XEXP (SET_DEST (body), 0), 0);

		nra_value = reg_value[REGNO (SET_SRC (body))];
		nra_mode = GET_MODE (SET_DEST (body));
	    }

	    /**********************************************************/
	    /* for others, note affected registers and stack slots as */
	    /* undefined.                                             */
	    /**********************************************************/

	    else
	    {
		kloads_mark_touched (body);

		/* also, we might have boottied the memory location pointed
		   to by the non-reload address. to be conservative, we
		   must nix its value. */

		non_reload_address = NULL;
		nra_value = NULL;
	    }
	    insn = NEXT_INSN (insn);
	}
    }
#if defined( USE_C_ALLOCA )
    (void) alloca (0);
#endif
}

/* an array of chars, one char corresponding to each stack slot. a 'u'
   means that the associated stack slot has been used since the last
   def, in a bottom up traversal of the basic block. a 'd' meas that
   the last reference to the slot was a def. back to back defs indicate
   that the earlier def (in the order of execution) is useless. 

   storage space for the array is allocated by kill_stores. */

static char *slot_ref;

static void
kill_stores (block_num)
int block_num;
{
    /* the maximum stack slot offset in this function. */
    int slot_count = slot_bias + 1;

    /* allocate storage for slot_ref array */
    slot_ref = (char *) alloca (sizeof (char) * slot_count);

    /* initialize slots to all 'u', anticipating potential use refs beyond the
       end of the basic block. */
    {
	int i = slot_count;

	while (i--)
	{
	    slot_ref[i] = 'u';
	}
    }

    /* traverse the block in bottom up fashion, removing useless defs of
       reload generated stack slots. */
    {
	rtx head = basic_block_head[block_num];
	rtx insn = basic_block_end[block_num];

	while (insn != head)
	{
	    if (INSN == GET_CODE (insn))
	    {
		rtx body = PATTERN (insn);

		if (SET == GET_CODE (body))
		{
		    int offset = get_sf_offset (SET_DEST (body));

		    /* do we have a store to a stack slot ? */

		    if (0 <= offset)
		    {
			/* is it a store to a reload generated slot,
			   or a possibly aliased stack slot */

			if ((SUBREG == GET_CODE (SET_DEST (body)) &&
			     RTX_RELOAD_GENERATED_P
			     (XEXP (XEXP (SET_DEST (body), 0), 0))) ||
			    RTX_RELOAD_GENERATED_P
			    (XEXP (SET_DEST (body), 0)))
			{
			    /* it is reload generated. */

			    if ('d' == slot_ref[offset])
			    {
				/* a def followed by a def. this is the
				   criteria for a 'useless' def.
				   eliminate this def. */

				something_changed_p = TRUE;
				PUT_CODE (insn, NOTE);
				NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
				NOTE_SOURCE_FILE (insn) = 0;
			    }
			    else
				/* 'u' */
			    {
				/* make the slot as defined, and continue. */

				slot_ref[offset] = 'd';
			    }
			}
			else
			{
			    /* it isn't reload generated. */

			    if ('m' == slot_ref[offset])
			    {
				/* a def followed by another def of the
				   same slot, with no possible reads from 
				   that slot inbetween. ace this def. */

				something_changed_p = TRUE;
				PUT_CODE (insn, NOTE);
				NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
				NOTE_SOURCE_FILE (insn) = 0;
			    }
			    else
			    {
				/* mark the slot as defined, and continue. */

				slot_ref[offset] = 'm';
			    }
			}
		    }
		    else
		    {
			/* note any stack slot uses within the insn. */

			kstores_mark_touched (body);
		    }
		}
		else
		{
		    /* note any stack slot uses within the insn. */

		    kstores_mark_touched (body);
		}
	    }
	    insn = PREV_INSN (insn);
	}
    }
#if defined( USE_C_ALLOCA )
    (void) alloca (0);
#endif
}

static int
modes_compatable_p (m1, m2)
enum machine_mode m1,
 m2;
{
    if ((VOIDmode == m1 || VOIDmode == m2 || m1 == m2) ||
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	((SImode == m1 || QImode == m1) &&
	 (SImode == m2 || QImode == m2)))
#elif defined( DSP56000 )
	((SImode == m1 || QImode == m1 || PSImode == m1) &&
	 (SImode == m2 || QImode == m2 || PSImode == m2)))
#elif defined( DSP96000 )
	    ((DImode == m1 || SImode == m1 || QImode == m1 || PSImode == m1) &&
	   (DImode == m2 || SImode == m2 || QImode == m2 || PSImode == m2)))
#endif
    {
	return 1;
    }
    return 0;
}

/* if the possibly SUBREGed MEM is a reference that uses the stack pointer:
   figure out the stack offset and possibly convert it to a positive number
   that excludes the arguments. if it references the frame pointer return the
   (positive) offset. otherwise, return -1. */

static int
 get_sf_offset (what)
rtx what;
{
/* horrible kludge for now */
#if defined( DSP56300 ) || defined( DSP56600 )
    static int slot_bias;

#endif

    what = clean_subregs (what);

    if (MEM == GET_CODE (what))
    {
	what = XEXP (what, 0);

	if (PLUS == GET_CODE (what) && REG_P (XEXP (what, 0))
	    && CONST_INT == GET_CODE (XEXP (what, 1)))
	{
	    int offset = INTVAL (XEXP (what, 1));

#if defined( HAVE_L_MEMORY ) && defined( TARGET_PACK_L_MEMORY )
	    if (RTX_ALTERNATE_MEMORY_P (what))
	    {
		offset += x_slot_base;
	    }
#endif
#if defined( HAVE_FRAME_POINTER )
	    if (FRAME_POINTER_REGNUM == REGNO (XEXP (what, 0)) &&
		offset < slot_bias && offset >= 0)
	    {
		return offset;
	    }
#else
	    if (STACK_POINTER_REGNUM == REGNO (XEXP (what, 0)) &&
		-offset < slot_bias && offset <= 0)
	    {
		return slot_bias + offset;
	    }
#endif
	}
    }
    return -1;
}

/* indicate whether, and how many, subsequent stack slots will be affected
   by a memory write with the given mode. */

static int
 trailing_affected_slots (mode)
enum machine_mode mode;
{
#if defined( DSP56100 )
    if ((DImode == mode) || (SFmode == mode) || (DFmode == mode))
    {
	return 1;
    }
#elif defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
    if ('l' == memory_model)
    {
	return 0;
    }
    else if ((DImode == mode) || (SFmode == mode) || (DFmode == mode))
    {
	return 1;
    }
#elif defined( DSP96000 )
    if ('l' == memory_model)
    {
	return 0;
    }
    else if ((DImode == mode) || (DFmode == mode))
    {
	return 1;
    }
#endif
    return 0;
}

/* do a and b imply the same value ? we have to check for UNKNOWNs, as 
   rtx_equal_p will abort on them. */

static int
 rtx_same_p (a, b)
rtx a,
 b;
{
    if ((NULL == a) ||
	(NULL == b))
    {
	return 0;
    }

    if ((UNKNOWN == GET_CODE (a)) ||
	(UNKNOWN == GET_CODE (b)))
    {
	return 0;
    }

    if (a == b)
    {
	return 1;
    }

    return rtx_equal_p (a, b);
}

static void
 kstores_mark_touched (body)
rtx body;
{
    if (SET == GET_CODE (body))
    {
	int offset,
	 i;

	body = clean_subregs (SET_SRC (body));

	if (MEM == GET_CODE (body))
	{
	    /* two cases to look for:
	       1) reads from a specific stack slot, and
	       2) mem reads in general. */

	    offset = get_sf_offset (body);

	    if (0 <= offset)
	    {
		/* case 1: the read is specific. */

		i = trailing_affected_slots (GET_MODE (body));

		while (i)
		{
		    slot_ref[(i--) + offset] = 'u';
		}
		slot_ref[offset] = 'u';
	    }
	    else
	    {
		/* case 2: we don't know the address. */

		for (i = 0; i < slot_bias + 1; ++i)
		{
		    if ('m' == slot_ref[i])
		    {
			slot_ref[i] = 'u';
		    }
		}
	    }
	}
    }
    else
    {
	char *fmt = GET_RTX_FORMAT (GET_CODE (body));
	char c,
	*p = fmt;

	while ('\0' != (c = *p))
	{
	    switch (c)
	    {
	    case 'e':
		kstores_mark_touched (XEXP (body, p - fmt));
		break;

	    case 'E':
		{
		    int i = XVECLEN (body, p - fmt);

		    while (i--)
		    {
			kstores_mark_touched (XVECEXP (body, p - fmt, i));
		    }
		}
	    }
	    ++p;
	}
    }
}

static void
 kloads_mark_touched (body)
rtx body;
{
    if ((SET == GET_CODE (body)) ||
	(CLOBBER == GET_CODE (body)) ||
	(PRE_INC == GET_CODE (body)) ||
	(PRE_DEC == GET_CODE (body)) ||
	(POST_INC == GET_CODE (body)) ||
	(POST_DEC == GET_CODE (body)))
    {
	body = XEXP (body, 0);

	if (((MEM == GET_CODE (body)) &&
	     (RTX_RELOAD_GENERATED_P (XEXP (body, 0)))) ||
	    ((SUBREG == GET_CODE (body)) &&
	     (MEM == GET_CODE (XEXP (body, 0))) &&
	     (RTX_RELOAD_GENERATED_P (XEXP (XEXP (body, 0), 0)))))
	{
	    /* blow out any previous value overlapping this load
	       target. */

	    int offset = get_sf_offset (body);
	    int i = trailing_affected_slots (GET_MODE (body));

	    kill_prior_contents (slot_value, offset);

	    slot_value[offset] = rtx_alloc (UNKNOWN);
	    slot_value_mode[offset] = GET_MODE (body);

	    while (i)
	    {
		slot_value[(i--) + offset] = look_to_previous_elem;
	    }
	    kloads_mark_touched (XEXP (body, 0));

	    return;
	}

	else if (REG == GET_CODE (body))
	{
	    /* waste any register values overlapping with the target
	       of this INSN. */

	    int regno = REGNO (body);
	    int i = HARD_REGNO_NREGS (regno, GET_MODE (body));

	    kill_prior_contents (reg_value, regno);

	    reg_value[regno] = rtx_alloc (UNKNOWN);
	    reg_value_mode[regno] = GET_MODE (body);

	    while (--i)
	    {
		reg_value[i + regno] = look_to_previous_elem;
	    }
	}
    }
    else
    {
	char *fmt = GET_RTX_FORMAT (GET_CODE (body));
	char c,
	*p = fmt;

	while ('\0' != (c = *p))
	{
	    switch (c)
	    {
	    case 'e':
		kloads_mark_touched (XEXP (body, p - fmt));
		break;

	    case 'E':
		{
		    int i = XVECLEN (body, p - fmt);

		    while (i--)
		    {
			kloads_mark_touched (XVECEXP (body, p - fmt, i));
		    }
		}
	    }
	    ++p;
	}
    }
}
/* note the uses of stack slots within the RTL expression passed as an 
   argument. here, I'm being conservative in that I'm not checking whether
   or not a reference to a reload generated stack slot is a use, or a def:
   I'm counting all references as uses. I don't think we'll ever see a
   reload address inside a complex expression such as PARALLEL, but the
   extra code protects against deleting needed defs just in case. */

static void
 compliv_mark_touched (body)
rtx body;
{
    if ((SET == GET_CODE (body)) ||
	(CLOBBER == GET_CODE (body)) ||
	(PRE_INC == GET_CODE (body)) ||
	(PRE_DEC == GET_CODE (body)) ||
	(POST_INC == GET_CODE (body)) ||
	(POST_DEC == GET_CODE (body)))
    {
	body = XEXP (body, 0);

	if (((MEM == GET_CODE (body)) &&
	     (RTX_RELOAD_GENERATED_P (XEXP (body, 0)))) ||
	    ((SUBREG == GET_CODE (body)) &&
	     (MEM == GET_CODE (XEXP (body, 0))) &&
	     (RTX_RELOAD_GENERATED_P (XEXP (XEXP (body, 0), 0)))))
	{
	    /* blow out any previous value overlapping this load
	       target. */

	    int offset = get_sf_offset (body);
	    int i = trailing_affected_slots (GET_MODE (body));

	    kill_prior_contents (slot_value, offset);

	    slot_value[offset] = NULL;
	    slot_value_mode[offset] = VOIDmode;

	    while (i)
	    {
		slot_value[(i--) + offset] = NULL;
	    }
	    compliv_mark_touched (XEXP (body, 0));

	    return;
	}
    }
    else
    {
	char *fmt = GET_RTX_FORMAT (GET_CODE (body));
	char c,
	*p = fmt;

	while ('\0' != (c = *p))
	{
	    switch (c)
	    {
	    case 'e':
		compliv_mark_touched (XEXP (body, p - fmt));
		break;

	    case 'E':
		{
		    int i = XVECLEN (body, p - fmt);

		    while (i--)
		    {
			compliv_mark_touched (XVECEXP (body, p - fmt, i));
		    }
		}
	    }
	    ++p;
	}
    }
}

/* this function takes an array of rtx values that represent either adjacent
   memory locations or registers. it mark as invalid any data value that 
   encompasses the element specified by the index. e.g. if we do a SImode load
   into a DImode stack slot, then we must kill ALL of the DImode value. We
   could be less conservative with constant values, maybe we'd want to add that
   later (yeah-right). */

static void
 kill_prior_contents (rtx_array, index)
rtx *rtx_array;
int index;
{
    while (look_to_previous_elem == rtx_array[index])
    {
	rtx_array[index--] = NULL;
    }
    rtx_array[index] = NULL;
}

/* this function traverses a basic block and tries to eliminate some spills
   that are needlessly created by poor lifetime management. we often see
   the following code sequence:

   (1) ( SET ( MEMx ) ( R1 ))
   ...
   (2) ( SET ( R1 ) ( any value ))
   ...
   (3) ( SET ( R2 ) ( MEMx ))

   we want to reorder this as (1) (3) (2). This will allow kill_stores to 
   possibly elininate (1), and allow kill_loads to transform (3) into a 
   register copy. The only dflow requirement is that R2 cannot be live between
   (2) and (3) in the original insn ordering. It is easier for us 
   (computationally) to instead ensure that R2 is not live between (1) and (3).
   This could cause us to miss a potential opportunity, but in the code
   I've examined (durbin.c), it wouldn't */

/* we do this in one traversal of the basic block, making an extra subpass
   for each optimization opportunity. */

static void
 compress_lives (block_num)
int block_num;
{
    /* the maximum stack slot offset in this function. */
    int slot_count = slot_bias + 1;

    /* scan through the basic block once, in execution order. */
    rtx insn = basic_block_head[block_num];
    rtx tail = basic_block_end[block_num];

    /* allocate memory for the grouping marker. */
    look_to_previous_elem = (rtx) alloca (sizeof (struct rtx_def));

    /* we keep an array of last sets, an entry for each stack slot. the
       insn rtx of the last insn to SET the slot is kept in the array. 
       NULL implies ambiguity. look_to_previous_elem indicates a mult-slot
       SET. */
    slot_value = (rtx *) alloca (sizeof (rtx) * slot_count);

    /* we keep an array of modes corresponding to the above insn array. */
    slot_value_mode = (enum machine_mode *)
	alloca (sizeof (enum machine_mode) * slot_count);

    /* initialize the data-structures. */
    {
	int i = slot_count;

	while (i--)
	{
	    slot_value[i] = NULL;
	    slot_value_mode[i] = VOIDmode;
	}
    }

    /* algorithm (loosely speaking): 
       traverse the block in execution order. if the current insn is a 
       store to a memory stack slot, then we set the slot_value element 
       corrsponding to the slot to point to the current insn. the 
       slot_value_mode
       array element is given the mode used access the stack slot.
       if the current insn is a load from a stack slot, see if we can 
       compress a lifetime. otherwise, note any stack slot references by
       setting the corresponding slot_value array element to NULL. 

       to compress a lifetime, check whether R2 is referenced within the
       interval bounded by the current insn and the insn specified by
       the corresponding element of the slot_value array. if not, copy the 
       PATTERN of the current insn into a new insn following the insn
       specified by the slot_value array (whew!). delete the current insn.
       kill_loads, (and possibly kill_stores) will make the code more
       efficient. */

    while (insn != tail)
    {
	rtx body;

	if (INSN != GET_CODE (insn))
	{
	    insn = NEXT_INSN (insn);

	    continue;
	}
	body = PATTERN (insn);

	/* do we have a store to a stack slot ? */

	if ((SET == GET_CODE (body)) &&
	    (REG == GET_CODE (SET_SRC (body))) &&
	    (MEM == GET_CODE (SET_DEST (body))) &&
	    (RTX_RELOAD_GENERATED_P (XEXP (SET_DEST (body), 0))))
	{
	    int offset = get_sf_offset (SET_DEST (body));

	    kill_prior_contents (slot_value, offset);

	    slot_value[offset] = insn;
	    slot_value_mode[offset] = GET_MODE (SET_SRC (body));

	    /* handle the case where the the mode uses more than one memory
	       location. */
	    {
		int i = trailing_affected_slots (slot_value_mode[offset]);

		while (i)
		{
		    slot_value[(i--) + offset] = look_to_previous_elem;
		}
	    }
	}

	/* do we have a load from a stack slot ? */

	else if ((SET == GET_CODE (body)) &&
		 (REG == GET_CODE (SET_DEST (body))) &&
		 (MEM == GET_CODE (SET_SRC (body))) &&
		 (RTX_RELOAD_GENERATED_P (XEXP (SET_SRC (body), 0))))
	{
	    int offset = get_sf_offset (SET_SRC (body));
	    enum machine_mode mode = GET_MODE (SET_SRC (body));

	    if ((NULL != slot_value[offset]) &&
		(insn != NEXT_INSN (slot_value[offset])) &&
		(look_to_previous_elem != slot_value[offset]) &&
		(mode == slot_value_mode[offset]) &&
		(!reg_possibly_live_between_p (SET_DEST (body),
					       slot_value[offset], insn)))
	    {
		/* shorten the lifetime of a spilled register. */

		/* move the load to immediately after the store. */

		emit_insn_after (copy_rtx (PATTERN (insn)),
				 slot_value[offset]);

		/* eliminate the second load. */

		something_changed_p = TRUE;
		PUT_CODE (insn, NOTE);
		NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
		NOTE_SOURCE_FILE (insn) = 0;
	    }
	}

	/* note any stack slots that are clobbered by the current insn. */

	else
	{
	    compliv_mark_touched (PATTERN (insn));
	}

	/* move on to the next insn */

	insn = NEXT_INSN (insn);
    }
#if defined( USE_C_ALLOCA )
    (void) alloca (0);
#endif
}

/* we go through the function body once and find all of the stores to 
   unused reload stack slots. we keep an array of void*, with an element
   corresponding to each reload stack offset location. initially, all
   elements are set to NULL. if a slot is ever loaded from, its element
   is changed to & is_used_p. if a slot is stored to, and its element is
   is still NULL, a list is created and placed in the element. the list 
   will contain all of the insns that store to this slot. 
   after we've passed through the function, we go back through the array
   and delete all of the insns in each list. these isns are the stores into
   unused stack slots. */

struct store_note
{
    rtx which_insn;
    struct store_note *link;
};

static void
 global_kill_stores (insns)
rtx insns;
{
    rtx look = insns;

    struct store_note *free_list = NULL;

    /* the maximum stack slot offset in this function. */
    int i,
     subreg = 0,
     is_used_p,
     slot_count = slot_bias + 1;

    /* the array used to track unused stack slots. */
    void **usage_array = (void **)
     xmalloc (sizeof (void *) * slot_count);

    /* initialize record keeper array */
    for (i = 0; i < slot_count; ++i)
    {
	usage_array[i] = NULL;
    }

    /* look through the function. */
    while (NULL != look)
    {
	/* is this possibly a read from a reload stack slot ? */
	if (((subreg = 0), (INSN == GET_CODE (look))) &&
	    SET == GET_CODE (PATTERN (look)) &&
	    (MEM == GET_CODE (SET_SRC (PATTERN (look))) ||
	     (subreg = (SUBREG == GET_CODE (SET_SRC (PATTERN (look))) &&
			MEM == GET_CODE (XEXP (SET_SRC (PATTERN (look))
					       ,0))))) &&
	    RTX_RELOAD_GENERATED_P
	    ((subreg) ? XEXP (XEXP (SET_SRC (PATTERN (look)), 0), 0) :
	     XEXP (SET_SRC (PATTERN (look)), 0)))
	{
	    int slot_num = get_sf_offset (SET_SRC (PATTERN (look)));
	    int trailing_slots = trailing_affected_slots
	     (GET_MODE ((subreg) ?
			XEXP (SET_SRC (PATTERN (look)), 0)
			: SET_SRC (PATTERN (look))));

	    do
	    {
		/* deallocate any previously built chain. */
		if (NULL != usage_array[slot_num] &&
		    (&is_used_p) != usage_array[slot_num])
		{
		    struct store_note *dead,
		    *next = usage_array[slot_num];

		    do
		    {
			dead = next;
			next = next->link;
			dead->link = free_list;
			free_list = dead;
		    }
		    while (NULL != next);
		}
		usage_array[slot_num++] = (void *) &is_used_p;
	    }
	    while (trailing_slots--);
	}

	/* otherwise is this possibly a write to a reload stack slot ? */
	else if (((subreg = 0), (INSN == GET_CODE (look))) &&
		 SET == GET_CODE (PATTERN (look)) &&
		 (MEM == GET_CODE (SET_DEST (PATTERN (look))) ||
		  (subreg =
		   (SUBREG == GET_CODE (SET_DEST (PATTERN (look))) &&
		    MEM == GET_CODE (XEXP (SET_DEST (PATTERN (look))
					   ,0))))) &&
		 RTX_RELOAD_GENERATED_P
		 ((subreg) ?
		  XEXP (XEXP (SET_DEST (PATTERN (look)), 0), 0) :
		  XEXP (SET_DEST (PATTERN (look)), 0)))
	{
	    int dont_bother = 0;
	    int i,
	     slot_num = get_sf_offset (SET_DEST (PATTERN (look)));
	    int j,
	     trailing_slots = trailing_affected_slots
	     (GET_MODE ((subreg) ?
			XEXP (SET_DEST (PATTERN (look)), 0) :
			SET_DEST (PATTERN (look))));

	    i = slot_num, j = trailing_slots;
	    do
	    {
		dont_bother |= (((void *) &is_used_p)
				== usage_array[i++]);
	    }
	    while (j--);

	    /* note the store if we should. */

	    if (!dont_bother)
	    {
		struct store_note *new;

		if (NULL == free_list)
		{
		    new = (struct store_note *) xmalloc
			(sizeof (struct store_note));
		}
		else
		{
		    new = free_list;
		    free_list = free_list->link;
		}
		new->which_insn = look;
		new->link = (struct store_note *) usage_array[slot_num];
		usage_array[slot_num] = (void *) new;
	    }
	}

	/* move on to the next insn */
	look = NEXT_INSN (look);
    }

    /* now, blow out the worthless stores. */
    for (i = 0; i < slot_count; ++i)
    {
	if (NULL != usage_array[i] &&
	    ((void *) &is_used_p) != usage_array[i])
	{
	    struct store_note *dead;
	    struct store_note *next = (struct store_note *) usage_array[i];

	    do
	    {
		dead = next;
		PUT_CODE (dead->which_insn, NOTE);
		NOTE_LINE_NUMBER (dead->which_insn) = NOTE_INSN_DELETED;
		NOTE_SOURCE_FILE (dead->which_insn) = 0;
		next = dead->link;
		free (dead);
	    }
	    while (NULL != next);
	}
    }
    /* deallocate the whole array. */
    free (usage_array);

    while (NULL != free_list)
    {
	struct store_note *dead = free_list;

	free_list = free_list->link;
	free (dead);
    }
}

/* return 1 if the register(s) referenced by the REG rtx hard_reg might
   be referenced by any of the insns between first and last. return 0 
   otherwise. */

static int
 reg_possibly_live_between_p (hard_reg, first, last)
rtx hard_reg,
 first,
 last;
{

    if (first == last)
    {
	return 0;
    }
    first = NEXT_INSN (first);

    while (first != last)
    {
	if ((INSN == GET_CODE (first)) ||
	    (JUMP_INSN == GET_CODE (first)))
	{
	    if (reg_possibly_live (hard_reg, PATTERN (first)))
	    {
		return 1;
	    }
	}
	else if (CALL_INSN == GET_CODE (first))
	{
	    /* look for the usual usage */
	    if (reg_possibly_live (hard_reg, PATTERN (first)))
	    {
		return 1;
	    }

	    /* check for CALL_USED/CLOBBERED_REGS */
	    {
		int i = REGNO (hard_reg);
		int trailers = HARD_REGNO_NREGS (i, GET_MODE (hard_reg));

		do
		{
		    if ((TEST_HARD_REG_BIT (call_used_reg_set, i)) ||
			(TEST_HARD_REG_BIT (fixed_reg_set, i++)))
		    {
			return 1;
		    }
		}
		while (trailers--);
	    }
	}
	first = NEXT_INSN (first);
    }
    return 0;
}

static int
 reg_possibly_live (reg, pat)
rtx reg,
 pat;
{
    if (REG == GET_CODE (pat))
    {
	int r1 = REGNO (reg),
	 r2 = REGNO (pat);
	int l1 = HARD_REGNO_NREGS (r1, GET_MODE (reg)) - 1;
	int l2 = HARD_REGNO_NREGS (r2, GET_MODE (pat)) - 1;

	if (((r1 <= r2) && ((r1 + l1) >= r2)) ||
	    ((r2 <= r1) && ((r2 + l2) >= r1)))
	{
	    return 1;
	}
	return 0;
    }
    else
    {
	char *fmt = GET_RTX_FORMAT (GET_CODE (pat));
	char c,
	*p = fmt;

	while ('\0' != (c = *p))
	{
	    switch (c)
	    {
	    case 'e':
		if (reg_possibly_live (reg, XEXP (pat, p - fmt)))
		{
		    return 1;
		}
		break;

	    case 'E':
		{
		    int i = XVECLEN (pat, p - fmt);

		    while (i--)
		    {
			if (reg_possibly_live (reg,
					       XVECEXP (pat, p - fmt, i)))
			{
			    return 1;
			}
		    }
		}
		break;
	    }
	    ++p;
	}
	return 0;
    }
}

/* like reg_set_p, as provided by GNU, but only looks for mods that are
   the object of a set. */
static int
 addr_reg_set_p (reg, pat)
rtx reg,
 pat;
{
    if (SET == GET_CODE (pat))
    {
	enum machine_mode mode;

	pat = XEXP (pat, 0);
	mode = GET_MODE (pat);

	while (SUBREG == GET_CODE (pat))
	{
	    pat = XEXP (pat, 0);
	}

	if (REG == GET_CODE (pat))
	{
	    int p_regno = REGNO (pat);
	    int p_len = HARD_REGNO_NREGS (p_regno, GET_MODE (pat));
	    int regno = REGNO (reg);
	    int len = HARD_REGNO_NREGS (regno, GET_MODE (reg));

	    if (((p_regno <= regno) && ((p_regno + p_len) >= regno)) ||
		((regno <= p_regno) && ((regno + len) >= p_regno)))
	    {
		return 1;
	    }
	}
    }

    {
	char *fmt = GET_RTX_FORMAT (GET_CODE (pat));
	char c,
	*p = fmt;

	while ('\0' != (c = *p))
	{
	    switch (c)
	    {
	    case 'e':
		if (addr_reg_set_p (reg, XEXP (pat, p - fmt)))
		{
		    return 1;
		}
		break;

	    case 'E':
		{
		    int i = XVECLEN (pat, p - fmt);

		    while (i--)
		    {
			if (addr_reg_set_p (reg,
					    XVECEXP (pat, p - fmt, i)))
			{
			    return 1;
			}
		    }
		}
	    }
	    ++p;
	}
    }
    return 0;
}

static char *
 fms (has_ats)
const char *has_ats;
{
    static char buffer[BUFSIZ];
    char *at_here = buffer;

    (void) strcpy (buffer, has_ats);
    while (NULL != (at_here = strchr (at_here, '@')))
    {
	*at_here = mem_space;
    }
    return buffer;
}

void
 fix_sp_fp (template_string)
char *template_string;
{
    char *p;

    for (p = template_string; '\0' != *p; ++p)
    {
	if ('F' == *p)
	{
	    ++p;

	    if ('P' == *p)
	    {
		*(--p) = 'r';
		*(++p) = '0' +
		    (FRAME_POINTER_REGNUM - FIRST_ADDRESS_REGISTER);
	    }
	    else if ('O' == *p)
	    {
		*(--p) = 'n';
		*(++p) = '0' +
		    (FRAME_POINTER_REGNUM - FIRST_ADDRESS_REGISTER);
	    }
	}
	else if ('S' == *p)
	{
	    ++p;

	    if ('P' == *p)
	    {
		*(--p) = 'r';
		*(++p) = '0' +
		    (STACK_POINTER_REGNUM - FIRST_ADDRESS_REGISTER);
	    }
	    else if ('O' == *p)
	    {
		*(--p) = 'n';
		*(++p) = '0' +
		    (STACK_POINTER_REGNUM - FIRST_ADDRESS_REGISTER);
	    }
	}
    }
}

/*
 * Print out the section header and assembler defines / options at the
 * start of file.
 */
#if !defined( MSIL_DISABLE )
char *sec_name;

void
 mark_start_of_zini (file)
FILE *file;
{
    fprintf (file, "%s\n", DATA_SECTION_ASM_OP);
    fprintf (file, "%s_start_zero_data\n", sec_name);

}

void
 select_init_section (file)
FILE *file;
{
    fprintf (file, "\tsection %s_init_data\n", sec_name);
    fprintf (file, "%s\n", INIT_SECTION_ASM_OP);

}

void
 end_init_section (file)
FILE *file;
{
    fprintf (file, "\tendsec\n\n");
    fprintf (file, "%s\n", DATA_SECTION_ASM_OP);

}

#endif

void
 asm_file_start (file)
FILE *file;
{
    char *section,
    *replace;

#if defined( SDB_DEBUGGING_INFO )
    if (write_symbols == SDB_DEBUG)
    {
	sdbout_filename (file, main_input_filename);
    }
#endif

    section = strcpy ((char *) xmalloc (strlen (main_input_filename) + 1),
		      main_input_filename);

    while (replace = strpbrk (section, ".-"))
    {
	*replace = '_';
    }

    /* use last part of path only */
    if ((NULL != (replace = strrchr (section, '/'))) ||
	(NULL != (replace = strrchr (section, '\\'))))
    {
	section = replace + 1;
    }

    fprintf (file, "\tsection\t%s\n", section);
#if defined( DSP56300 ) || defined( DSP56600 )

#if !defined( MSIL_DISABLE )
    fprintf (file, "\topt so,nomd,noidw\n\tpage 132,66,3,3\n");
#else
    fprintf (file, "\topt so,nomd,noidw,nonde\n\tpage 132,66,3,3\n");
#endif
#else
    fprintf (file, "\topt so,nomd\n\tpage 132,66,3,3\n");
#endif

#if !defined( MSIL_DISABLE )
    if (TARGET_AUTO_INIT)
    {
	sec_name = section;

	fprintf (file, "%s\n", DATA_SECTION_ASM_OP);
	fprintf (file, "%s_start\n", sec_name);

	SELECT_INIT_SECTION (file);

	fprintf (file, "%s_start_init\n", sec_name);
	fprintf (file, "\tsection init_table\n");
	fprintf (file, "\torg p:*-1\n");
	fprintf (file, "\tdc\t%s_start\n", sec_name);
	fprintf (file, "%s\n", INIT_SECTION_ASM_OP);
	fprintf (file, "\tdc\t%s_start_init\n", sec_name);
	fprintf (file, "\n\tendsec\n");
	END_INIT_SECTION (file);
    }

#endif

}

void
 asm_file_end (file)
FILE *file;
{
#if !defined( MSIL_DISABLE )
    if (TARGET_AUTO_INIT)
    {
	fprintf (file, "%s\n", DATA_SECTION_ASM_OP);
	fprintf (file, "%s_end\n", sec_name);
	SELECT_INIT_SECTION (file);

	fprintf (file, "%s_end_init\n", sec_name);
	fprintf (file, "\tsection init_table\n");
	fprintf (file, "\torg p:\n");
	fprintf (file, "F__init_size\tset\t(%s_end_init-%s_start_init)\n", sec_name, sec_name);
	fprintf (file, "F__zinit_size\tset\t(%s_end-%s_start-F__init_size)\n", sec_name, sec_name);
	fprintf (file, "\tdc\tF__init_size\n");
	fprintf (file, "\tdc\tF__zinit_size\n");
	fprintf (file, "\tdc\t(-1)\n");

	fprintf (file, "F__data_size\tset\t(%s_start_zero_data-%s_start)\n", sec_name, sec_name);
	fprintf (file, "\n\tIF (F__data_size!=F__init_size)\n");
	fprintf (file, "\n\tFAIL 'Internal compiler error: contact dsphelp@dsp.sps.mot.com'");
	fprintf (file, "\n\tENDIF\n");

	END_INIT_SECTION (file);
	fprintf (file, "\n\tendsec\n");
    }

#endif

    fprintf (file, "\n\tendsec\n");
}

void
 output_ascii (file, P, size)
FILE *file;
char *P;
int size;
{
    int i;

    for (i = 0; i < size; i++)
    {
	if (0 == (i % 15))
	{
	    fprintf (file, "\n\tdc\t");
	}
	else
	{
	    putc (',', file);
	}

	if ((0x40 < *P && 0x5b > *P) || (0x60 < *P && 0x7b > *P) ||
	    (0x2f < *P && 0x3a > *P))
	{
	    putc ('\'', file);
	    fputc (*P++, file);
	    putc ('\'', file);
	}
	else
	{
	    fprintf (file, "$%02X", *P++);
	}
    }
    putc ('\n', file);
}

/*
 * For local lables to be useful within inline assembly calls, A unique
 * global lable must be emmited before and after the each such invocation.
 */
char *
 asm_app_toggle ()
{
    static int LabelNo = 0;
    static char Label[20];

    sprintf (Label, "ASM_APP_%X\n", LabelNo++);
    return (Label);
}

int
 multiply_operator (op, mode)
rtx op;
enum machine_mode mode;
{
    return (SImode == mode &&
	    (MULT == GET_CODE (op) || UMULT == GET_CODE (op)));
}

int
 plus_or_minus_operator (op, mode)
rtx op;
enum machine_mode mode;
{
    return (PLUS == GET_CODE (op) || MINUS == GET_CODE (op));
}

int
 alu_one_operand_operator (op, mode)
rtx op;
enum machine_mode mode;
{
    enum rtx_code code = GET_CODE (op);

    return (NEG == code || ABS == code);
}

int
 alu_two_operand_operator (op, mode)
rtx op;
enum machine_mode mode;
{
    enum rtx_code code = GET_CODE (op);

    return (PLUS == code || MINUS == code ||
	    ASHIFT == code || ASHIFTRT == code ||
	    LSHIFT == code || LSHIFTRT == code);
}

/* return 1 iff we've got a conditional. */
int
 cond_operator (op, mode)
rtx op;
enum machine_mode mode;
{
    switch (GET_CODE (op))
    {
    case NE:
    case EQ:
    case GE:
    case GT:
    case LE:
    case LT:
#if ! defined( HAVE_UNSIGNED_BRANCH_CODES )
    case NEU:
    case EQU:
#endif
    case GEU:
    case GTU:
    case LEU:
    case LTU:
    case CC:
    case CS:
	return 1;

    default:
	return 0;
    }
}

int
 hard_register_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    if (!register_operand (op, mode))
    {
	return 0;
    }
    op = clean_subregs (op);

    return FIRST_PSEUDO_REGISTER > REGNO (op);
}

int
 crement_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    if (CONST_DOUBLE == GET_CODE (op))
    {
	return (((1 == CONST_DOUBLE_LOW (op)) &&
		 (0 == CONST_DOUBLE_HIGH (op))) ||
		((-1 == CONST_DOUBLE_LOW (op)) &&
		 (-1 == CONST_DOUBLE_HIGH (op))));

    }
    else if (CONST_INT == GET_CODE (op))
    {
	return (1 == INTVAL (op) || -1 == INTVAL (op));
    }
    return 0;
}

int
 register_or_immediate_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (immediate_operand (op, mode) || register_operand (op, mode));
}

int
 register_or_crement_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (crement_operand (op, mode) || register_operand (op, mode));
}

int
 shift_count_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (register_operand (op, mode) ||
	    (immediate_operand (op, mode) &&
	     0 < INTVAL (op) && 24 > INTVAL (op)));
}

int
 register_or_1_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (register_operand (op, mode) || const1_rtx == op);
}

int
 one_inclusive_eight_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (CONST_INT == GET_CODE (op) &&
	    0 < INTVAL (op) && 9 > INTVAL (op));
}

int
 minus_two_inclusive_two_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (CONST_INT == GET_CODE (op) &&
	    -3 < INTVAL (op) && 3 > INTVAL (op));
}

int
 pow_two_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    if (CONST_INT != GET_CODE (op) || !INTVAL (op))
    {
	return 0;
    }

    {
	/* find the first bit. */
	unsigned int mask = 1,
	 count = HOST_BITS_PER_INT;
	unsigned int value = INTVAL (op);

	while (--count)
	{
	    if (mask & value)
	    {
		break;
	    }
	    mask <<= 1;
	}
	/* clear that bit and see if we get a zero. */
	value &= (~mask);

	return (0 == value);
    }
}

int
 register_or_pow_two_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    if (pow_two_operand (op, mode))
    {
	return TRUE;
    }
    return register_operand (op, mode);
}

register_or_indirect_memory_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (register_operand (op, mode) ||
	    indirect_memory_operand (op, mode));
}

int
 pic_sensitive_call_operand (op, mode)
rtx op;
enum machine_mode mode;
{
#if defined( TARGET_PIC )
    if (TARGET_PIC)
    {
	return restricted_memory_operand (op, mode);
    }
#endif
    return register_or_restricted_memory_operand (op, mode);
}

int
 register_or_restricted_memory_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (register_operand (op, mode) ||
	    restricted_memory_operand (op, mode));
}

int
 indirect_memory_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (MEM == GET_CODE (op) &&
	    (GET_MODE (op) == mode || VOIDmode == mode) &&
	    indirect_memref_p (op));
}

int
 direct_memory_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (MEM == GET_CODE (op) &&
	    (GET_MODE (op) == mode || VOIDmode == mode) &&
	    direct_memref_p (op));
}

int
 restricted_memory_operand (op, mode)
rtx op;
enum machine_mode mode;
{
    return (MEM == GET_CODE (op) &&
	    (GET_MODE (op) == mode || VOIDmode == mode) &&
	    restricted_memref_p (op));
}

/* this function returns 1 if it is sure that mem isn't going to
   yeild a (rn+#xx) addressing mode in final code generation, 0 otherwise. */
int
 restricted_memref_p (mem)
rtx mem;
{
    rtx addr = XEXP (mem, 0);

    if (CONSTANT_ADDRESS_P (addr))
    {
	return 1;
    }

    switch (GET_CODE (addr))
    {
    case PLUS:
	return 0;

    case POST_INC:
    case POST_DEC:
	return REG_P (XEXP (addr, 0));

    case REG:
	if (FRAME_POINTER_REGNUM == REGNO (addr))
	{
	    return 0;
	}
	return 1;

    default:
	return 0;
    }
}

/* this function returns 1 if it is sure that mem isn't going to
   yeild a (rn+#xx) or direct addressing mode in final code
   generation, 0 otherwise. */
int
 indirect_memref_p (mem)
rtx mem;
{
    rtx addr = XEXP (mem, 0);

    if (CONSTANT_ADDRESS_P (addr))
    {
	return 0;
    }

    switch (GET_CODE (addr))
    {
    case PLUS:
	return 0;

    case POST_INC:
    case POST_DEC:
	return REG_P (XEXP (addr, 0));

    case REG:
	if (FRAME_POINTER_REGNUM == REGNO (addr))
	{
	    return 0;
	}
	return 1;

    default:
	return 0;
    }
}

/* this function returns 1 if it is sure that mem is going to be
   a direct address. */
int
 direct_memref_p (mem)
rtx mem;
{
    rtx addr = XEXP (mem, 0);

    if (CONSTANT_ADDRESS_P (addr))
    {
	return 1;
    }

    if (PLUS == GET_CODE (addr))
    {
	return CONSTANT_ADDRESS_P (XEXP (addr, 0)) &&
	    CONSTANT_ADDRESS_P (XEXP (addr, 0));
    }

    return 0;
}

/* The following is an ad-hoc scheme for making sure we don't generate
 * do loops with pipline conflicts. Note that there is an insn for the "do",
 * and another for the "od" (end-o-do). The latter never actually causes 
 * an instruction in the assembly language output; it is used to indicate
 * the implicit back-branch at the do target label to the optimizer.
 *
 * 1) When we emit a do insn, peek forward to the next insn. If this insn
 *    uses an address register, mark the corresponding bit in the stack of
 *    register use bit masks, and increment the stack pointer.
 *
 * 2) When we emit an od insn, peek backward to the previous insn. if it
 *    "move"s to an address register, check the corresponding bit in the bit
 *    mask at the top of the stack. if it is set, we have a conflict. Emit
 *    a nop to avert the pipeline hazard. In either case, decrement the
 *    bit mask stack pointer before reading from the stack.
 *
 * NOTE: we use unsigned char to contain the bit mask because we have 
 *       only eight address registers on either chip.
 */

unsigned int do_loop_nest_level = 0;
unsigned char bit_mask_stack[MAX_DO_LOOP_NESTING];

#if defined( DSP56100 )
#define ADDR_REG_MASK 0x0f
#else
#define ADDR_REG_MASK 0xff
#endif

void
 record_address_regs_used (pattern)
rtx pattern;
{
    unsigned char mask = 0,
     flag = 1;
    rtx reg;

#if defined( DSP56000 ) || defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
    reg = gen_rtx (REG, Pmode, 6);
#elif defined( DSP96000 )
    reg = gen_rtx (REG, Pmode, 25);
#endif

    for (; ADDR_REG_MASK & flag; flag <<= 1, ++REGNO (reg))
    {
	if (reg_mentioned_p (reg, pattern))
	{
	    mask |= flag;
	}
    }
    bit_mask_stack[do_loop_nest_level++] = mask;
}

#ifdef INSN_MACHINE_INFO
const INSN_MACHINE_INFO insn_machine_info[];

#endif

int
 conflicting_address_regs_set_p (pattern)
rtx pattern;
{
    rtx reg;
    unsigned char mask = bit_mask_stack[--do_loop_nest_level];

    reg = gen_rtx (REG, Pmode, 6);

    for (; ADDR_REG_MASK & mask; mask >>= 1, ++REGNO (reg))
    {
	if ((1 & mask) && (addr_reg_set_p (reg, pattern)))
	{
	    return 1;
	}
    }
    return 0;
}

int
 target_mbtowc (pwc, s, n)
int *pwc;
char *s;
int n;
{
    if (NULL == s)
    {
	return 0;
    }

    if (0 == n)
    {
	return -1;
    }

    {
	char skinny = *(s++);

	if ('\0' == skinny)
	{
	    if (NULL != pwc)
	    {
		*pwc = 0;
	    }
	    return 0;
	}

	{
	    int skinnys_left;
	    int wide = 0;

	    n = (n < (CHAR_TYPE_SIZE / 8)) ? n : (CHAR_TYPE_SIZE / 8);

	    skinnys_left = n;

	    if (NULL == pwc)
	    {
		do
		{
		    skinny = *(s++);
		}
		while ((--skinnys_left) && ('\0' != skinny));
	    }
	    else
	    {
		do
		{
		    wide = (wide << 8) | skinny;
		    skinny = *(s++);
		}
		while ((--skinnys_left) && ('\0' != skinny));

		*pwc = wide;
	    }
	    return (int) (n - skinnys_left);
	}
    }
}

int
 target_mbstowcs (pwcs, s)
int *pwcs;
char *s;
{
    int converted;
    int i;

    for (i = 0;; i++)
    {
	/* check for invalid character */
	if ((converted = target_mbtowc (pwcs, s, CHAR_TYPE_SIZE / 8)) == -1)
	{
	    return ((int) -1);
	}

	/* check for NULL termination character */
	if (converted == 0)
	{
	    break;
	}

	/* bump pwcs and s */
	pwcs++;
	s += converted;
    }

    return i;
}

/* this global variable is used to cut down on the number of stupid org
   statements. when we see a counter pragma we set it, when we org we clear
   it. */
int counter_pragma_seen = 0;

/* process allowed pragmas. warn about unrecognized pragmas. */
int
 dsp_process_pragma_line (f)
FILE *f;
{
    /* read up to EOF or \n to get the pragma line. */
    char pragma_text[128];

    int retval,
     last = 0,
     bytes_left = sizeof (pragma_text) - 1;

    while (bytes_left--)
    {
	retval = getc (f);
	pragma_text[last] = retval;

	if (EOF == retval || '\n' == retval)
	{
	    pragma_text[last] = '\0';
	    break;
	}
	++last;
    }

    /* we got the string, now use strtok to get the parts. right now we
       use brute force to parse, maybe if we add a bunch of pragmas, we
       should use some CS technique. */

    {
	char **mod_spot,
	*default_val,
	*counter,
	*new_value;

#if ! defined( MIPSEL ) && ( defined( sun ) && defined( __GNUC__ ) )
	char delim[] = " \n\t\v\b\r";

#else
	char delim[7];

	strcpy (delim, " \n\t\v\b\r");
#endif

	/* get the counter */
	counter = strtok (pragma_text, delim);

	if (NULL == counter)
	{
	    goto fail;
	}

#if defined( HAVE_L_MEMORY )
	if (0 == strcmp ("y_run", counter))
	{
	    mod_spot = &y_run;
	    default_val = y_run_default;
	}
	else if (0 == strcmp ("y_load", counter))
	{
	    mod_spot = &y_load;
	    default_val = y_load_default;
	}
	else if (0 == strcmp ("l_run", counter))
	{
	    mod_spot = &l_run;
	    default_val = l_run_default;
	}
	else if (0 == strcmp ("l_load", counter))
	{
	    mod_spot = &l_load;
	    default_val = l_load_default;
	}
#endif
	else if (0 == strcmp ("x_run", counter))
	{
	    mod_spot = &x_run;
	    default_val = x_run_default;
	}
	else if (0 == strcmp ("x_load", counter))
	{
	    mod_spot = &x_load;
	    default_val = x_load_default;
	}
	else if (0 == strcmp ("p_run", counter))
	{
	    mod_spot = &p_run;
	    default_val = p_run_default;
	}
	else if (0 == strcmp ("p_load", counter))
	{
	    mod_spot = &p_load;
	    default_val = p_load_default;
	}
#if defined( TARGET_PID )
	else if (0 == strcmp ("rti_run", counter))
	{
	    mod_spot = &rti_run;
	    default_val = rti_run_default;
	}
	else if (0 == strcmp ("rti_load", counter))
	{
	    mod_spot = &rti_load;
	    default_val = rti_load_default;
	}
#endif
	else
	{
	    goto fail;
	}

	/* get the new value for orgs */
	new_value = strtok (NULL, delim);

	if (NULL == new_value)
	{
	    /* switch back to the default_val. */
	    if (safe_strcmp (*mod_spot, default_val))
	    {
		counter_pragma_seen = 1;
	    }
	    *mod_spot = default_val;
	}
	else
	{
	    /* put the new counter into effect. */
	    if (safe_strcmp (*mod_spot, new_value))
	    {
		counter_pragma_seen = 1;
	    }
	    *mod_spot = strcpy
		((char *) xmalloc (strlen (new_value) + 1), new_value);
	}
    }
    return retval;

  fail:
    warning ("invalid #pragma string has been ignored");

    return retval;
}

/* all org asm statements are sprintf'ed into this array */
static char org_text[BUFSIZ];

/* create the proper org statement for data. */
char *
 text_section_asm_op ()
{
    /* org statements are assumed to be of the form blah:blech. the blech is
       optional. if we switch back and forth betweem program and data, we 
       don't want to org multiple sections of code at blech, but we still want
       to use the counter blah. we waste everything after the colon after
       the full string is printed once. */
    char *colon;

    if (NULL == p_load)
    {
	sprintf (org_text, "\torg\t%s", p_run);
    }
    else
    {
	sprintf (org_text, "\torg\t%s,%s", p_run, p_load);
	if (p_load_default != p_load)
	{
	    colon = strchr (p_load, ':');
	    if (NULL != colon)
	    {
		*(colon + 1) = '\0';
	    }
	}
    }
    if (p_run_default != p_run)
    {
	colon = strchr (p_run, ':');
	if (NULL != colon)
	{
	    *(colon + 1) = '\0';
	}
    }

    return org_text;
}

/* create the proper org statement for data. */
char *
 data_section_asm_op ()
{
    /* org statements are assumed to be of the form blah:blech. the blech is
       optional. if we switch back and forth betweem program and data, we 
       don't want to org multiple sections of code at blech, but we still want
       to use the counter blah. we waste everything after the colon after
       the full string is printed once. */
    char *colon,
    *the_load,
    *the_run,
    *the_default_load,
    *the_default_run;

#if defined( DSP56100 )
    the_load = x_load;
    the_run = x_run;
    the_default_load = x_load_default;
    the_default_run = x_run_default;

    if (NULL == x_load)
    {
	sprintf (org_text, "\torg\t%s", x_run);
    }
    else
    {
	sprintf (org_text, "\torg\t%s,%s", x_run, x_load);
    }
#else
    switch (memory_model)
    {
    case 'x':
	the_load = x_load;
	the_run = x_run;
	the_default_load = x_load_default;
	the_default_run = x_run_default;

	if (NULL == x_load)
	{
	    sprintf (org_text, "\torg\t%s", x_run);
	}
	else
	{
	    sprintf (org_text, "\torg\t%s,%s", x_run, x_load);
	}
	break;

    case 'y':
	the_load = y_load;
	the_run = y_run;
	the_default_load = y_load_default;
	the_default_run = y_run_default;

	if (NULL == y_load)
	{
	    sprintf (org_text, "\torg\t%s", y_run);
	}
	else
	{
	    sprintf (org_text, "\torg\t%s,%s", y_run, y_load);
	}
	break;

    case 'l':
	the_load = l_load;
	the_run = l_run;
	the_default_load = l_load_default;
	the_default_run = l_run_default;

	if (NULL == l_load)
	{
	    sprintf (org_text, "\torg\t%s", l_run);
	}
	else
	{
	    sprintf (org_text, "\torg\t%s,%s", l_run, l_load);
	}
	break;

    default:
	abort ();
    }
#endif

    if (NULL != the_load && the_load != the_default_load)
    {
	colon = strchr (the_load, ':');
	if (NULL != colon)
	{
	    *(colon + 1) = '\0';
	}
    }
    if (the_run != the_default_run)
    {
	colon = strchr (the_run, ':');
	if (NULL != colon)
	{
	    *(colon + 1) = '\0';
	}
    }

    return org_text;
}

#if defined( TARGET_PID )
char *
 rti_section_asm_op ()
{
    /* org statements are assumed to be of the form blah:blech. the blech is
       optional. if we switch back and forth betweem program and data, we 
       don't want to org multiple sections of code at blech, but we still want
       to use the counter blah. we waste everything after the colon after
       the full string is printed once. */
    char *colon;

    if (NULL == rti_load)
    {
	sprintf (org_text, "\torg\t%s", rti_run);
    }
    else
    {
	sprintf (org_text, "\torg\t%s,%s", rti_run, rti_load);
	if (rti_load_default != rti_load)
	{
	    colon = strchr (rti_load, ':');
	    if (NULL != colon)
	    {
		*(colon + 1) = '\0';
	    }
	}
    }
    if (rti_run_default != rti_run)
    {
	colon = strchr (rti_run, ':');
	if (NULL != colon)
	{
	    *(colon + 1) = '\0';
	}
    }

    return org_text;
}
#endif

static struct counter_bind
{
    char *x_run,
    *x_load,
    *y_run,
    *y_load,
    *l_run,
    *l_load,
    *p_run,
    *p_load;
    struct counter_bind *link;
}
*binding_stack = NULL;

/* save the current state of the run-time and load-time counters. change the
   counters to the values associated with the given declaration. */

void
 push_counters (decl)
tree decl;
{
    char *new_x_run,
    *new_x_load;

#if defined( HAVE_L_MEMORY )
    char *new_y_run,
    *new_y_load;
    char *new_l_run,
    *new_l_load;

#endif
    char *new_p_run,
    *new_p_load;

    struct counter_bind *new = (struct counter_bind *)
     xmalloc (sizeof (struct counter_bind));

    new->x_run = x_run, new->x_load = x_load;
#if defined( HAVE_L_MEMORY )
    new->y_run = y_run, new->y_load = y_load;
    new->l_run = l_run, new->l_load = l_load;
#endif
    new->p_run = p_run, new->p_load = p_load;

    new->link = binding_stack;
    binding_stack = new;

    /* run time counters should never be NULL, load time counters sometimes
       are. */
    new_x_run = (NULL == DECL_X_RUN (decl)) ? x_run : DECL_X_RUN (decl);
    new_x_load = DECL_X_LOAD (decl);
#if defined( HAVE_L_MEMORY )
    new_y_run = (NULL == DECL_Y_RUN (decl)) ? y_run : DECL_Y_RUN (decl);
    new_y_load = DECL_Y_LOAD (decl);
    new_l_run = (NULL == DECL_L_RUN (decl)) ? l_run : DECL_L_RUN (decl);
    new_l_load = DECL_L_LOAD (decl);
#endif
    new_p_run = (NULL == DECL_P_RUN (decl)) ? p_run : DECL_P_RUN (decl);
    new_p_load = DECL_P_LOAD (decl);

    if (safe_strcmp (new_x_run, x_run) ||
#if defined( HAVE_L_MEMORY )
	safe_strcmp (new_y_run, y_run) ||
	safe_strcmp (new_l_run, l_run) ||
#endif
	safe_strcmp (new_p_run, p_run) ||
	safe_strcmp (new_x_load, x_load) ||
#if defined( HAVE_L_MEMORY )
	safe_strcmp (new_y_load, y_load) ||
	safe_strcmp (new_l_load, l_load) ||
#endif
	safe_strcmp (new_p_load, p_load))
    {
	counter_pragma_seen = 1;

	x_run = new_x_run;
	x_load = new_x_load;
#if defined( HAVE_L_MEMORY )
	y_run = new_y_run;
	y_load = new_y_load;
	l_run = new_l_run;
	l_load = new_l_load;
#endif
	p_run = new_p_run;
	p_load = new_p_load;
    }
}

/* restore the counters to the state before push_counters was last called. */

void
 pop_counters ()
{
    struct counter_bind *dead = binding_stack;

    if (safe_strcmp (x_run, dead->x_run) ||
	safe_strcmp (x_load, dead->x_load) ||
#if defined( HAVE_L_MEMORY )
	safe_strcmp (y_run, dead->y_run) ||
	safe_strcmp (y_load, dead->y_load) ||
	safe_strcmp (l_run, dead->l_run) ||
	safe_strcmp (l_load, dead->l_load) ||
#endif
	safe_strcmp (p_run, dead->p_run) ||
	safe_strcmp (p_load, dead->p_load))
    {
	counter_pragma_seen = 1;

	x_run = dead->x_run, x_load = dead->x_load;
#if defined( HAVE_L_MEMORY )
	y_run = dead->y_run, y_load = dead->y_load;
	l_run = dead->l_run, l_load = dead->l_load;
#endif
	p_run = dead->p_run, p_load = dead->p_load;
    }
    binding_stack = binding_stack->link;
    free (dead);
}

static int
 safe_strcmp (s1, s2)
char *s1,
*s2;
{
    if (NULL == s1)
    {
	if (NULL == s2)
	{
	    return 0;
	}
	return 1;
    }
    if (NULL == s2)
    {
	return 1;
    }
    return strcmp (s1, s2);
}

#if defined( HAVE_L_MEMORY )

char *insn_template[];

/* replace all of the '@'s in the insn templates with the actual memory 
   space specification character. */

void
 harden_mem_space ()
{
    int i;

    for (i = 0; i < MAX_INSN_CODE; ++i)
    {
	if (NULL != insn_template[i] &&
	    NULL != strchr (insn_template[i], '@'))
	{
	    char *new = (char *) xmalloc (strlen (insn_template[i]) + 1);
	    char *point = new;

	    (void) strcpy (new, insn_template[i]);

	    point = strchr (point, '@');
	    do
	    {
		*point = mem_space;
		point = strchr (point, '@');
	    }
	    while (NULL != point);

	    insn_template[i] = new;
	}
    }
}

#endif

#if defined( TARGET_PID )
/* this code supports fixing data structure addresses at run-time. */

/* return 1 if the passed rtl expression references the DATA_POINTER,
   0 otherwise. */
int
 expr_refs_data_pointer_p (x)
rtx x;
{
    if (REG_P (x) && DATA_POINTER_REGNUM == REGNO (x))
    {
	return 1;
    }
    else
    {
	int i,
	 j;
	char *format_ptr = GET_RTX_FORMAT (GET_CODE (x));

	for (i = 0; i < GET_RTX_LENGTH (GET_CODE (x)); ++i)
	{
	    switch (*format_ptr++)
	    {
	    case 'e':
		if (expr_refs_data_pointer_p (XEXP (x, i)))
		{
		    return 1;
		}
		break;

	    case 'E':
		if (XVEC (x, i) != NULL)
		{
		    for (j = 0; j < XVECLEN (x, i); ++j)
		    {
			if (expr_refs_data_pointer_p (XVECEXP (x, i, j)))
			{
			    return 1;
			}
		    }
		}
		break;
	    }
	}
	return 0;
    }
}

/* return a copy of the passed rtl expression that no longer references
   the DATA_POINTER. No part of the new expression is part of the passed
   expression. note that the DATA_POINTER is only used as one of the two 
   operands. also the DATA_POINTER will appear only once in the
   expression. */

rtx
copy_and_remove_data_pointer_ref (x)
rtx x;
{
    /* eliminate those nasty core droppings. */

    if (NULL == x)
    {
	return NULL;
    }

    /* if we've got a PLUS with the DATA_POINTER as one of its operands,
       then return the other operand. */

    if (PLUS == GET_CODE (x))
    {
	rtx x0 = XEXP (x, 0),
	 x1 = XEXP (x, 1);

	if (REG_P (x0) && DATA_POINTER_REGNUM == REGNO (x0))
	{
	    return copy_rtx (x1);
	}
	else if (REG_P (x1) && DATA_POINTER_REGNUM == REGNO (x1))
	{
	    return copy_rtx (x0);
	}
    }

    /* otherwise, duplicate the current node and run this function
       on each of the subexpressions. this code was boosted from
       copy_rtx ( ) in rtl.c. */

    {
	rtx copy;
	int i,
	 j;
	RTX_CODE code = GET_CODE (x);
	char *format_ptr;

	switch (code)
	{
	case REG:
	case QUEUED:
	case CONST_INT:
	case CONST_DOUBLE:
	case SYMBOL_REF:
	case CODE_LABEL:
	case PC:
	case CC0:
	    return x;
	}

	copy = rtx_alloc (code);
	PUT_MODE (copy, GET_MODE (x));
	copy->in_struct = x->in_struct;
	copy->volatil = x->volatil;
	copy->unchanging = x->unchanging;
	copy->integrated = x->integrated;
#if defined( HAVE_FRACTIONAL_ACCUMULATORS )
	/* added fields for DSP enhancements. */
	copy->reload_subreg = x->reload_subreg;
#endif
#if defined( MOTOROLA_DSP )
	copy->reload_generated = x->reload_generated;
#endif
	format_ptr = GET_RTX_FORMAT (code);

	for (i = 0; i < GET_RTX_LENGTH (code); ++i)
	{
	    switch (*format_ptr++)
	    {
	    case 'e':
		XEXP (copy, i) =
		    copy_and_remove_data_pointer_ref (XEXP (x, i));
		break;

	    case 'E':
		XVEC (copy, i) = XVEC (x, i);
		if (XVEC (x, i) != NULL)
		{
		    XVEC (copy, i) = rtvec_alloc (XVECLEN (x, i));
		    for (j = 0; j < XVECLEN (copy, i); ++j)
		    {
			XVECEXP (copy, i, j) =
			    copy_and_remove_data_pointer_ref
			    (XVECEXP (x, i, j));
		    }
		}
		break;

	    default:
		XINT (copy, i) = XINT (x, i);
		break;
	    }
	}
	return copy;
    }
}
#endif

int
 go_if_legitimate_address (mode, x, strict)
int mode;
rtx x;
int strict;
{
    /* constant pointer */

    if (CONSTANT_ADDRESS_P (x))
    {
	return 1;
    }

    /* note that REG_OK_STRICT may NOT be defined here! */
    /* register pointer */

    if (REG_P (x) &&
	((strict) ?
	 (REGNO_OK_FOR_INDEX_P (REGNO (x))) :
	 (REG_OK_FOR_INDEX_P (x))))
    {
	return 1;
    }

    /* register arithmetic */

    switch (GET_CODE (x))
    {
    case SUBREG:
	return go_if_legitimate_address (mode, XEXP (x, 0), strict);

    case POST_DEC:
    case POST_INC:
    case PRE_INC:
    case PRE_DEC:
	if (REG_P (XEXP (x, 0)) &&
	    ((strict) ?
	     (REGNO_OK_FOR_INDEX_P (REGNO (XEXP (x, 0)))) :
	     (REG_OK_FOR_INDEX_P (XEXP (x, 0)))))
	{
	    return 1;
	}
	break;

    case PLUS:
	if (REG_P (XEXP (x, 0)) &&
	    (Pmode == GET_MODE (XEXP (x, 0))) &&
	    (CONSTANT_ADDRESS_P (XEXP (x, 1))) &&
	    ((strict) ?
	     (REGNO_OK_FOR_INDEX_P (REGNO (XEXP (x, 0)))) :
	     (REG_OK_FOR_INDEX_P (XEXP (x, 0)))))
	{
	    return 1;
	}

	if (REG_P (XEXP (x, 1)) &&
	    (Pmode == GET_MODE (XEXP (x, 1))) &&
	    (CONSTANT_ADDRESS_P (XEXP (x, 0))) &&
	    ((strict) ?
	     (REGNO_OK_FOR_INDEX_P (REGNO (XEXP (x, 1)))) :
	     (REG_OK_FOR_INDEX_P (XEXP (x, 1)))))
	{
	    return 1;
	}
    }
    return 0;
}

#if defined( HAVE_FRACTIONAL_ACCUMULATORS )
rtx
next_cc_relevancy (insn)
rtx insn;
{
    rtx peek = NEXT_INSN (insn);

    for (; NULL != peek; peek = NEXT_INSN (peek))
    {
	switch (GET_CODE (peek))
	{
	case JUMP_INSN:
	case CALL_INSN:
	case CODE_LABEL:
	case BARRIER:
	    return peek;

	case INSN:
	    if ((PARALLEL == GET_CODE (PATTERN (peek))) &&
		(SET == GET_CODE (XVECEXP (PATTERN (peek), 0, 0))) &&
		(0 == strcmp (GET_RTX_FORMAT (GET_CODE (XEXP (XVECEXP (PATTERN (peek), 0, 0), 1))), "ee")) &&
	    (cc0_rtx == XEXP (XEXP (XVECEXP (PATTERN (peek), 0, 0), 1), 0)))
	    {
		return peek;
	    }
	}
    }
    return insn;
}

enum mdep_cc_info
 next_cc_use (insn)
rtx insn;
{
    enum rtx_code jump_code;
    rtx peek = NEXT_INSN (insn);

    for (; NULL != peek; peek = NEXT_INSN (peek))
    {
	if (JUMP_INSN == GET_CODE (peek))
	{
	    break;
	}
	if (CALL_INSN == GET_CODE (peek) ||
	    CODE_LABEL == GET_CODE (peek) ||
	    BARRIER == GET_CODE (peek))
	{
	    return CC_UNKNOWN;
	}
	if (INSN == GET_CODE (peek) &&
	    PARALLEL == GET_CODE (PATTERN (peek)) &&
	    SET == GET_CODE (XVECEXP (PATTERN (peek), 0, 0)) &&
	    0 == strcmp (GET_RTX_FORMAT (GET_CODE (XEXP (XVECEXP (PATTERN (peek), 0, 0), 1))), "ee") &&
	    (cc0_rtx == XEXP (XEXP (XVECEXP (PATTERN (peek), 0, 0), 1), 0) ||
	     cc0_rtx == XEXP (XEXP (XVECEXP (PATTERN (peek), 0, 0), 1), 1)))
	{
	    jump_code = GET_CODE (XEXP (XVECEXP (PATTERN (peek), 0, 0), 1));

	    if (GEU == jump_code || GTU == jump_code || LEU == jump_code
		|| LTU == jump_code)
	    {
		return CC_UNSIGNED;
	    }
	    else
	    {
		return CC_SIGNED;
	    }
	}
    }

    if ((NULL == peek) ||
	(JUMP_INSN != GET_CODE (peek)) ||
	(SET != GET_CODE (PATTERN (peek))) ||
	(IF_THEN_ELSE != GET_CODE (XEXP (PATTERN (peek), 1))))
    {
	return CC_UNKNOWN;
    }
    jump_code = GET_CODE (XEXP (XEXP (PATTERN (peek), 1), 0));

    if (GEU == jump_code || GTU == jump_code || LEU == jump_code
	|| LTU == jump_code || EQU == jump_code || NEU == jump_code)
    {
	return CC_UNSIGNED;
    }
    else
    {
	return CC_SIGNED;
    }
}
#endif

#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56100 ) || \
    defined( DSP96000 ) || defined( DSP56600 )
#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
void
 pushregs (file)
FILE *file;
{
    int reg_index;

    if (TARGET_CONSERVE_P_MEM)
    {
	/* function logue will take care of it. */
	return;
    }

    FirstSave = -1;

    for (reg_index = 0; reg_index < FIRST_PSEUDO_REGISTER; ++reg_index)
    {
	if (regs_ever_live[reg_index] && !call_used_regs[reg_index])
	{
	    if (0 > FirstSave)
	    {
		FirstSave = reg_index;
	    }

	    if (DST_REGS == REGNO_REG_CLASS (reg_index))
	    {
		if ('l' == memory_model)
		{
		    fprintf (file, "\
	move	%c10,l:(r6)+\n", *reg_names[reg_index]);
		}
		else
		{
		    fprintf (file, "\
	move	%s0,%c:(r6)+\n\
	move	%s1,%c:(r6)+\n", reg_names[reg_index], mem_space,
			     reg_names[reg_index], mem_space);
		}

	    }
	    else if (SRC_REGS == REGNO_REG_CLASS (reg_index) ||
		     MPY_REGS == REGNO_REG_CLASS (reg_index))
	    {
		if ('l' == memory_model)
		{
		    /* try to save x or y as a pair. */
		    if (reg_index & 1)
		    {
			/* x1 or y1 */
			if (!(regs_ever_live[reg_index - 1] &&
			      !call_used_regs[reg_index - 1]))
			{
			    /* was the 0 portion saved ? if so, then we
			       don't need to do anything, else save the 1 ! */

			    fprintf (file, "\
	move	%s,%c:(r6)+\n", reg_names[reg_index], mem_space);
			}
		    }
		    else
		    {
			/* x0 or y0 */
			if (regs_ever_live[reg_index + 1] &&
			    !call_used_regs[reg_index + 1])
			{
			    /* are we also saving the 1 portion ? if so,
			       save the 1 as well as the 0 with an l: move.
			       otherwise, just save the 0. */
			    fprintf (file, "\
	move	%c,l:(r6)+\n", *reg_names[reg_index]);
			}
			else
			{
			    fprintf (file, "\
	move	%s,%c:(r6)+\n", reg_names[reg_index], mem_space);
			}
		    }
		}
		else
		{
		    fprintf (file, "\
	move	%s,%c:(r6)+\n", reg_names[reg_index], mem_space);
		}
	    }
	    else
	    {
		fprintf (file, "\
	move	%s,%c:(r6)+\n", reg_names[reg_index], mem_space);
	    }

	}
    }
}
#elif defined( DSP56100 )
void
 pushregs (file)
FILE *file;
{
    int reg_index;

    if (TARGET_CONSERVE_P_MEM)
    {
	/* the prolog will handle this. */
	return;
    }

    FirstSave = -1;

    for (reg_index = 0; reg_index < FIRST_PSEUDO_REGISTER; ++reg_index)
    {
	if (regs_ever_live[reg_index] &&
	    !call_used_regs[reg_index])
	{
	    if (0 > FirstSave)
	    {
		FirstSave = reg_index;
	    }

	    if (DST_REGS == REGNO_REG_CLASS (reg_index))
	    {
		fprintf (file, "\
	move	%s0,x:(r2)+\n\
	move	%s1,x:(r2)+\n", reg_names[reg_index], reg_names[reg_index]);
	    }
	    else
	    {
		fprintf (file, "\
	move	%s,x:(r2)+\n", reg_names[reg_index]);
	    }

	}
    }
}

#endif

#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
void
 popregs (file)
FILE *file;
{
    int reg_index = FIRST_PSEUDO_REGISTER;
    int first_save = 0;

#define SAVE_CHECK \
{\
    if ( ! first_save )\
    {\
	first_save = 1;\
	fprintf ( file, "\tmove	(r6)-\n" );\
    }\
}

    if (TARGET_CONSERVE_P_MEM)
    {
	/* the epilog will handle this. */
	return;
    }

    while (--reg_index >= 0)
    {
	if (regs_ever_live[reg_index] &&
	    !call_used_regs[reg_index])
	{
	    SAVE_CHECK;

	    if (DST_REGS == REGNO_REG_CLASS (reg_index))
	    {
		if (reg_index == FirstSave)
		{
		    if ('l' == memory_model)
		    {
			fprintf (file, "\
	move	l:(r6),%c\n", *reg_names[reg_index]);
		    }
		    else
		    {
			fprintf (file, "\
	move	%c:(r6)-,%s\n\
	move	%c:(r6),%s0\n", mem_space, reg_names[reg_index],
				 mem_space, reg_names[reg_index]);
		    }
		}
		else
		{
		    if ('l' == memory_model)
		    {
			fprintf (file, "\
	move	l:(r6)-,%c\n", *reg_names[reg_index]);
		    }
		    else
		    {
			fprintf (file, "\
	move	%c:(r6)-,%s\n\
	move	%c:(r6)-,%s0\n", mem_space, reg_names[reg_index],
				 mem_space, reg_names[reg_index]);
		    }
		}
	    }
	    else if (SRC_REGS == REGNO_REG_CLASS (reg_index) ||
		     MPY_REGS == REGNO_REG_CLASS (reg_index))
	    {
		if ('l' == memory_model)
		{
		    if (reg_index & 1)
		    {
			/* x1 or y1 */
			if (!(regs_ever_live[reg_index - 1] &&
			      !call_used_regs[reg_index - 1]))
			{
			    /* if the 0 portion wasn't saved we need to
			       restore the 1, otherwise do nothing. */

			    if (reg_index == FirstSave)
			    {
				fprintf (file, "\
	move	%c:(r6),%s\n", mem_space, reg_names[reg_index]);
			    }
			    else
			    {
				fprintf (file, "\
	move	%c:(r6)-,%s\n", mem_space, reg_names[reg_index]);
			    }
			}
		    }
		    else
		    {
			/* x0 or y0 */
			if (regs_ever_live[reg_index + 1] &&
			    !call_used_regs[reg_index + 1])
			{
			    /* if the 1 portion was also saved, then we
			       need to to do an l: restore, otherwise a y:. */

			    if (reg_index == FirstSave)
			    {
				fprintf (file, "\
	move	l:(r6),%c\n", *reg_names[reg_index]);
			    }
			    else
			    {
				fprintf (file, "\
	move	l:(r6)-,%c\n", *reg_names[reg_index]);
			    }
			}
			else
			{
			    if (reg_index == FirstSave)
			    {
				fprintf (file, "\
	move	%c:(r6),%s\n", mem_space, reg_names[reg_index]);
			    }
			    else
			    {
				fprintf (file, "\
	move	%c:(r6)-,%s\n", mem_space, reg_names[reg_index]);
			    }
			}
		    }
		}
		else
		{
		    if (reg_index == FirstSave)
		    {
			fprintf (file, "\
	move	%c:(r6),%s\n", mem_space, reg_names[reg_index]);
		    }
		    else
		    {
			fprintf (file, "\
	move	%c:(r6)-,%s\n", mem_space, reg_names[reg_index]);
		    }
		}
	    }
	    else
	    {
		if (reg_index == FirstSave)
		{
		    fprintf (file, "\
	move	%c:(r6),%s\n", mem_space, reg_names[reg_index]);
		}
		else
		{
		    fprintf (file, "\
	move	%c:(r6)-,%s\n", mem_space, reg_names[reg_index]);
		}
	    }
	}
    }
}

#elif defined( DSP56100 )
void
 popregs (file)
FILE *file;
{
    int reg_index = FIRST_PSEUDO_REGISTER;
    int first_save = 0;

#define SAVE_CHECK \
{\
    if ( ! first_save )\
    {\
	first_save = 1;\
	fprintf ( file, "\tmove	(r2)-\n" );\
    }\
}

    if (TARGET_CONSERVE_P_MEM)
    {
	/* function logue will handle it. */
	return;
    }

    while (--reg_index >= 0)
    {
	if (regs_ever_live[reg_index] && !call_used_regs[reg_index])
	{
	    SAVE_CHECK;

	    if (DST_REGS == REGNO_REG_CLASS (reg_index))
	    {
		if (reg_index == FirstSave)
		{
		    fprintf (file, "\
	move	x:(r2)-,%s\n\
	move	x:(r2),%s0\n", reg_names[reg_index], reg_names[reg_index]);
		}
		else
		{
		    fprintf (file, "\
	move	x:(r2)-,%s\n\
	move	x:(r2)-,%s0\n", reg_names[reg_index], reg_names[reg_index]);
		}
	    }
	    else
	    {
		if (reg_index == FirstSave)
		{
		    fprintf (file, "\
	move	x:(r2),%s\n", reg_names[reg_index]);
		}
		else
		{
		    fprintf (file, "\
	move	x:(r2)-,%s\n", reg_names[reg_index]);
		}
	    }
	}
    }
}

#endif

void
 print_operand_address (file, addr)
FILE *file;
rtx addr;
{
    rtx reg0;
    rtx reg1;
    void print_operand ();

    switch (GET_CODE (addr))
    {

    case REG:
	fprintf (file, "(%s)", reg_names[REGNO (addr)]);
	break;

    case POST_INC:
	fprintf (file, "(%s)+", reg_names[REGNO (XEXP (addr, 0))]);
	break;

    case POST_DEC:
	fprintf (file, "(%s)-", reg_names[REGNO (XEXP (addr, 0))]);
	break;

#if defined( DSP56600 ) || defined( DSP56300 ) || defined( DSP96000 )
    case PLUS:
	reg0 = XEXP (addr, 0);
	reg1 = XEXP (addr, 1);

	if (CONSTANT_ADDRESS_P (reg0))
	{
	    putc ('(', file);
	    print_operand (file, reg1, 'c');
	    putc ('+', file);
	    output_addr_const (file, reg0);
	    putc (')', file);
	    return;
	}

	if (CONSTANT_ADDRESS_P (reg1))
	{
	    putc ('(', file);
	    print_operand (file, reg0, 'c');
	    putc ('+', file);
	    output_addr_const (file, reg1);
	    putc (')', file);
	    return;
	}
#endif

    default:
	output_addr_const (file, addr);
    }
}

#if ! defined( DSP96000 )
void
 notice_update_cc (exp, insn)
rtx exp,
 insn;
{
    int code = recog_memoized (insn);
    enum info_spec flags = get_spec_bits (insn_machine_info[code].spec_bits);

    if (CALL_INSN == GET_CODE (insn))
    {
	cc_status.mdep = CC_SIGNED;
	cc_status.value1 = cc0_rtx;
	cc_status.value2 = gen_rtx (REG, VOIDmode, DSP_A_REGNUM);
    }
    else if (CCS_SET_CORRECTLY & flags)
    {
	/* the INSN alters the CCs. We assume that no PARALLELs set the
	   CCs correctly. */

	if (SET == GET_CODE (exp))
	{
	    cc_status.value1 = cc0_rtx;

	    if (CC0 == GET_CODE (SET_DEST (exp)))
	    {
		cc_status.mdep =
		    (UNSIGNED_COMPARE_P (insn)) ? CC_UNSIGNED : CC_SIGNED;

		cc_status.value2 = SET_SRC (exp);
	    }
	    else
	    {
		cc_status.mdep = CC_SIGNED;
		cc_status.value2 = SET_DEST (exp);
	    }
	}
    }
    else if (CCS_NOT_AFFECTED & flags)
    {
	/* the INSN leave the CCs alone. we do need to check that
	   the value that set the CCs isn't altered, though. */

	if (NULL != cc_status.value2 &&
	    ((REG == GET_CODE (cc_status.value2) &&
	      reg_mentioned_p (cc_status.value2, exp)) ||
	     (COMPARE == GET_CODE (cc_status.value2) &&
	      (reg_mentioned_p (XEXP (cc_status.value2, 0), exp) ||
	       reg_mentioned_p (XEXP (cc_status.value2, 1), exp)))))
	{
	    CC_STATUS_INIT;
	}
    }
    else
    {
	/* we're not sure how the CCs make it through. */
	CC_STATUS_INIT;
    }
    added_flags = NO_FLAGS;
}
#else
void
 notice_update_cc (Exp, Insn)
rtx Exp;
rtx Insn;
{
    if (SET == GET_CODE (Exp))
    {
	switch (GET_CODE (SET_DEST (Exp)))
	{

	case CC0:
	    cc_status.flags = 0;
	    cc_status.value1 = SET_DEST (Exp);
	    cc_status.value2 = SET_SRC (Exp);
	    break;

	case REG:
	    switch (GET_CODE (SET_SRC (Exp)))
	    {
	    case CALL:
		if (SImode == GET_MODE (SET_DEST (Exp)))
		{
		    cc_status.flags = 0;
		    cc_status.value1 = SET_DEST (Exp);
		    cc_status.value2 = SET_SRC (Exp);
		}
		else
		{
		    CC_STATUS_INIT;
		}
		break;

	    case MEM:
	    case REG:
	    case EQ:
	    case NE:
	    case GT:
	    case GTU:
	    case GE:
	    case GEU:
	    case LT:
	    case LTU:
	    case LE:
	    case LEU:
		CC_STATUS_INIT;
		break;

	    default:
		CC_STATUS_INIT;
		break;
	    }

	    break;

	case MEM:
	    CC_STATUS_INIT;
	    break;
	}
    }
}
#endif

/* hard_regno_mode_ok -
 * this global function implements the tm-m56k.h file macro of the same name.
 * a 1 is returned if regno can be used to hold a value of type mode.
 */

int
 hard_regno_mode_ok (regno, mode)
int regno;
int mode;
{
    switch (mode)
    {
    case VOIDmode:
	return 1;

    case SFmode:
    case DFmode:
    case DImode:
#if defined( DSP56000 ) || defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	/* for doubles, floats, and longs, we need 48-bit registers. */
	if (DSP_X0_REGNUM == regno || DSP_Y0_REGNUM == regno ||
	    DSP_A_REGNUM == regno || DSP_B_REGNUM == regno)
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
#elif defined( DSP96000 )
	return (DSP_D0L_REGNUM <= regno && DSP_D7L_REGNUM >= regno &&
		(0 == regno % 3));
#endif
	break;

#if defined( DSP56000 )

    case PSImode:
	/* for pointers we can use any register. */
	return 1;

    default:
	/* for all other cases, we assume that a 24-bit reg is needed. */
	if (regno < DSP_R0_REGNUM)
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
	break;

#elif defined( DSP56600 ) || defined( DSP56300 ) || defined( DSP56100 ) || defined( DSP96000 )

    default:
	return 1;
#endif
    }
}

/* modes_tieable_p -
 * this global function implements the tm-m56k.h file macro of the same name.
 * a 1 is returned if both modes can always use the same pseudo register.
 */

int
 modes_tieable_p (mode1, mode2)
int mode1;
int mode2;
{
    if (mode1 == mode2)
    {
	return 1;
    }

    switch (mode1)
    {
    case SFmode:
    case DFmode:
    case DImode:
	if (SFmode == mode2 || DFmode == mode2 || DImode == mode2)
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
	break;
    }
    return 0;
}

/* regno_reg_class -
 * this global function implements the tm-m56k.h file macro of the same name.
 * the class number of the smallest class containing reg number regno is
 * returned.
 */

enum reg_class
 regno_reg_class (regno)
int regno;
{
#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
    switch (regno)
    {
    case DSP_X0_REGNUM:
    case DSP_Y0_REGNUM:
	return MPY_REGS;

    case DSP_X1_REGNUM:
    case DSP_Y1_REGNUM:
	return SRC_REGS;

    case DSP_A_REGNUM:
    case DSP_B_REGNUM:
	return DST_REGS;

    default:
#if defined( DSP56300 ) || defined( DSP56600 )
	if (regno <= LAST_ADDRESS_REGISTER)
	{
	    return ADDR_REGS;
	}
	else
	{
	    return INDX_REGS;
	}
#elif defined( DSP56000 )
	return ADDR_REGS;
#endif
    }
#elif defined( DSP56100 )
    switch (regno)
    {
    case DSP_X1_REGNUM:
	return MPY1_REGS;

    case DSP_Y1_REGNUM:
	return MPY2_REGS;

    case DSP_X0_REGNUM:
	return SHF1_REGS;

    case DSP_Y0_REGNUM:
	return Y_DBL_REGS;

    case DSP_A_REGNUM:
    case DSP_B_REGNUM:
	return DST_REGS;

    default:
	return ADDR_REGS;
    }
#elif defined( DSP96000 )
    /* this sucks. rewrite it. */
    return ((24 <= regno) ? ((32 > regno) ? ADDR_REGS : INDX_REGS) : ((0 == regno % 3) ? LO_DATA_REGS : ((1 == regno % 3) ? DATA_OR_ADDR_REGS : HI_DATA_REGS)));
#endif
}

/* reg_class_from_letter -
 * this global function implements the tm-m56k.h file macro of the same name.
 * the class number of the given letter is returned.
 */

enum reg_class
 reg_class_from_letter (c)
char c;
{
#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
    switch (c)
    {
    case 'S':
	return SRC_REGS;
    case 'D':
	return DST_REGS;
    case 'A':
	return ADDR_REGS;
    case 'R':
	return MPY_REGS;

#if defined( DSP56300 ) || defined( DSP56600 )
    case 'N':
	return INDX_REGS;
#endif

    case 'r':
	return ALL_REGS;

    default:
	return NO_REGS;
    }
#elif defined( DSP56100 )
    switch (c)
    {
    case 'q':
	return SRC1_REGS;
    case 'D':
	return DST_REGS;
    case 'A':
	return ADDR_REGS;
    case 't':
	return MPY1_REGS;
    case 'T':
	return MPY2_REGS;
    case 'X':
	return SHF1_REGS;
    case 'Q':
	return SRC_REGS;
    case 'x':
	return X_DBL_REGS;
    case 'y':
	return Y_DBL_REGS;
    case 'R':
	return INC_REGS;

    default:
	return NO_REGS;
    }
#elif defined( DSP96000 )
    switch (c)
    {
    case 'd':
	return LO_DATA_REGS;
    case 'h':
	return HI_DATA_REGS;
    case 'a':
	return ADDR_REGS;
    case 'f':
	return FP_REGS;
    case 'n':
	return INDX_REGS;
    case 'D':
	return DI_REGS;

    default:
	return NO_REGS;
    }
#endif
}

int
 class_max_nregs (class, mode)
int class;
int mode;
{
#if defined( DSP96000 )

    if ((SFmode == mode || DFmode == mode) && FP_REGS == class)
    {
	return 3;
    }
    if (DImode == mode && DI_REGS == class)
    {
	return 2;
    }

#elif defined( DSP56000 ) || defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )

    if ((SFmode == mode || DFmode == mode || DImode == mode) &&
#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
	(SRC_REGS == class)
#elif defined( DSP56100 )
	(SRC1_REGS == class || MPY1_REGS == class || SRC_REGS == class ||
	 MPY2_REGS == class || X_DBL_REGS == class || Y_DBL_REGS == class ||
	 SHF1_REGS == class)
#endif
	)
    {
	return 2;
    }

#endif

    return 1;
}

#if defined( DSP56300 ) || defined( DSP56000 )
/* ---------------------------------------------------------------- *\

   name         dgetmant

   synopsis     dgetmant(dfptval, wordnum)
   double dfptval;
   int wordnum;

   description
   Inputs:
   dfptval = double precision value from which mantissa is extracted
   wordnum:  0 = upper 20 bits (leading 1 suppressed)
   1 = lower 32 bits

   return value 

   \* ---------------------------------------------------------------- */

long dgetmant (dfptval, wordnum)
double dfptval;
int wordnum;
{
    long *ivalptr;

    ivalptr = (long *) &dfptval;

#if defined( _MSDOS ) || defined( MIPSEL )
    if (wordnum == 0)
	return (((*(ivalptr + 1))) & 0x000fffffL);
    else			/* lower word */
	return (*(ivalptr));
#else
    if (wordnum == 0)
	return (((*ivalptr)) & 0x000fffff);
    else			/* lower word */
	return (*(ivalptr + 1));
#endif
}

long dgetexp (dfptval)
double dfptval;
{
    long *ivalptr;

    ivalptr = (long *) &dfptval;
#if defined( _MSDOS ) || defined( MIPSEL )
    ivalptr++;
#endif
    return (((*ivalptr) >> 20) & 0x07ff);	/* masks sign bit too */
}

void
 dtok (file, davey)
FILE *file;
double davey;
{
    union real_extract dval;
    unsigned long L_word,
     H_word,
     tmp,
     tmp2,
     dsgn;
    long increase_exponent;

    dval.d = davey;

    if (dval.d == 0)
    {
	H_word = L_word = 0;
    }
    else
    {
	L_word = dgetexp (dval.d);	/* ieee exponent */
	H_word = dgetmant (dval.d, 0);	/* grab upper 20 bits of double */
	tmp2 = dgetmant (dval.d, 1);	/* lower 32 bits of ieee mantissa */

	/* ieee denormalized number */
	if (L_word == 0)
	{
	    L_word++;		/* try it */

	    /* normalize it */
	    while (!(H_word & 0x100000))
	    {
		L_word--;	/* exponent decreases */
		H_word <<= 1;	/* shift entire 52 bit ieee mantissa left */

		if (tmp2 & 0x80000000l)
		{
		    H_word |= 1;
		}

		tmp2 <<= 1;
	    }
	}

	/* normalized ieee number with implied leading 1 */
	else
	{
	    H_word |= 0x00100000;	/* add leading 1 */
	}

	H_word <<= 10;
	tmp = ((tmp2 & 0xc0000000) >> 1);	/* grab lower 2 bits */
	tmp >>= 21;		/* ------------------- */
	H_word |= tmp;		/* combine upper and lower */
	increase_exponent = 0;

	/* round up for 1/2 lsb of kp mantissa */
	if (tmp2 & 0x20000000l)
	{
	    if (H_word >= 0x7fffff00)
	    {
		increase_exponent = 1;
		H_word += 0x100;
		H_word >>= 1;
	    }

	    else
	    {
		H_word += 0x100;
	    }
	}

#if defined( _MSDOS ) || defined( MIPSEL )
	dsgn = (*(long *) &(dval.i[1])) & 0x80000000l;
#else /* real computers */
	dsgn = (*(long *) &(dval.i[0])) & 0x80000000l;
#endif

	/* check for negative value in ieee format */
	if (dsgn)
	{
	    H_word = -H_word;
	}

	H_word >>= 8;
	L_word -= 1023;		/* IEEE double precision bias  */
	L_word += 8191;		/* kevFP little precision bias */
	L_word++;		/* lameFP s.xxx..xx */
	L_word += increase_exponent;	/* in case rounding modified exponent */

	/* negative weird case ( -0.5 == 800000  not c00000 ) */
	if ((H_word & 0x800000) && (H_word & 0x400000))
	{
	    H_word <<= 1;
	    L_word--;
	}
    }

    if (NULL != file)
    {
	if ('l' == memory_model)
	{
	    fprintf (file, "\tdc\t$%06x%06x\n", H_word & 0xffffff,
		     L_word & 0xffffff);
	}
	else
	{
	    fprintf (file, "\tdc\t$%06x\n\tdc\t$%06x\n", L_word & 0xffffff,
		     H_word & 0xffffff);
	}
    }

    adjusted_dval.i[1] = (0x00ffffff & H_word);
    adjusted_dval.i[0] = (0x00ffffff & L_word);
}
#endif

void
 simplify_cse_victims (first)
rtx first;
{
    rtx point,
     elem;

    while (first)
    {
	/* is it a 3 element PARALLEL ? */
	if ((INSN == GET_CODE (first)) &&
	    (point = PATTERN (first), PARALLEL == GET_CODE (point)) &&
	    (3 == XVECLEN (point, 0)) &&

	/* is the 1st element a ( SET ( REG ) ( REG / CONST_INT )) ? */
	    (elem = XVECEXP (point, 0, 0), SET == GET_CODE (elem)) &&
	    (REG == GET_CODE (SET_DEST (elem))) &&
	    ((REG == GET_CODE (SET_SRC (elem))) ||
	     (CONST == GET_CODE (SET_SRC (elem))) ||
	     (CONST_INT == GET_CODE (SET_SRC (elem)))) &&

	/* is the 2nd element a ( CLOBBER ( REG )) ? */
	    (elem = XVECEXP (point, 0, 1), CLOBBER == GET_CODE (elem)) &&
	    (REG == GET_CODE (XEXP (elem, 0))) &&

	/* is the 3rd element a ( CLOBBER ( REG )) ? */
	    (elem = XVECEXP (point, 0, 2), CLOBBER == GET_CODE (elem)) &&
	    (REG == GET_CODE (XEXP (elem, 0))))
	{
	    /* we've got a blown-out div template ! substitute with a simple
	       ( SET ( REG ) ( REG / CONST_INT )) (note: really the 1st 
	       element of the parallel). */

	    PATTERN (first) = XVECEXP (point, 0, 0);
	}
	first = NEXT_INSN (first);
    }
}

static int
 mpy_insn_p (insn)
rtx insn;
{
    return ((INSN == GET_CODE (insn)) &&
	    (SET == GET_CODE (PATTERN (insn))) &&
	    (REG == GET_CODE (SET_DEST (PATTERN (insn)))) &&
	    (MULT == GET_CODE (SET_SRC (PATTERN (insn)))));
}

static int
 mac_insn_p (insn)
rtx insn;
{
    /* mac +/- shape detector. */
    return ((INSN == GET_CODE (insn)) &&
	    (SET == GET_CODE (PATTERN (insn))) &&
	    (REG == GET_CODE (SET_DEST (PATTERN (insn)))) &&
	    ((PLUS == GET_CODE (SET_SRC (PATTERN (insn)))) ||
	     (MINUS == GET_CODE (SET_SRC (PATTERN (insn))))) &&
	    ((MULT == GET_CODE (XEXP (SET_SRC (PATTERN (insn)), 0))) ||
	     (MULT == GET_CODE (XEXP (SET_SRC (PATTERN (insn)), 1)))));
}

static rtx
 mac_addative_input (mac)
rtx mac;
{
    return (MULT == GET_CODE (XEXP (SET_SRC (PATTERN (mac)), 0))) ?
	XEXP (SET_SRC (PATTERN (mac)), 1) :
	XEXP (SET_SRC (PATTERN (mac)), 0);
}

#if defined( DSP56000 ) || defined( DSP56300 )|| defined( DSP56600 )
void
 reduce_norm_freq (first)
rtx first;
{
    /* point to the last insn.
       search up the list for each mac insn:
       detemine the addative_input.
       foreach insn in LOG_LINKS( point ):
       if insn SETs the addative_input, and insn is a mac or mpy, and
       there are no reload generated references to the operand
       between these two instructions, then:
       mark the pointed to mac as not needing input normalization
       mark the mpy/mac insn as not needing output normalization.

       the templates for mpy/mac will check for these flags. */

    rtx last = first,
     point;

    /* find the last insn, search up from it */
    while (NULL != NEXT_INSN (last))
    {
	last = NEXT_INSN (last);
    }

    while (last != first)
    {
	if (INSN == GET_CODE (last) &&
	    (get_spec_bits
	     (insn_machine_info[recog_memoized (last)].spec_bits) &
	     INPUT_NORM_OPTIONAL))
	{
	    rtx add_in = mac_addative_input (last);
	    rtx insn_chain = LOG_LINKS (last);

	    while (NULL != insn_chain)
	    {
		rtx this_insn = XEXP (insn_chain, 0);

		if (INSN == GET_CODE (this_insn) &&
		    (get_spec_bits
		     (insn_machine_info
		      [recog_memoized (this_insn)].spec_bits) &
		     OUTPUT_NORM_OPTIONAL) &&
		    add_in == SET_DEST (PATTERN (this_insn)) &&
		    !reg_used_between_p (add_in, this_insn, last))
		{
		    /* kludge up: re-use existing flags, in_struct and used, 
		       to inhibit the emission of normalization code at the
		       front and back end of mac (and mpy) instructions. */

		    /* turn off last's pre-norm. */
		    MEM_IN_STRUCT_P (last) = 1;

		    /* turn off this_insn's post-norm. */
		    RTX_UNCHANGING_P (this_insn) = 1;
		}
		insn_chain = XEXP (insn_chain, 1);
	    }
	}
	last = PREV_INSN (last);
    }
}
#endif

rtx
local_function_incoming_arg (cum, mode, type, named)
CUMULATIVE_ARGS cum;
enum machine_mode mode;
tree type;
int named;
{
    rtx arg = FUNCTION_ARG (cum, mode, type, named);

    if (NULL == arg || REG != GET_CODE (arg))
    {
	current_func_info |= FUNC_HAS_STACK_PARMS;
    }

    return arg;
}

#if ! defined( HAVE_UNSIGNED_BRANCH_CODES )
/* sort the compares into the signed and unsigned varieties. */
void
 sort_compares (insns)
rtx insns;
{
    while (insns)
    {
	if ((INSN == GET_CODE (insns)) &&
	    (SET == GET_CODE (PATTERN (insns))) &&
	    (cc0_rtx == SET_DEST (PATTERN (insns))))
	{
	    if (CC_UNSIGNED == next_cc_use (insns))
	    {
		/* note that this is an unsigned (slow) compare. we must
		   re-recog it in case this means we'll match a different
		   insn. making the insn code -1 ensures that recog will 
		   be re-run. */
		UNSIGNED_COMPARE_P (insns) = 1;
	    }
	    INSN_CODE (insns) = -1;
	}
	insns = NEXT_INSN (insns);
    }
}
#endif

/* the following is a lot of bullshit to ensure that constants in the 
   assembly output *in no way* ever assume that the assembler will discard
   the carry from arithmetic operations on constants. */

static rtx ensure_signed_int_offsets ();
static int find_symbol ();

static rtx
 ensure_integer_result_size (x)
rtx x;
{
    if (find_symbol (x) || CONST_INT == GET_CODE (x))
    {
	x = ensure_signed_int_offsets (x);
    }
    else
    {
	char *p,
	*fmt = GET_RTX_FORMAT (GET_CODE (x));

	p = fmt;

	while ('\0' != (*p))
	{
	    if ('e' == (*p))
	    {
		XEXP (x, p - fmt) =
		    ensure_integer_result_size (XEXP (x, p - fmt));
	    }
	    ++p;
	}
    }
    return x;
}

static rtx
 ensure_signed_int_offsets (x)
rtx x;
{
    if (CONST_INT == GET_CODE (x))
    {
	x = gen_rtx (CONST_INT, VOIDmode,
		     ((INTVAL (x) << (HOST_BITS_PER_INT - BITS_PER_UNIT))
		      >> (HOST_BITS_PER_INT - BITS_PER_UNIT)));
    }
    else
    {
	char *p,
	*fmt = GET_RTX_FORMAT (GET_CODE (x));

	p = fmt;

	while ('\0' != (*p))
	{
	    if ('e' == (*p))
	    {
		XEXP (x, p - fmt) =
		    ensure_signed_int_offsets (XEXP (x, p - fmt));
	    }
	    ++p;
	}
    }
    return x;
}

static int
 find_symbol (x)
rtx x;
{
    int found_symbol = 0;
    char *p,
    *fmt = GET_RTX_FORMAT (GET_CODE (x));

    if (SYMBOL_REF == GET_CODE (x))
    {
	return 1;
    }

    p = fmt;

    while ('\0' != (*p))
    {
	if ('e' == (*p))
	{
	    found_symbol |= find_symbol (XEXP (x, p - fmt));
	}
	++p;
    }
    return found_symbol;
}

#if defined( DSP56000 )
static rtx
 ensure_signed_pointer_offsets (x)
rtx x;
{
    if (CONST_INT == GET_CODE (x))
    {
	x = gen_rtx (CONST_INT, VOIDmode,
		     ((INTVAL (x) << (HOST_BITS_PER_INT - POINTER_SIZE))
		      >> (HOST_BITS_PER_INT - POINTER_SIZE)));
    }
    else
    {
	char *p,
	*fmt = GET_RTX_FORMAT (GET_CODE (x));

	p = fmt;

	while ('\0' != (*p))
	{
	    if ('e' == (*p))
	    {
		XEXP (x, p - fmt) =
		    ensure_signed_pointer_offsets (XEXP (x, p - fmt));
	    }
	    ++p;
	}
    }
    return x;
}

static rtx
 ensure_pointer_result_size (x)
rtx x;
{
    int mask = (~((-1) << POINTER_SIZE));

    if (CONST_INT == GET_CODE (x))
    {
	int value = INTVAL (x) & mask;

	/* sign extend it to provisionally help the Chimera people. */
	value <<= (HOST_BITS_PER_INT - POINTER_SIZE);
	x = gen_rtx (CONST_INT, VOIDmode,
		     value >> (HOST_BITS_PER_INT - POINTER_SIZE));
    }
    else if (find_symbol (x))
    {
	x = ensure_signed_pointer_offsets (x);
    }
    else
    {
	char *p,
	*fmt = GET_RTX_FORMAT (GET_CODE (x));

	p = fmt;

	while ('\0' != (*p))
	{
	    if ('e' == (*p))
	    {
		XEXP (x, p - fmt) =
		    ensure_pointer_result_size (XEXP (x, p - fmt));
	    }
	    ++p;
	}
    }
    return x;
}
#endif

void
 print_operand (file, x, code)
FILE *file;
rtx x;
char code;
{
    rtx ensure_integer_result_size ();

#if defined( DSP96000 )
    if ('n' == code)
    {
	fprintf (file, "\n");
	return;
    }
#endif

#if defined( DSP56000 )
    if ('p' == code)
    {
	x = ensure_pointer_result_size (copy_rtx (x));

	output_addr_const (file, x);
	return;
    }
#endif

    if ('q' == code)
    {
	x = ensure_integer_result_size (copy_rtx (x));

	output_addr_const (file, x);
	return;
    }

    switch (GET_CODE (x))
    {
    case NE:
    case EQ:
    case GE:
    case GT:
    case LE:
    case LT:
#if ! defined( HAVE_UNSIGNED_BRANCH_CODES )
    case NEU:
    case EQU:
#endif
    case CC:
    case CS:
    case GEU:
    case GTU:
    case LEU:
    case LTU:
	{
	    const char *names[] =
	    {
		"ne", "eq", "ge", "gt", "le", "lt",
#if defined( HAVE_FRACTIONAL_ACCUMULATORS )
		    "ne", "eq",
#endif
		    "ge", "gt", "le", "lt", "cs", "cc"
	    };

	    if ('f' == code)
	    {
		fputs (names[reverse_condition (GET_CODE (x)) - NE],
		       file);
	    }
	    else
	    {
		fputs (names[GET_CODE (x) - NE], file);
	    }
	}
	break;

    case REG:
#if defined( DSP96000 )
	/*
	 * Registers need special handling on m96k. When they are targets
	 * of moves, data alu registers must be suffiexed by .l,.m,.h,.s
	 * or .d Operands that may need this extension use the Code 'z'.
	 */
	if ('z' == code)
	{
	    fprintf (file, "%s", Rname (REGNO (x), GET_MODE (x)));
	    return;
	}

	fprintf (file, "%s", RegNames[REGNO (x)]);
	return;
#else

	switch (code)
	{
	case 'd':
	    fprintf (file, "%s", reg_names[REGNO (x)]);
	    break;

	case 'e':
	    if (DST_REGS == REGNO_REG_CLASS (REGNO (x)))
	    {
		if (MEM_IN_STRUCT_P (x))
		{
		    /* see comment in final.c */
		    fprintf (file, "%s0", reg_names[REGNO (x)]);
		}
		else
		{
#if defined( HAVE_FRACTIONAL_SUPPORT )
		    enum machine_mode mode = GET_MODE (x);

		    if (SZmode == mode || DZmode == mode ||
			XSZmode == mode || XDZmode == mode)
		    {
			/* use saturation with frac operands. */

			fprintf (file, "%s", reg_names[REGNO (x)]);
		    }
		    else
		    {
			fprintf (file, "%s1", reg_names[REGNO (x)]);
		    }
#else
		    fprintf (file, "%s1", reg_names[REGNO (x)]);
#endif
		}
	    }
	    else
	    {
		fprintf (file, "%s", reg_names[REGNO (x)]);
	    }
	    break;

#if defined( HAVE_FRACTIONAL_ACCUMULATORS )
	case 'g':
	    if (DSP_Y1_REGNUM >= REGNO (x) &&
		DSP_X0_REGNUM <= REGNO (x))
	    {
		fprintf (file, "%s", reg_names[REGNO (x) ^ 1]);
	    }
	    else
	    {
		fputs (reg_names[REGNO (x)], file);
	    }
	    break;

	case 'h':
	    fputs (reg_names[REGNO (x)], file);

	    if (DSP_B_REGNUM >= REGNO (x))
	    {
		fputc ('0', file);
	    }
	    break;
#endif

	case 'i':
	    fprintf (file, "%c", *(reg_names[REGNO (x)]));
	    break;

	case 'j':
	    fprintf (file, "n%c", *(reg_names[REGNO (x)] + 1));
	    break;

	case 'k':
	    fprintf (file, "%s2", reg_names[REGNO (x)]);
	    break;

#if defined( HAVE_FRACTIONAL_ACCUMULATORS )
	case 'm':
	    if (DSP_B_REGNUM < REGNO (x))
	    {
		fprintf (file, "%c", (reg_names[REGNO (x)])[0]);
	    }
	    else
	    {
		fprintf (file, "%c10", (reg_names[REGNO (x)])[0]);
	    }
	    break;
#endif

#if defined( DSP56100 )
	case 't':
	    fputc (reg_names[REGNO (x)][0], file);
	    fputc ('1', file);
	    break;
#endif

	default:
	    if (MEM_IN_STRUCT_P (x))
	    {
		/* see comment in final.c */
		fprintf (file, "%s0", reg_names[REGNO (x)]);
	    }
	    else
	    {
		fprintf (file, "%s", reg_names[REGNO (x)]);
	    }
	    break;
	}
	break;
#endif

    case MEM:
#if defined( HAVE_L_MEMORY ) && 0
    case XMEM:
#endif

#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
	if ('f' != code)
	{
	    /* print out a 'l' instead of an x or y when we need to */
	    if (('l' == memory_model) &&
		((DImode == GET_MODE (x)) || (SFmode == GET_MODE (x)) ||
		 (DFmode == GET_MODE (x))))
	    {
		putc ('l', file);
	    }

#if defined( HAVE_L_MEMORY )
#if 0
	    else if (XMEM == GET_CODE (x))
#else
	    else if (RTX_ALTERNATE_MEMORY_P (x))
#endif
	    {
		putc (('y' == mem_space) ? 'x' : 'y', file);
	    }
#endif
	    else
	    {
		putc (mem_space, file);
	    }
	    putc (':', file);
	}
	output_address (ensure_integer_result_size
			(copy_rtx (XEXP (x, 0))));
	return;

#elif defined( DSP56100 )
	switch (code)
	{
	case 'f':
	    fprintf (file, "x:");
	    output_address (XEXP (x, 0));
	    break;

	default:
	    output_address (XEXP (x, 0));
	}
	return;
#endif

    case CONST_INT:
	/* sign extend "negative" numbers so that the asm printout 
	   looks better to most humans. Also, limit the precision to 
	   24 bits. */
	{
#if defined( DSP56000 ) || defined( DSP56300 )
	    int value = 0xffffff & INTVAL (x);

#elif defined( DSP56100 ) || defined( DSP56600 )
	    int value = 0xffff & INTVAL (x);

#elif defined( DSP96000 )
	    int value = INTVAL (x);

#endif
	    int mask = -1 << BITS_PER_UNIT;

	    if (value >> (BITS_PER_UNIT - 1))
	    {
		value |= mask;
	    }
	    fprintf (file, "%d", value);
	}
	return;

#if defined( DSP56000 ) || defined( DSP56300 )
    case CONST_DOUBLE:
	switch (code)
	{
	case 'e':
	    {
		int value = upper_24 (x);

		if (0x00800000 & value)
		{
		    value = ((value << 8) >> 8);
		}
		fprintf (file, "%d", value);
	    }
	    break;

	case 'h':
	    {
		int value = lower_24 (x);

		if (0x00800000 & value)
		{
		    value = ((value << 8) >> 8);
		}
		fprintf (file, "%d", value);
	    }
	    break;

	default:
	    output_operand_lossage ("expression isn't a printable operand.");
	}
	return;
#elif defined( DSP56600 )
    case CONST_DOUBLE:
	switch (code)
	{
	case 'e':
	    {
		int value = upper_16 (x);

		if (0x00008000 & value)
		{
		    value = ((value << 16) >> 16);
		}
		fprintf (file, "%d", value);
	    }
	    break;

	case 'h':
	    {
		int value = lower_16 (x);

		if (0x00008000 & value)
		{
		    value = ((value << 16) >> 16);
		}
		fprintf (file, "%d", value);
	    }
	    break;

	default:
	    output_operand_lossage ("expression isn't a printable operand.");
	}
	return;
#elif defined( DSP96000 )
    case CONST_DOUBLE:
	if (GET_MODE (x) == SFmode)
	{
	    union
	    {
		double d;
		int i[2];
	    }
	    u;
	    union
	    {
		float f;
		int i;
	    }
	    u1;
	    u.i[0] = CONST_DOUBLE_LOW (x);
	    u.i[1] = CONST_DOUBLE_HIGH (x);
	    u1.f = u.d;

	    if (code == 'f')
	    {
		fprintf (file, "%.9g", u1.f);
		return;
	    }

	    fprintf (file, "#$%x", u1.i);
	    return;
	}

	if (GET_MODE (x) != DImode)
	{
	    union
	    {
		double d;
		int i[2];
	    }
	    u;
	    u.i[0] = CONST_DOUBLE_LOW (x);
	    u.i[1] = CONST_DOUBLE_HIGH (x);

	    fprintf (file, "%.20g", u.d);
	    return;
	}
#endif

    case SYMBOL_REF:
    case LABEL_REF:
	output_address (x);
	break;

    case CONST:
	output_addr_const (file, x);
	break;

    default:
	output_operand_lossage ("expression isn't a printable operand.");
    }
}

#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
int
 local_fix_frame_pointer_address (addr_ptr, depth)
rtx *addr_ptr;
int depth;
{
    int frame_offset,
     stack_offset,
     i;

    if (PLUS == GET_CODE (*addr_ptr))
    {
	if (REG == GET_CODE (XEXP (*addr_ptr, 0)) &&
	    FRAME_POINTER_REGNUM == REGNO (XEXP (*addr_ptr, 0)))
	{
	    frame_offset = INTVAL (XEXP (*addr_ptr, 1));
	}
	else if (REG == GET_CODE (XEXP (*addr_ptr, 1)) &&
		 FRAME_POINTER_REGNUM == REGNO (XEXP (*addr_ptr, 1)))
	{
	    frame_offset = INTVAL (XEXP (*addr_ptr, 0));
	}
	else
	{
	    return 0;
	}
    }
    else if (REG == GET_CODE (*addr_ptr) &&
	     FRAME_POINTER_REGNUM == REGNO (*addr_ptr))
    {
	frame_offset = 0;
    }
    else
    {
	return 0;
    }

    stack_offset = frame_offset - depth - get_frame_size ();

    /* we must also account for the registers that have been saved, since
       they're between the stack pointer and all local data and paramters. */

    if (TARGET_CONSERVE_P_MEM)
    {
#if defined( DSP56100 )
	int push_pop_offset = 8,
	 pro_epi_offset = 9;

#else	/* if defined( DSP56300 ) */
	int push_pop_offset = 8,
	 pro_epi_offset = 9;

#endif
	/* if the frame size > 0 or this isn't a leaf, then we
	   call the full prolog / epilog function. */
	if (get_frame_size () || (current_func_info & FUNC_ISNT_LEAF))
	{
	    stack_offset -= pro_epi_offset;
	}
	/* otherwise, if the are register saves needed, just call the 
	   register save / restore function. */
	else
	{
	    /* index bit 1 is set if there are register saves needed, bit 0
	       is set if there is a size offset to compute. */

	    int i;

	    for (i = 0; i < FIRST_PSEUDO_REGISTER; ++i)
	    {
		if (regs_ever_live[i] && !call_used_regs[i])
		{
		    break;
		}
	    }
	    if (i != FIRST_PSEUDO_REGISTER)
	    {
		stack_offset -= push_pop_offset;
	    }
	}
    }
    else
    {
	/* parameters may be one unit further away from the stack pointer due
	   to a saved ssh. paramter references always have negative offsets
	   from the frame pointer. */

	if ((current_func_info & FUNC_ISNT_LEAF) ||
	    !TARGET_CALL_OVERHEAD_REDUCTION)
	{
	    --stack_offset;
	}

	for (i = 0; i < FIRST_PSEUDO_REGISTER; ++i)
	{
	    if (DSP_FAKE_REGNUM == i)
	    {
		continue;
	    }
	    if (regs_ever_live[i] && !call_used_regs[i])
	    {
		--stack_offset;

		if (DST_REGS == REGNO_REG_CLASS (i)
#if defined( DSP56300 ) || defined( DSP56600 )
		    && !TARGET_L_MEMORY
#endif
		    )
		{
		    /* it takes two words to save an accumulator if we're
		       not using the L memory model. */

		    --stack_offset;
		}
	    }
	}
#if defined( DSP56300 ) || defined( DSP56600 )
	/* if we're using the L memory model, it only takes one address
	   to save both halves of a source register. */

	if (TARGET_L_MEMORY)
	{
	    if ((regs_ever_live[DSP_X0_REGNUM] &&
		 !call_used_regs[DSP_X0_REGNUM]) &&
		(regs_ever_live[DSP_X1_REGNUM] &&
		 !call_used_regs[DSP_X1_REGNUM]))
	    {
		++stack_offset;
	    }

	    if ((regs_ever_live[DSP_Y0_REGNUM] &&
		 !call_used_regs[DSP_Y0_REGNUM]) &&
		(regs_ever_live[DSP_Y1_REGNUM] &&
		 !call_used_regs[DSP_Y1_REGNUM]))
	    {
		++stack_offset;
	    }
	}
#endif
    }

    if (!stack_offset)
    {
	*addr_ptr = stack_pointer_rtx;
    }
    else
    {
	*addr_ptr = gen_rtx (PLUS, Pmode, stack_pointer_rtx,
			     gen_rtx (CONST_INT, VOIDmode, stack_offset));
    }
    return stack_offset;
}
#endif

/* functions used to facilitate the generation of jscc instructions. */
int
 jscc_combo_follows_p (insn, label)
rtx insn,
 label;
{
    rtx next = NEXT_INSN (insn);
    rtx body = PATTERN (insn);

    /* make sure we're dealing with a call */
    if (CALL_INSN != GET_CODE (insn))
    {
	return 0;
    }

    /* check for a single label reference. */
    if (1 != LABEL_NUSES (label))
    {
	return 0;
    }

    /* verify that our CODE_LABEL follows that. */
    while (NULL != next && NOTE == GET_CODE (next))
    {
	next = NEXT_INSN (next);
    }

    if (NULL == next || label != next)
    {
	return 0;
    }
    return 1;
}

#if ! defined( DSP96000 )
int
 tcc_combo_follows_p (insn, label)
rtx insn,
 label;
{
    rtx next = NEXT_INSN (insn);
    rtx body = PATTERN (insn);

    /* make sure we're dealing with an alu tfr opcode. */
    if (GET_MODE (SET_SRC (body)) != GET_MODE (SET_DEST (body)) ||
	REGNO (SET_DEST (body)) > DSP_B_REGNUM ||
	REGNO (SET_SRC (body)) > DSP_Y1_REGNUM)
    {
	return 0;
    }

    /* check for a single label reference. */
    if (1 != LABEL_NUSES (label))
    {
	return 0;
    }

    /* verify that our CODE_LABEL follows that. */
    while (NULL != next && NOTE == GET_CODE (next))
    {
	next = NEXT_INSN (next);
    }

    if (NULL == next || label != next)
    {
	return 0;
    }
    return 1;
}
#endif

#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
char *
 output_long_mem_bitop (name, operands)
char *name;
rtx *operands;
{
    operands[2] = gen_rtx (SYMBOL_REF, VOIDmode, name);

    if (TARGET_L_MEMORY)
    {
	/* LSW in x memory, MSW in y memory. */

	operands[0] = XEXP (operands[0], 0);

	if ((BITS_PER_UNIT-1) < INTVAL (operands[1]))
	{
	    operands[1] = gen_rtx
		(CONST_INT, VOIDmode, INTVAL (operands[1]) - BITS_PER_UNIT);

	    return "b%2	#%1,x:(%0)";
	}
	else
	{
	    return "b%2	#%1,y:(%0)";
	}
    }
    else
    {
	operands[0] = XEXP (operands[0], 0);

	/* LSW @ address, MSW @ address + 1. */

	if ((BITS_PER_UNIT-1) < INTVAL (operands[1]))
	{
	    operands[1] = gen_rtx
		(CONST_INT, VOIDmode, INTVAL (operands[1]) - BITS_PER_UNIT);

	    return fms ("move	(%0)+\n\tb%2	#%1,@:(%0)-");
	}
	else
	{
	    return fms ("b%2	#%1,@:(%0)");
	}
    }
}
#endif

#if defined( DSP56000 ) || defined( DSP56300 ) || defined( DSP56600 )
char *
 output_long_reg_bitop (name, operands)
char *name;
rtx *operands;
{
    operands[3] = gen_rtx (SYMBOL_REF, VOIDmode, name);

    if ((BITS_PER_UNIT-1) < INTVAL (operands[2]))
    {
	operands[2] = gen_rtx
	    (CONST_INT, VOIDmode, INTVAL (operands[2]) - BITS_PER_UNIT);

	if (DSP_B_REGNUM >= REGNO (operands[0]))
	{
	    return "b%3	#%2,%e0";
	}
	else
	{
	    return "b%3	#%2,%g0";
	}
    }
    else
    {
	return "b%3	#%2,%h0";
    }
}
#endif

static char *
 output_jscc (insn, operands, flip_p)
rtx insn,
 operands[];
int flip_p;
{
    /* if we need to, flip the conditional. */
    if (flip_p)
    {
	PUT_CODE (operands[1], reverse_condition (GET_CODE (operands[1])));
    }

    /* delete the follow-on label. */
    {
	rtx next = NEXT_INSN (insn);

	while (CODE_LABEL != GET_CODE (next))
	{
	    next = NEXT_INSN (next);
	}
	PUT_CODE (next, NOTE);
	INSN_DELETED_P (next) = 1;
    }

#if defined( TARGET_PIC )
    return (TARGET_PIC) ? "bs%1	%f0" : "js%1	%f0";
#else
    return "js%1	%f0";
#endif
}
#endif

static char *
 output_tcc (insn, operands, flip_p)
rtx insn,
 operands[];
int flip_p;
{
    /* if we need to, flip the conditional. */
    if (flip_p)
    {
	PUT_CODE (operands[2], reverse_condition (GET_CODE (operands[2])));
    }

    /* delete the follow-on label. */
    {
	rtx next = NEXT_INSN (insn);

	while (CODE_LABEL != GET_CODE (next))
	{
	    next = NEXT_INSN (next);
	}
	PUT_CODE (next, NOTE);
	INSN_DELETED_P (next) = 1;
    }

    return "t%2	%1,%0";
}

static char *
 output_jsbit (insn, operands, set_p)
rtx insn,
 operands[];
int set_p;
{
    /* delete the follow-on label. */
    {
	rtx next = NEXT_INSN (insn);

	while (CODE_LABEL != GET_CODE (next))
	{
	    next = NEXT_INSN (next);
	}
	PUT_CODE (next, NOTE);
	INSN_DELETED_P (next) = 1;
    }

    if (set_p)
    {
#if defined( TARGET_PIC )
	return (TARGET_PIC) ?
	    "bsset	#%2,%1,%f0" :

	    "jsset	#%2,%1,%f0";
#else
	return "jsset	#%2,%1,%f0";
#endif
    }
    else
    {
#if defined( TARGET_PIC )
	return (TARGET_PIC) ?
	    "bsclr	#%2,%1,%f0" :

	    "jsclr	#%2,%1,%f0";
#else
	return "jsclr	#%2,%1,%f0";
#endif
    }
}

/* this vairable is set by function_logue to tell sdbout.c which pro/ep
   sequence is being used. */
int logue_index;

void
 function_logue (file, size, epilogue_p)
FILE *file;
int size,
 epilogue_p;
{
    static mem_space_fixed = 0;

    char *tst = (current_func_info & FUNC_RETURNS_FLOAT) ?
#if defined( DSP96000 )
     "ftst	d0" : "tst	d0";

#else
     "tst	a":"tst	a";

#endif

    static struct
    {
	char *pro,
	*epi;
    }
    *chosen, stack_check_logues =
    /* a special epilog/prolog that watches for stack heap collision. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
    {
	"\
	move	#%d,SO\n\
	move	(SP)+SO\n\
	move	ssh,@:(SP)+\n\
	jsr	F__stack_check\n",
	    "\
	move	#-%d,SO\n\
	move	@:-(SP),ssh\n\
	%s	(SP)+SO\n\
	rts\n\n"
    }
    ,
#else
    {
	"\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n\
	bsr	F__stack_check\n",
	    "\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
    }
    ,
#endif

#if defined( TARGET_CONSERVE_P_MEM )
	tiny_logues[] =
    {
#if defined( DSP56000 )
	/* 00 - no reg saves, no frame temps, no stack parms. */
	{
	    "\
        move	#%d,SO\n\
	jsr	prolog\n",
		"\
	jmp	epilog\n"
	}
	,

	/* 01 - no reg saves, but frame temps or stack parms. */
	{
	    "\
        move	#%d,SO\n\
	jsr	prolog\n",
		"\
	jmp	epilog\n"
	}
	,

	/* 10 - reg saves, but no frame temps, no stack parms. */
	{
	    "\
        move	#%d,SO\n\
	jsr	prolog\n",
		"\
	jmp	epilog\n"
	}
	,

	/* 11 - reg saves, frame temps or stack parms. */
	{
	    "\
        move	#%d,SO\n\
	jsr	prolog\n",
		"\
	jmp	epilog\n"
	}
	,
#else
	/* 00 - no reg saves, no frame temps, no stack parms. */
	{
	    "", "\
	tst	a\n\
	rts\n"
	}
	,

	/* 01 - no reg saves, but frame temps or stack parms. */
	{
	    "\
        move	#%d,SO\n\
	jsr	prolog\n",
		"\
        move	#%d,SO\n\
	jmp	epilog\n"
	}
	,

	/* 10 - reg saves, but no frame temps, no stack parms. */
	{
	    "\
	jsr	push_regs\n",
		"\
	jsr	pop_regs\n\
	rts\n"
	}
	,

	/* 11 - reg saves, frame temps or stack parms. */
	{
	    "\
        move	#%d,SO\n\
	jsr	prolog\n",
		"\
        move	#%d,SO\n\
	jmp	epilog\n"
	}
	,
#endif
	/* terminator */
	{
	    NULL, NULL
	}

    }
    ,
#endif

	logues[] =
    {
	/* 00000: non-void, no stack parms, leaf, no frame temps. */
	{
	    "",
		"\
	%s\n\
	rts\n\n"
	}
	,

	/* 00001: non-void, stack parms, leaf, no frame temps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "",
		"\
	%s\n\
	rts\n\n"
	}
	,
#elif defined( DSP96000 )
	{
	    "\
	move	FP,@:(SP)+\n\
	lea	(SP)+,FP\n",
		"\
	%s	@:-(SP),FP\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	lua	(SP)+,FP\n",
		"\
	%s	@:-(SP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 00010: non-void, no stack parms, not leaf, no frame temps. */
	{
	    "\
	move	ssh,@:(SP)+\n",
		"\
	%s	(SP)-\n\
	move	@:(SP),ssh\n\
	rts\n\n"
	}
	,

	/* 00011: non-void, stack parms, not leaf, no frame temps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	ssh,@:(SP)+\n",
		"\
	%s	(SP)-\n\
	move	@:(SP),ssh\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	SP,FP\n",
		"\
	move	(SP)-\n\
	move	@:(SP)-,ssh\n\
	%s	@:(SP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 00100: non-void, no stack parms, leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n",
		"\
	%s	(SP)-\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n",
		"\
	%s	(SP)-\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(FP)-\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 00101: non-void, stack parms, leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n",
		"\
	%s	(SP)-\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n",
		"\
	%s	(SP)-\n\
	rts\n\n"
	}
	,
#elif defined( DSP96000 )
	{
	    "\
	move	FP,@:(SP)+\n\
	move	(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	lea	(SP+-3),SP\n\
	%s	@:(SP),FP\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(SP)-\n\
	move	(SP)-\n\
	move	(SP)-\n\
	%s	@:(SP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 00110: non-void, no stack parms, not leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n\
	move	ssh,@:(SP)+\n",
		"\
	%s	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n\
	move	ssh,@:(SP)+\n",
		"\
	%s	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 00111: non-void, stack parms, not leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n\
	move	ssh,@:(SP)+\n",
		"\
	%s	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n\
	move	ssh,@:(SP)+\n",
		"\
	%s	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 01000 thru 01011: illegal. */
	{
	    "", ""
	}
	,
	{
	    "", ""
	}
	,
	{
	    "", ""
	}
	,
	{
	    "", ""
	}
	,

	/* 01100: non-void, no stack parms, leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n",
		"\
	move	#-%d,SO\n\
	%s	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 01101: non-void, stack parms, leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n",
		"\
	move	#-%d,SO\n\
	%s	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	(FP)-\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 01110: non-void, no stack parms, not leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	#-%d,SO\n\
	move	@:(SP),ssh\n\
	%s	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 01111: non-void, stack parms, not leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	@:(SP),ssh\n\
	move	#-%d,SO\n\
	%s	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	%s	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 10000: void, no stack parms, leaf, no frame temps. */
	{
	    "",
		"\
	rts\n\n"
	}
	,

	/* 10001: void, stack parms, leaf, no frame temps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "",
		"\
	rts\n\n"
	}
	,
#elif defined( DSP96000 )
	{
	    "\
	move	FP,@:(SP)+\n\
	lea	(SP)+,FP\n",
		"\
	move	@:-(SP),FP\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	lua	(SP)+,FP\n",
		"\
	move	@:-(SP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 10010: void, no stack parms, not leaf, no frame temps. */
	{
	    "\
	move	ssh,@:(SP)+\n",
		"\
	move	@:-(SP),ssh\n\
	rts\n\n"
	}
	,

	/* 10011: void, stack parms, not leaf, no frame temps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	ssh,@:(SP)+\n",
		"\
	move	@:-(SP),ssh\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	SP,FP\n",
		"\
	move	(SP)-\n\
	move	@:(SP)-,ssh\n\
	move	@:(SP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 10100: void, no stack parms, leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n",
		"\
	move	(SP)-\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n",
		"\
	move	(SP)-\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(FP)-\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 10101: void, stack parms, leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n",
		"\
	move	(SP)-\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n",
		"\
	move	(SP)-\n\
	rts\n\n"
	}
	,
#elif defined( DSP96000 )
	{
	    "\
	move	FP,@:(SP)+\n\
	move	(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	lea	(SP+-3),SP\n\
	move	@:(SP),FP\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(SP)-\n\
	move	(SP)-\n\
	move	(SP)-\n\
	move	@:(SP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 10110: void, no stack parms, not leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 10111: void, stack parms, not leaf, 1 frame tmp. */
#if defined( DSP56100 )
	{
	    "\
	lea	(SP)+,SP\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#elif defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	(SP)+\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	@:(SP)-,ssh\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	SP,FP\n\
	move	(SP)+\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 11000 thru 11011: illegal. */
	{
	    "", ""
	}
	,
	{
	    "", ""
	}
	,
	{
	    "", ""
	}
	,
	{
	    "", ""
	}
	,

	/* 11100: void, no stack parms, leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n",
		"\
	move	#-%d,SO\n\
	move	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 11101: void, stack parms, leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n",
		"\
	move	#-%d,SO\n\
	move	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	(FP)-\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 11110: void, no stack parms, not leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	#-%d,SO\n\
	move	@:(SP),ssh\n\
	move	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif

	/* 11111: void, stack parms, not leaf, n frame tmps. */
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	{
	    "\
	move	#%d,SO\n\
	move	(SP)+SO\n\
	move	ssh,@:(SP)+\n",
		"\
	move	(SP)-\n\
	move	#-%d,SO\n\
	move	@:(SP),ssh\n\
	move	(SP)+SO\n\
	rts\n\n"
	}
	,
#else
	{
	    "\
	move	FP,@:(SP)+\n\
	move	ssh,@:(SP)+\n\
	move	#%d,SO\n\
	move	SP,FP\n\
	move	(SP)+SO\n",
		"\
	move	(FP)-\n\
	move	@:(FP)-,ssh\n\
	move	FP,SP\n\
	move	@:(FP),FP\n\
	rts\n\n"
	}
	,
#endif
	/* terminator. */
	{
	    NULL, NULL
	}
    };

    if (!mem_space_fixed)
    {
	int i;

	for (i = 0; NULL != logues[i].pro; ++i)
	{
	    logues[i].pro = strcpy
		((char *) xmalloc
		 (strlen (logues[i].pro) + 1), fms (logues[i].pro));
	    fix_sp_fp (logues[i].pro);

	    logues[i].epi = strcpy
		((char *) xmalloc
		 (strlen (logues[i].epi) + 1), fms (logues[i].epi));
	    fix_sp_fp (logues[i].epi);
	}

	for (i = 0; NULL != tiny_logues[i].pro; ++i)
	{
	    tiny_logues[i].pro = strcpy
		((char *) xmalloc
		 (strlen (tiny_logues[i].pro) + 1),
		 fms (tiny_logues[i].pro));

	    fix_sp_fp (tiny_logues[i].pro);

	    tiny_logues[i].epi = strcpy
		((char *) xmalloc
		 (strlen (tiny_logues[i].epi) + 1),
		 fms (tiny_logues[i].epi));

	    fix_sp_fp (tiny_logues[i].epi);
	}

	/* don't forget to initialize the stack checking logues! */
	stack_check_logues.pro = strcpy
	    ((char *) xmalloc (strlen
			       (stack_check_logues.pro) + 1),
	     fms (stack_check_logues.pro));
	fix_sp_fp (stack_check_logues.pro);

	stack_check_logues.epi = strcpy
	    ((char *) xmalloc (strlen
			       (stack_check_logues.epi) + 1),
	     fms (stack_check_logues.epi));
	fix_sp_fp (stack_check_logues.epi);

	mem_space_fixed = 1;
    }

    /* switch bits are set if (0th bit is LSBit):
       0: the function has stack parameters.
       1: the function is not a leaf function (has call(s)).
       3,2: the function has no frame based temps (00), the function
       has one frame based temp (01), or the function has more than
       one frame based temp (11).
       4: the function is of type void.
     */

    if (TARGET_CONSERVE_P_MEM)
    {
	/* index bit 1 is set if there are register saves needed, bit 0
	   is set if there is a size offset to compute. */

	int i;

	for (i = 0; i < FIRST_PSEUDO_REGISTER; ++i)
	{
	    if (regs_ever_live[i] && !call_used_regs[i])
	    {
		break;
	    }
	}
	logue_index =
	    -(((i != FIRST_PSEUDO_REGISTER) << 1) |
	      ((0 != (current_func_info & FUNC_ISNT_LEAF)) ||
	       (size != 0))) - 1;

	chosen = &tiny_logues[-logue_index - 1];
    }
    else if (TARGET_STACK_CHECK)
    {
	chosen = &stack_check_logues;
    }
    else if (TARGET_CALL_OVERHEAD_REDUCTION)
    {
	if (epilogue_p && NULL != cc_status.value2 &&
	    FUNCTION_VALUE_REGNO_P (REGNO (cc_status.value2)))
	{
	    current_func_info |= FUNC_RETURNS_VOID;
	}

	logue_index = ((0 != (current_func_info & FUNC_HAS_STACK_PARMS)) |
		       ((0 != (current_func_info & FUNC_ISNT_LEAF)) << 1) |
		       ((0 != size) << 2) |
		       ((1 < size) << 3) |
		       ((0 != (current_func_info & FUNC_RETURNS_VOID))
			<< 4));

	chosen = &logues[logue_index];
    }
    else
    {
	/* if they've turned off call overhead reduction, go for the safest
	   possible pro/epilogue: stack parms, non-leaf, > 1 frame temps, and
	   not a void function. */

	chosen = &logues[logue_index = 0x0f];
    }

    if (epilogue_p)
    {
	popregs (file);

	if (TARGET_CONSERVE_P_MEM)
	{
	    fprintf (file, chosen->epi, size);
	}
#if defined( DSP56100 ) || defined( DSP56300 ) || defined( DSP56600 )
	else if (TARGET_STACK_CHECK || 1 < size ||
		 !TARGET_CALL_OVERHEAD_REDUCTION)
	{
	    fprintf (file, chosen->epi, size, tst);
	}
#endif
	else
	{
	    fprintf (file, chosen->epi, tst);
	}
    }
    else
    {
#if defined( DSP56000 ) || defined( DSP56100 )
	/* on the 56k, we need to clear out the assumed n register values
	   at the beginning of the function. */
	clear_n_reg_values ();
#endif
	fprintf (file, chosen->pro, size, size);

#if defined( DSP96000 )
	SetLocalBusy ();
#endif
	pushregs (file);
    }
}

int
 really_reg_p (operand)
rtx operand;
{
    while (SUBREG == GET_CODE (operand))
    {
	operand = XEXP (operand, 0);
    }
    return REG_P (operand);
}

#if defined( HAVE_FRACTIONAL_ACCUMULATORS )
char *
 output_compare (insn, operands)
rtx insn,
*operands;
{
    extern enum mdep_cc_info next_cc_use ();
    extern rtx next_cc_relevancy ();
    rtx relevant_insn,
     relevant_body;

    /* sometimes this compiler manages attempt ( compare ( reg x ) ( reg x )).
       the condition codes implied by this are obvious. we peek down the insn
       chain and discover the next_cc0_relevancy ( ). if we discover
       that the ccs are being set gratuitiously, we ignore this compare, 
       otherwise we change the cc0 consumer to be a constant load in the case
       of an Scond or a hard branch or nop in the case of a Bcond. */

    if (
#if defined( DSP56300 ) || defined( DSP56600 )
	   REG_P (operands[0]) && REG_P (operands[1]) &&
#endif
	   (REGNO (operands[0]) == REGNO (operands[1])))
    {
	relevant_insn = next_cc_relevancy (insn);
	relevant_body = PATTERN (relevant_insn);

	if (JUMP_INSN == GET_CODE (relevant_insn))
	{
	    if ((IF_THEN_ELSE == GET_CODE (relevant_body =
					   XEXP (relevant_body, 1))) &&
		(CONST_INT !=
		 GET_CODE (XEXP (relevant_body, 0))))
	    {
		if ((GEU == GET_CODE (XEXP (relevant_body, 0))) ||
		    (GE == GET_CODE (XEXP (relevant_body, 0))) ||
		    (LEU == GET_CODE (XEXP (relevant_body, 0))) ||
		    (LE == GET_CODE (XEXP (relevant_body, 0))) ||
		    (EQU == GET_CODE (XEXP (relevant_body, 0))) ||
		    (EQ == GET_CODE (XEXP (relevant_body, 0))))
		{
		    /* change ineq into const1_rtx because x == x always. */

		    if (pc_rtx == XEXP (XEXP (PATTERN (relevant_insn),
					      1), 1))
		    {
			PUT_CODE (relevant_insn, NOTE);
			INSN_DELETED_P (relevant_insn) = 1;
		    }
		    else
		    {
			INSN_CODE (relevant_insn) = -1;

			XEXP (PATTERN (relevant_insn), 1) =
			    XEXP (XEXP (PATTERN (relevant_insn), 1), 1);
		    }
		}
		else
		{
		    /* otherwise change ineq into const0_rtx. */

		    if (pc_rtx == XEXP (XEXP (PATTERN (relevant_insn),
					      1), 1))
		    {
			INSN_CODE (relevant_insn) = -1;

			XEXP (PATTERN (relevant_insn), 1) =
			    XEXP (XEXP (PATTERN (relevant_insn), 1), 2);
		    }
		    else
		    {
			PUT_CODE (relevant_insn, NOTE);
			INSN_DELETED_P (relevant_insn) = 1;
		    }
		}
	    }
	}
	else if ((INSN == GET_CODE (relevant_insn)) &&
		 (PARALLEL == GET_CODE (relevant_body)) &&
		 (SET == GET_CODE (relevant_body =
				   XVECEXP (relevant_body, 0, 0))) &&
		 (0 == strcmp ("ee",
		    GET_RTX_FORMAT (GET_CODE (XEXP (relevant_body, 1))))) &&
		 (cc0_rtx == XEXP (XEXP (relevant_body, 1), 0)))
	{
	    if ((GEU == GET_CODE (XEXP (relevant_body, 1))) ||
		(GE == GET_CODE (XEXP (relevant_body, 1))) ||
		(LEU == GET_CODE (XEXP (relevant_body, 1))) ||
		(LE == GET_CODE (XEXP (relevant_body, 1))) ||
		(EQU == GET_CODE (XEXP (relevant_body, 1))) ||
		(EQ == GET_CODE (XEXP (relevant_body, 1))))
	    {
		/* change ineq into const1_rtx because x == x always. */

		XEXP (relevant_body, 1) = const1_rtx;
	    }
	    else
	    {
		/* otherwise change ineq into const0_rtx. */

		XEXP (relevant_body, 1) = const0_rtx;
	    }
	    INSN_CODE (relevant_insn) = -1;
	}
	return "";
    }

    if (UNSIGNED_COMPARE_P (insn))
    {
#if defined( DSP56300 ) || defined( DSP56600 )
	if (REG_P (operands[1]))
	{
	    return "cmpu	%1,%0";
	}
	else
	{
	    return "move	#0,%k0" SC
		"cmp	#%1,%0";
	}
#elif defined( DSP56100 )
	return "zero	%0" SC
	    "zero	%1" SC
	    "cmp	%1,%0";
#else
	return "move	#0,%k0" SC
	    "move	#0,%k1" SC
	    "cmp	%1,%0";
#endif
    }
    else
    {
	cc_status.mdep = CC_SIGNED;

#if defined( DSP56300 ) || defined( DSP56600 )
	return (which_alternative && SImode == GET_MODE (operands[0])) ?
	    "cmp	#%1,%0" : "cmp	%1,%0";
#else
	return "cmp	%1,%0";
#endif
    }
}

char *
 output_tst (insn, operands)
rtx insn,
*operands;
{
    if (CC_UNSIGNED == (cc_status.mdep = next_cc_use (insn)))
    {
#if defined( DSP56100 )
	if (SImode == GET_MODE (operands[0]) &&
	    DST_REGS != REGNO_REG_CLASS (REGNO (operands[0])))
	{
	    return "tst	%0" SC
		"andi	#$f7,ccr";
	}
	else
	{
	    return "zero	%0" SC
		"tst	%0";
	}
#else
	return "move	#0,%k0" SC
	    "tst	%0";
#endif
    }
    else
    {
	if (cc_status.value2 &&
	    VOIDmode == GET_MODE (cc_status.value2))
	{
	    PUT_MODE (cc_status.value2, GET_MODE (operands[0]));
	}
	return (rtx_equal_p (operands[0], cc_status.value2))
	    ? "" : "tst	%0";
    }
}

#endif



#if defined ( DSP56600 ) || defined ( DSP56300 ) 
char *
 output_do (insn, operands)
rtx insn,
*operands;
{
    /* we must scan forward and make sure that there is at least one non-do
     * insn between this insn and the target label. If not, 86 the loop. 
     * BTW, a nop instruction doesn't count.
     */
    rtx scan = insn;

    while (scan && (operands[1] != scan))
    {
	if ((INSN == GET_CODE (scan)) &&
	    (const0_rtx != PATTERN (scan)))
	{
	    break;
	}
	scan = NEXT_INSN (scan);
    }

    if (!scan)
    {
	/* do without corresponding od ! */
	abort ();
    }
    if ((scan == operands[1]) ||
	(which_alternative && (1 == INTVAL (operands[0]))))
    {
	/* decrement the number of uses at the target label. */

	--LABEL_NUSES (operands[1]);

	return "";
    }

    return (CONSTANT_P (operands[0])) ? "do	#%0,%l1" :
	"do	%e0,%l1";
}


static int
insn_p (insn)
rtx insn;
{
    return  insn && (
                     (INSN == GET_CODE (insn)) &&
            /* ignore latent nop insns! */
                     (const0_rtx != PATTERN (insn))
                    );

}

static int
jump_insn_p (insn)
rtx insn;
{
    return  insn && (JUMP_INSN == GET_CODE (insn));
}


char *
 output_od (insn, operands)
rtx insn,
*operands;
{

#if defined ( DSP56600 ) 
    int nop_cnt = 3;
#else
    int nop_cnt = 1;
#endif
    rtx scan = PREV_INSN (insn);

#if defined ( DSP56600 ) 
    /* currently we have this ugly solution */
    /* should be fixed asap in assembler */
    return "nop" SC "nop" SC "nop";
#endif

    
    while ( scan && nop_cnt )
    {
        if(jump_insn_p (scan))
            break;
        else if(insn_p (scan))
            nop_cnt--;
        
        scan = PREV_INSN (scan);
    }

    if(scan == 0)
    {
        nop_cnt = 0;
    }
    
    switch(nop_cnt)
    {
    case 1:
        return "nop";
    case 2:
        return "nop" SC "nop";
    case 3:
        return "nop" SC "nop" SC "nop";
    default:
        return "";
    }
}




#else      /* not (defined ( DSP56600 ) || defined ( DSP56300 )) */


char *
 output_do (insn, operands)
rtx insn,
*operands;
{
    extern void record_address_regs_used ();

    /* we must scan forward and make sure that there is at least one non-do
     * insn between this insn and the target label. If not, 86 the loop. 
     * BTW, a nop instruction doesn't count.
     */
    rtx scan = insn;

    while (scan && (operands[1] != scan))
    {
	if ((INSN == GET_CODE (scan)) &&
	    (const0_rtx != PATTERN (scan)))
	{
	    break;
	}
	scan = NEXT_INSN (scan);
    }

    if (!scan)
    {
	/* do without corresponding od ! */
	abort ();
    }
    if ((scan == operands[1]) ||
	(which_alternative && (1 == INTVAL (operands[0]))))
    {
	/* we're not really doing a do; record that no conflicts are 
	 * possible.
	 */
	record_address_regs_used (const0_rtx);

	/* decrement the number of uses at the target label. */

	--LABEL_NUSES (operands[1]);

	return "";
    }

    /* record potential r/n register conflicts. */
    record_address_regs_used (PATTERN (scan));

#if defined( TARGET_PIC )
    if (TARGET_PIC)
    {
	return (CONSTANT_P (operands[0])) ? "dor	#%0,%l1" :
	    "dor	%e0,%l1";
    }
#endif

    return (CONSTANT_P (operands[0])) ? "do	#%0,%l1" :
	"do	%e0,%l1";
}

char *
 output_od (insn, operands)
rtx insn,
*operands;
{
    rtx scan = PREV_INSN (insn);

    while (scan &&
	   ((INSN != GET_CODE (scan)) ||
    /* ignore latent nop insns! */
	    (const0_rtx == PATTERN (scan))) &&
	   (JUMP_INSN != GET_CODE (scan)))
    {
	scan = PREV_INSN (scan);
    }

    if (JUMP_INSN == GET_CODE (scan))
    {
	/* pop the stack - no chance of conflict. */
	(void) conflicting_address_regs_set_p (const0_rtx);

	/* but hey- we can't have a jmp as the last insn within the DO! */
        
        return "nop";

    }
    if (INSN == GET_CODE (scan))
    {
	/* if we have a conflict, emit a noop to sw interlock. */
	if (conflicting_address_regs_set_p (PATTERN (scan)))
	{
            return "nop";
	}
#if defined TARGET_PIC
	if (TARGET_PIC &&
	    (BSR_LAST_OP & get_spec_bits
	     (insn_machine_info[recog_memoized (scan)].spec_bits)))
	{
	    return "nop";
	}
	scan = PREV_INSN (scan);

	while (scan &&
	       ((INSN != GET_CODE (scan)) ||
	/* ignore latent nop insns! */
		(const0_rtx == PATTERN (scan))) &&
	       (JUMP_INSN != GET_CODE (scan)))
	{
	    scan = PREV_INSN (scan);
	}
	if (TARGET_PIC && scan &&
	    ((JUMP_INSN == GET_CODE (scan)) ||
	     (BSR_LAST_OP & get_spec_bits
	      (insn_machine_info[recog_memoized (scan)].spec_bits))))
	{
	    return "nop";
	}
#endif
    }
    return "";
}

#endif     /* defined ( DSP56600 ) || defined ( DSP56300 ) */


rtx
clean_subregs (op)
rtx op;
{
    while (SUBREG == GET_CODE (op))
    {
	op = SUBREG_REG (op);
    }
    return op;
}

/* this is normally implemented in parse.y. we moved the keyword recognizer
   into this file, now that we've extended C to include extra DSP keywords.
   the new keywords are run-time switched, so we need to include both
   the standard and extended perfect hash functions. */

#if defined( TARGET_C_FOR_DSP )

#define NORID RID_UNUSED

struct resword *
 is_reserved_word (str, len)
char *str;
unsigned int len;
{
    struct resword *funky_is_reserved_word ();
    struct resword *normal_is_reserved_word ();

    if (TARGET_C_FOR_DSP)
    {
	return funky_is_reserved_word (str, len);
    }
    return normal_is_reserved_word (str, len);
}

struct resword
{
    char *name;
    short token;
    enum rid rid;
};

#if defined( HAVE_L_MEMORY )
#if defined( FRACTIONAL_SUPPORT )
/* this is a MSIL fractional chip (DSP5600X, DSP5630X):
   add keywords: modulo, frac, extended, x_memory. */

#define TOTAL_KEYWORDS 55
#define MIN_WORD_LENGTH 2
#define MAX_WORD_LENGTH 13
#define MIN_HASH_VALUE 7
#define MAX_HASH_VALUE 95
/* maximum key range = 89, duplicates = 0 */

#ifdef __GNUC__
inline
#endif
static unsigned int
 funky_hash (str, len)
register char *str;
register int unsigned len;
{
    static unsigned char asso_values[] =
    {
	96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 1, 96, 2, 1, 16,
	    10, 5, 18, 40, 12, 17, 96, 1, 10, 1,
	    31, 9, 25, 96, 19, 20, 1, 42, 33, 15,
	    4, 1, 4, 96, 96, 96, 96, 96,
    };
    register int hval = len;

    switch (hval)
    {
    default:
    case 3:
	hval += asso_values[str[2]];
    case 2:
    case 1:
	hval += asso_values[str[0]];
    }
    return hval + asso_values[str[len - 1]];
}

#ifdef __GNUC__
inline
#endif
struct resword *
 funky_is_reserved_word (str, len)
register char *str;
register unsigned int len;
{
    static struct resword wordlist[] =
    {
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "asm", ASM, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__asm", ASM, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__asm__", ASM, NORID
	}
	,
	{
	    "break", BREAK, NORID
	}
	,
	{
	    "__typeof__", TYPEOF, NORID
	}
	,
	{
	    "x_memory", SCSPEC, RID_X_MEMORY
	}
	,
	{
	    "__alignof__", ALIGNOF, NORID
	}
	,
	{
	    "auto", SCSPEC, RID_AUTO
	}
	,
	{
	    "__attribute__", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__attribute", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "do", DO, NORID
	}
	,
	{
	    "int", TYPESPEC, RID_INT
	}
	,
	{
	    "",
	}
	,
	{
	    "extended", TYPESPEC, RID_EXTENDED
	}
	,
	{
	    "__const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "modulo", SCSPEC, RID_MODULO
	}
	,
	{
	    "__const__", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "__typeof", TYPEOF, NORID
	}
	,
	{
	    "__inline__", SCSPEC, RID_INLINE
	}
	,
	{
	    "__alignof", ALIGNOF, NORID
	}
	,
	{
	    "__inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "__signed__", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "float", TYPESPEC, RID_FLOAT
	}
	,
	{
	    "else", ELSE, NORID
	}
	,
	{
	    "short", TYPESPEC, RID_SHORT
	}
	,
	{
	    "default", DEFAULT, NORID
	}
	,
	{
	    "if", IF, NORID
	}
	,
	{
	    "inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "__signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "frac", TYPESPEC, RID_FRAC
	}
	,
	{
	    "char", TYPESPEC, RID_CHAR
	}
	,
	{
	    "while", WHILE, NORID
	}
	,
	{
	    "extern", SCSPEC, RID_EXTERN
	}
	,
	{
	    "static", SCSPEC, RID_STATIC
	}
	,
	{
	    "case", CASE, NORID
	}
	,
	{
	    "struct", STRUCT, NORID
	}
	,
	{
	    "__volatile__", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "sizeof", SIZEOF, NORID
	}
	,
	{
	    "__volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "typeof", TYPEOF, NORID
	}
	,
	{
	    "typedef", SCSPEC, RID_TYPEDEF
	}
	,
	{
	    "enum", ENUM, NORID
	}
	,
	{
	    "const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "goto", GOTO, NORID
	}
	,
	{
	    "switch", SWITCH, NORID
	}
	,
	{
	    "volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "return", RETURN, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "for", FOR, NORID
	}
	,
	{
	    "continue", CONTINUE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "double", TYPESPEC, RID_DOUBLE
	}
	,
	{
	    "void", TYPESPEC, RID_VOID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "unsigned", TYPESPEC, RID_UNSIGNED
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "long", TYPESPEC, RID_LONG
	}
	,
	{
	    "register", SCSPEC, RID_REGISTER
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "union", UNION, NORID
	}
	,
    };

    if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
	register int key = funky_hash (str, len);

	if (key <= MAX_HASH_VALUE && key >= 0)
	{
	    register char *s = wordlist[key].name;

	    if (*s == *str && !strcmp (str + 1, s + 1))
		return &wordlist[key];
	}
    }
    return 0;
}
#else
/* this is a MSIL non-fractional chip (DSP96002):
   add keywords: modulo, x_memory. */

#define TOTAL_KEYWORDS 53
#define MIN_WORD_LENGTH 2
#define MAX_WORD_LENGTH 13
#define MIN_HASH_VALUE 8
#define MAX_HASH_VALUE 78
/* maximum key range = 71, duplicates = 0 */

#ifdef __GNUC__
inline
#endif
static unsigned int
 funky_hash (str, len)
register char *str;
register int unsigned len;
{
    static unsigned char asso_values[] =
    {
	79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
	    79, 79, 79, 79, 79, 1, 79, 2, 1, 32,
	    7, 5, 18, 20, 13, 17, 79, 1, 26, 30,
	    28, 1, 23, 79, 12, 20, 1, 24, 7, 24,
	    18, 1, 11, 79, 79, 79, 79, 79,
    };
    register int hval = len;

    switch (hval)
    {
    default:
    case 3:
	hval += asso_values[str[2]];
    case 2:
    case 1:
	hval += asso_values[str[0]];
    }
    return hval + asso_values[str[len - 1]];
}

#ifdef __GNUC__
inline
#endif
struct resword *
 funky_is_reserved_word (str, len)
register char *str;
register unsigned int len;
{
    static struct resword wordlist[] =
    {
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "auto", SCSPEC, RID_AUTO
	}
	,
	{
	    "",
	}
	,
	{
	    "do", DO, NORID
	}
	,
	{
	    "__asm__", ASM, NORID
	}
	,
	{
	    "break", BREAK, NORID
	}
	,
	{
	    "__typeof__", TYPEOF, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__alignof__", ALIGNOF, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__attribute__", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__attribute", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__volatile__", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "int", TYPESPEC, RID_INT
	}
	,
	{
	    "__volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "",
	}
	,
	{
	    "float", TYPESPEC, RID_FLOAT
	}
	,
	{
	    "goto", GOTO, NORID
	}
	,
	{
	    "short", TYPESPEC, RID_SHORT
	}
	,
	{
	    "__typeof", TYPEOF, NORID
	}
	,
	{
	    "__inline__", SCSPEC, RID_INLINE
	}
	,
	{
	    "__alignof", ALIGNOF, NORID
	}
	,
	{
	    "__inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "__signed__", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "default", DEFAULT, NORID
	}
	,
	{
	    "else", ELSE, NORID
	}
	,
	{
	    "void", TYPESPEC, RID_VOID
	}
	,
	{
	    "__signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "if", IF, NORID
	}
	,
	{
	    "__asm", ASM, NORID
	}
	,
	{
	    "struct", STRUCT, NORID
	}
	,
	{
	    "extern", SCSPEC, RID_EXTERN
	}
	,
	{
	    "__const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "double", TYPESPEC, RID_DOUBLE
	}
	,
	{
	    "__const__", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "modulo", SCSPEC, RID_MODULO
	}
	,
	{
	    "for", FOR, NORID
	}
	,
	{
	    "volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "return", RETURN, NORID
	}
	,
	{
	    "typeof", TYPEOF, NORID
	}
	,
	{
	    "typedef", SCSPEC, RID_TYPEDEF
	}
	,
	{
	    "char", TYPESPEC, RID_CHAR
	}
	,
	{
	    "while", WHILE, NORID
	}
	,
	{
	    "register", SCSPEC, RID_REGISTER
	}
	,
	{
	    "signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "sizeof", SIZEOF, NORID
	}
	,
	{
	    "switch", SWITCH, NORID
	}
	,
	{
	    "x_memory", SCSPEC, RID_X_MEMORY
	}
	,
	{
	    "",
	}
	,
	{
	    "unsigned", TYPESPEC, RID_UNSIGNED
	}
	,
	{
	    "static", SCSPEC, RID_STATIC
	}
	,
	{
	    "case", CASE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "enum", ENUM, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "asm", ASM, NORID
	}
	,
	{
	    "const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "continue", CONTINUE, NORID
	}
	,
	{
	    "union", UNION, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "long", TYPESPEC, RID_LONG
	}
	,
    };

    if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
	register int key = funky_hash (str, len);

	if (key <= MAX_HASH_VALUE && key >= 0)
	{
	    register char *s = wordlist[key].name;

	    if (*s == *str && !strcmp (str + 1, s + 1))
		return &wordlist[key];
	}
    }
    return 0;
}
#endif
#else
/* this is an OHT fractional chip (DSP561XX, DSP564XX):
   add keywords: modulo, frac, extended. */

#define TOTAL_KEYWORDS 54
#define MIN_WORD_LENGTH 2
#define MAX_WORD_LENGTH 13
#define MIN_HASH_VALUE 7
#define MAX_HASH_VALUE 95
/* maximum key range = 89, duplicates = 0 */

#ifdef __GNUC__
inline
#endif
static unsigned int
 funky_hash (str, len)
register char *str;
register int unsigned len;
{
    static unsigned char asso_values[] =
    {
	96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 96, 96, 96, 96, 96,
	    96, 96, 96, 96, 96, 1, 96, 2, 1, 16,
	    10, 5, 18, 40, 12, 17, 96, 1, 10, 1,
	    31, 9, 25, 96, 19, 20, 1, 42, 33, 15,
	    96, 96, 4, 96, 96, 96, 96, 96,
    };
    register int hval = len;

    switch (hval)
    {
    default:
    case 3:
	hval += asso_values[str[2]];
    case 2:
    case 1:
	hval += asso_values[str[0]];
    }
    return hval + asso_values[str[len - 1]];
}

#ifdef __GNUC__
inline
#endif
struct resword *
 funky_is_reserved_word (str, len)
register char *str;
register unsigned int len;
{
    static struct resword wordlist[] =
    {
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "asm", ASM, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__asm", ASM, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__asm__", ASM, NORID
	}
	,
	{
	    "break", BREAK, NORID
	}
	,
	{
	    "__typeof__", TYPEOF, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__alignof__", ALIGNOF, NORID
	}
	,
	{
	    "auto", SCSPEC, RID_AUTO
	}
	,
	{
	    "__attribute__", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__attribute", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "do", DO, NORID
	}
	,
	{
	    "int", TYPESPEC, RID_INT
	}
	,
	{
	    "",
	}
	,
	{
	    "extended", TYPESPEC, RID_EXTENDED
	}
	,
	{
	    "__const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "modulo", SCSPEC, RID_MODULO
	}
	,
	{
	    "__const__", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "__typeof", TYPEOF, NORID
	}
	,
	{
	    "__inline__", SCSPEC, RID_INLINE
	}
	,
	{
	    "__alignof", ALIGNOF, NORID
	}
	,
	{
	    "__inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "__signed__", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "float", TYPESPEC, RID_FLOAT
	}
	,
	{
	    "else", ELSE, NORID
	}
	,
	{
	    "short", TYPESPEC, RID_SHORT
	}
	,
	{
	    "default", DEFAULT, NORID
	}
	,
	{
	    "if", IF, NORID
	}
	,
	{
	    "inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "__signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "frac", TYPESPEC, RID_FRAC
	}
	,
	{
	    "char", TYPESPEC, RID_CHAR
	}
	,
	{
	    "while", WHILE, NORID
	}
	,
	{
	    "extern", SCSPEC, RID_EXTERN
	}
	,
	{
	    "static", SCSPEC, RID_STATIC
	}
	,
	{
	    "case", CASE, NORID
	}
	,
	{
	    "struct", STRUCT, NORID
	}
	,
	{
	    "__volatile__", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "sizeof", SIZEOF, NORID
	}
	,
	{
	    "__volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "typeof", TYPEOF, NORID
	}
	,
	{
	    "typedef", SCSPEC, RID_TYPEDEF
	}
	,
	{
	    "enum", ENUM, NORID
	}
	,
	{
	    "const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "goto", GOTO, NORID
	}
	,
	{
	    "switch", SWITCH, NORID
	}
	,
	{
	    "volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "return", RETURN, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "for", FOR, NORID
	}
	,
	{
	    "continue", CONTINUE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "double", TYPESPEC, RID_DOUBLE
	}
	,
	{
	    "void", TYPESPEC, RID_VOID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "unsigned", TYPESPEC, RID_UNSIGNED
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "long", TYPESPEC, RID_LONG
	}
	,
	{
	    "register", SCSPEC, RID_REGISTER
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "union", UNION, NORID
	}
	,
    };

    if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
	register int key = funky_hash (str, len);

	if (key <= MAX_HASH_VALUE && key >= 0)
	{
	    register char *s = wordlist[key].name;

	    if (*s == *str && !strcmp (str + 1, s + 1))
		return &wordlist[key];
	}
    }
    return 0;
}
#endif

/* here's the default GNU C keyword matcher. */

#undef TOTAL_KEYWORDS
#undef MIN_WORD_LENGTH
#undef MAX_WORD_LENGTH
#undef MIN_HASH_VALUE
#undef MAX_HASH_VALUE
#define TOTAL_KEYWORDS 51
#define MIN_WORD_LENGTH 2
#define MAX_WORD_LENGTH 13
#define MIN_HASH_VALUE 7
#define MAX_HASH_VALUE 91
/* maximum key range = 85, duplicates = 0 */

#ifdef __GNUC__
inline
#endif
static unsigned int
 hash (str, len)
register char *str;
register int unsigned len;
{
    static unsigned char asso_values[] =
    {
	92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	    92, 92, 92, 92, 92, 1, 92, 2, 1, 32,
	    7, 5, 18, 20, 1, 17, 92, 1, 18, 1,
	    28, 1, 23, 92, 12, 20, 1, 41, 7, 15,
	    92, 92, 10, 92, 92, 92, 92, 92,
    };
    register int hval = len;

    switch (hval)
    {
    default:
    case 3:
	hval += asso_values[str[2]];
    case 2:
    case 1:
	hval += asso_values[str[0]];
    }
    return hval + asso_values[str[len - 1]];
}

#ifdef __GNUC__
inline
#endif
struct resword *
 normal_is_reserved_word (str, len)
register char *str;
register unsigned int len;
{
    static struct resword wordlist[] =
    {
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "asm", ASM, NORID
	}
	,
	{
	    "auto", SCSPEC, RID_AUTO
	}
	,
	{
	    "__asm", ASM, NORID
	}
	,
	{
	    "do", DO, NORID
	}
	,
	{
	    "__asm__", ASM, NORID
	}
	,
	{
	    "break", BREAK, NORID
	}
	,
	{
	    "__typeof__", TYPEOF, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__alignof__", ALIGNOF, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__attribute__", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__attribute", ATTRIBUTE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "__volatile__", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "int", TYPESPEC, RID_INT
	}
	,
	{
	    "__volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "",
	}
	,
	{
	    "float", TYPESPEC, RID_FLOAT
	}
	,
	{
	    "goto", GOTO, NORID
	}
	,
	{
	    "short", TYPESPEC, RID_SHORT
	}
	,
	{
	    "__typeof", TYPEOF, NORID
	}
	,
	{
	    "__inline__", SCSPEC, RID_INLINE
	}
	,
	{
	    "__alignof", ALIGNOF, NORID
	}
	,
	{
	    "__inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "__signed__", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "default", DEFAULT, NORID
	}
	,
	{
	    "else", ELSE, NORID
	}
	,
	{
	    "void", TYPESPEC, RID_VOID
	}
	,
	{
	    "__signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "if", IF, NORID
	}
	,
	{
	    "volatile", TYPE_QUAL, RID_VOLATILE
	}
	,
	{
	    "struct", STRUCT, NORID
	}
	,
	{
	    "extern", SCSPEC, RID_EXTERN
	}
	,
	{
	    "__const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "while", WHILE, NORID
	}
	,
	{
	    "__const__", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "switch", SWITCH, NORID
	}
	,
	{
	    "for", FOR, NORID
	}
	,
	{
	    "inline", SCSPEC, RID_INLINE
	}
	,
	{
	    "return", RETURN, NORID
	}
	,
	{
	    "typeof", TYPEOF, NORID
	}
	,
	{
	    "typedef", SCSPEC, RID_TYPEDEF
	}
	,
	{
	    "char", TYPESPEC, RID_CHAR
	}
	,
	{
	    "enum", ENUM, NORID
	}
	,
	{
	    "register", SCSPEC, RID_REGISTER
	}
	,
	{
	    "signed", TYPESPEC, RID_SIGNED
	}
	,
	{
	    "sizeof", SIZEOF, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "double", TYPESPEC, RID_DOUBLE
	}
	,
	{
	    "static", SCSPEC, RID_STATIC
	}
	,
	{
	    "case", CASE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "const", TYPE_QUAL, RID_CONST
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "long", TYPESPEC, RID_LONG
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "continue", CONTINUE, NORID
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "unsigned", TYPESPEC, RID_UNSIGNED
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "",
	}
	,
	{
	    "union", UNION, NORID
	}
	,
    };

    if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
	register int key = hash (str, len);

	if (key <= MAX_HASH_VALUE && key >= 0)
	{
	    register char *s = wordlist[key].name;

	    if (*s == *str && !strcmp (str + 1, s + 1))
		return &wordlist[key];
	}
    }
    return 0;
}

#endif

enum info_spec added_flags;

enum info_spec
 get_spec_bits (given)
enum info_spec given;
{
    return given | added_flags;
}

rtx
get_nonnote_insn (insn)
rtx insn;
{
    while (NULL != insn && NOTE == GET_CODE (insn))
    {
	insn = NEXT_INSN (insn);
    }
    return insn;
}

/* count the number of 0 bits in the bits least-significant bits of word. */
int
 zero_count (word, bits)
int word,
 bits;
{
    int count = 0;

    while (bits--)
    {
	count += !(word & 1);
	word >>= 1;
    }
    return count;
}

/* count the number of 1 bits in the bits least-significant bits of word. */
int
 one_count (word, bits)
int word,
 bits;
{
    int count = 0;

    while (bits--)
    {
	count += (word & 1);
	word >>= 1;
    }
    return count;
}

char *
 output_repeated_bitop (name, operands, immed_count)
char *name;
rtx *operands;
int immed_count;
{
    int isnt_and_p = strcmp (name + 1, "and");
    int mask = (isnt_and_p) ? 0 : -1;

    if (isnt_and_p)
    {
	do
	{
	    mask |= (1 << INTVAL (operands[immed_count--]));
	}
	while (immed_count);
    }
    else
    {
	do
	{
	    mask &= ~(1 << INTVAL (operands[immed_count--]));
	}
	while (immed_count);
    }

    operands[1] = gen_rtx (CONST_INT, VOIDmode, mask);
    operands[2] = gen_rtx (SYMBOL_REF, VOIDmode, name);

    return "%2	#%1,%0";
}

char *
 output_shift (name, operands)
char *name;
rtx *operands;
{
    operands[3] = gen_label_rtx ();
    operands[4] = gen_rtx (SYMBOL_REF, VOIDmode, name);

    if (TARGET_REP)
    {
	return "tst	%2" SC
	    "jeq	%l3" SC
	    "rep	%e2" SC
	    "%4	%0"
	    "\n%l3";
    }
    else
    {
	return "tst	%2" SC
	    "jeq	%l3" SC
	    "do	%e2,%l3" SC
	    "%4	%0"
	    "\n%l3";
    }
}

#if defined( DSP56300 ) || defined( DSP56000 )
unsigned int
 upper_24 (dbl)
rtx dbl;
{
    return (((CONST_DOUBLE_LOW (dbl) >> 24) & 0x000000ff) |
	    ((CONST_DOUBLE_HIGH (dbl) << 8) & 0x00ffff00));
}

unsigned int
 lower_24 (dbl)
rtx dbl;
{
    return CONST_DOUBLE_LOW (dbl) & 0x00ffffff;
}
#endif

#if defined( DSP56600 )
unsigned int
 upper_16 (dbl)
rtx dbl;
{
    return (CONST_DOUBLE_LOW (dbl) >> 16) & 0x0000ffff;
}

unsigned int
 lower_16 (dbl)
rtx dbl;
{
    return CONST_DOUBLE_LOW (dbl) & 0x0000ffff;
}
#endif

#if defined( DSP56600 ) || defined( DSP56300 ) || defined( DSP96000 )

void
 mark_2nd_parm_loads (first)
rtx first;
{
    rtx call,
     parm2_use,
     parm2_set,
     parm2_reg;

#if defined( DSP56300 ) || defined( DSP56600 )
    int parm2_regno = DSP_B_REGNUM;

#else	/* if defined( DSP96000 ) */
    int parm2_regno = DSP_D1L_REGNUM;

#endif

    /* run thru all the rtl in the function. */
    do
    {
	/* find the next CALL_INSN */
	for (call = first;
	     NULL != call && CALL_INSN != GET_CODE (call);
	     call = NEXT_INSN (call));

	/* find a possible 2nd parameter use. */
	for (parm2_use = call;
	     NULL != parm2_use &&
	     parm2_use != first &&
	     !(INSN == GET_CODE (parm2_use) &&
	       USE == GET_CODE (PATTERN (parm2_use)) &&
	       REG_P (XEXP (PATTERN (parm2_use), 0)) &&
	       parm2_regno == REGNO (XEXP (PATTERN (parm2_use), 0)));
	     parm2_use = PREV_INSN (parm2_use));

	if (NULL != parm2_use && parm2_use != first)
	{
	    /* we did have a 2nd parameter use. now find the INSN that
	       SETs the 2nd parameter. */

	    parm2_reg = XEXP (PATTERN (parm2_use), 0);

	    for (parm2_set = PREV_INSN (parm2_use);
		 !(INSN == GET_CODE (parm2_set) &&
		   reg_set_p (parm2_reg, PATTERN (parm2_set)));
		 parm2_set = PREV_INSN (parm2_set));

	    /* mark off the zone where the 2nd parameter might get 
	       clobbered. */

	    RTX_PARM_HAZARD_ZONE_P
		(emit_insn_after (const0_rtx, parm2_set)) = 1;
	    RTX_PARM_HAZARD_ZONE_P
		(emit_insn_before (const0_rtx, parm2_use)) = 1;
	}

	/* advance past the just examined call site. */
	first = (NULL == call) ? NULL : NEXT_INSN (call);
    }
    while (NULL != first);
}

void
 unclobber_2nd_parms (first)
rtx first;
{
    /* run thru all the rtl in the function. */
    do
    {
	if (RTX_PARM_HAZARD_ZONE_P (first))
	{
	    rtx cone1 = first,
	     cone2 = NEXT_INSN (first);
	    rtx parm2_reg;

	    /* first must the the SET of a 2nd parameter. find the
	       use of that parameter. */

	    for (; !RTX_PARM_HAZARD_ZONE_P (cone2);
		 cone2 = NEXT_INSN (cone2));

	    parm2_reg =
		XEXP (PATTERN (get_nonnote_insn (NEXT_INSN (cone2))), 0);

	    /* check whether the 2nd parameter got clobbered or not */

	    if (reg_set_between_p (parm2_reg, cone1, cone2))
	    {
		enum machine_mode mode = GET_MODE (parm2_reg);
		rtx stack_slot =
		 assign_stack_local (mode, GET_MODE_SIZE (mode));

		/* we have a problem. grab an extra frame slot and
		   place the 2nd parameter in the frame until just before
		   the call. */

		PATTERN (cone1) =
		    gen_rtx (SET, mode, stack_slot, parm2_reg);

		PATTERN (cone2) =
		    gen_rtx (SET, mode, parm2_reg, stack_slot);
	    }
	    first = cone2;
	}
	first = NEXT_INSN (first);
    }
    while (NULL != first);
}

#endif

late_rtx_equal_p (a, b)
rtx a,
 b;
{
    extern int rtx_equal_function_value_matters;
    int equal_p;

    rtx_equal_function_value_matters = 0;
    equal_p = rtx_equal_p (a, b);
    rtx_equal_function_value_matters = 1;

    return equal_p;
}

#if 0
void
 manage_accumulators (insn)
rtx insn;
{
    /* the various states of an accumulator segment: all zero bits, 
       all data bits, all sign bits or potential arithmetic overflow bits,
       or just plain unknown bits. */
    static enum
    {
	AS_ZERO, AS_DATA, AS_POFLOW, AS_UNKNOWN
    }
    accum_state[2][3];

    /* does each accumulator contain a live value? */
    static enum
    {
	AL_LIVE, AL_DEAD
    }
    accum_status[2];

    enum which_accum
    {
	REG_A, REG_B
    };

    /* RTL for standard moves and both SImode accumulators. */
    static rtx reg_a_rtx,
     reg_b_rtx;
    static rtx move_a0_a_rtx,
     move_a1_a_rtx,
     move_a1_a0_rtx;
    static rtx move_b0_b_rtx,
     move_b1_b_rtx,
     move_b1_b0_rtx;

    if (NULL == insn)
    {
	/* assume that upon block entry, both accumulators are filled
	   with real data, arranged in the canonical format:
	   (A2:A1:A0) is (AS_POFLOW:AS_DATA:AS_ZERO) */

	accum_state[REG_A][2] = AS_POFLOW;
	accum_state[REG_B][2] = AS_POFLOW;
	accum_state[REG_A][1] = AS_DATA;
	accum_state[REG_B][1] = AS_DATA;
	accum_state[REG_A][0] = AS_ZERO;
	accum_state[REG_B][0] = AS_ZERO;

	accum_state[REG_A] = AL_LIVE;
	accum_state[REG_B] = AL_LIVE;

	reg_a_rtx = gen_rtx (REG, SImode, DSP_A_REGNUM);
	reg_b_rtx = gen_rtx (REG, SImode, DSP_B_REGNUM);

	move_a0_a_rtx =
	    gen_rtx (SET, VOIDmode, reg_a_rtx,
		     gen_rtx (SUBREG, SImode,
			      gen_rtx (REG, DImode, DSP_A_REGNUM), 0));
	move_b0_b_rtx =
	    gen_rtx (SET, VOIDmode, reg_b_rtx,
		     gen_rtx (SUBREG, SImode,
			      gen_rtx (REG, DImode, DSP_B_REGNUM), 0));

	move_a1_a_rtx = gen_rtx (SET, VOIDmode, reg_a_rtx, reg_a_rtx);
	move_b1_b_rtx = gen_rtx (SET, VOIDmode, reg_b_rtx, reg_b_rtx);

	move_a1_a0_rtx =
	    gen_rtx (SET, VOIDmode, gen_rtx
		     (SUBREG, SImode,
		      gen_rtx (REG, DImode, DSP_A_REGNUM), 0), reg_a_rtx);
	move_b1_b0_rtx =
	    gen_rtx (SET, VOIDmode, gen_rtx
		     (SUBREG, SImode,
		      gen_rtx (REG, DImode, DSP_B_REGNUM), 0), reg_b_rtx);
	return;
    }

    if (INSN == GET_CODE (insn))
    {
	/* figure out if any, none or both accumulators are read by
	   this insn. figure out which, if any, accumulator is set by
	   this insn. */

	int reg_a_read = reg_read_p (reg_a_rtx, PATTERN (insn), TRUE);
	int reg_b_read = reg_read_p (reg_b_rtx, PATTERN (insn), TRUE);
	int reg_a_written =
	 reg_written_p (reg_a_rtx, PATTERN (insn), TRUE);
	int reg_b_written =
	 reg_written_p (reg_b_rtx, PATTERN (insn), TRUE);

	/* switch on the insn type - how is the insn affected by the 
	   accumulator contents, how does it affect the accumulator 
	   contents?

	   based on the insn type, pre-format the contents of any
	   accumulator references, and note how any accumulator set
	   would affect the format of the contents. */

	switch (insn_machine_info[recog_memoized (insn)].ait_spec)
	{
	  AIT_MOVE:
	    if (reg_a_read)
	    {
		if (AS_DATA != accum_state[REG_A][1])
		{
		    assert (AS_DATA == accum_state[REG_A][0]);

		    emit_insn_before (move_a0_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		    accum_state[REG_A][0] = AS_ZERO;
		}
	    }
	    if (reg_b_read)
	    {
		abort ();
	    }

	    if (reg_a_written)
	    {
		accum_state[REG_A][2] = AS_POFLOW;
		accum_state[REG_A][1] = AS_DATA;
		accum_state[REG_A][0] = AS_ZERO;
	    }
	    else if (reg_b_written)
	    {
		accum_state[REG_B][2] = AS_POFLOW;
		accum_state[REG_B][1] = AS_DATA;
		accum_state[REG_B][0] = AS_ZERO;
	    }
	    break;

	  AIT_ALU:
	    /* may use one or two accumulators, may set zero or one of
	       them. */

	    if (reg_a_read)
	    {
		if (AS_DATA != accum_state[REG_A][1])
		{
		    assert (AS_DATA == accum_state[REG_A][0]);

		    emit_insn_before (move_a0_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		    accum_state[REG_A][0] = AS_ZERO;
		}
		if (AS_ZERO != accum_state[REG_A][0])
		{
		    assert (AS_DATA == accum_state[REG_A][1]);

		    emit_insn_before (move_a1_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		    accum_state[REG_A][0] = AS_ZERO;
		}
	    }

	    if (reg_b_read)
	    {
		if (AS_DATA != accum_state[REG_B][1])
		{
		    assert (AS_DATA == accum_state[REG_B][0]);

		    emit_insn_before (move_b0_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		    accum_state[REG_B][1] = AS_DATA;
		    accum_state[REG_B][0] = AS_ZERO;
		}
		if (AS_ZERO != accum_state[REG_B][0])
		{
		    assert (AS_DATA == accum_state[REG_B][1]);

		    emit_insn_before (move_b1_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		    accum_state[REG_B][1] = AS_DATA;
		    accum_state[REG_B][0] = AS_ZERO;
		}
	    }

	    if (reg_a_written || reg_b_written)
	    {
		enum which_accum set_accum = (reg_a_written) ? REG_A : REG_B;

		if ((reg_a_read && AS_POFLOW != accum_state[REG_A][2]) ||
		    (reg_b_read && AS_POFLOW != accum_state[REG_B][2]))
		{
		    accum_state[set_accum][2] = AS_UNKNOWN;
		}
		else
		{
		    accum_state[set_accum][2] = AS_POFLOW;
		}
		accum_state[set_accum][1] = AS_DATA;
		accum_state[set_accum][0] = AS_ZERO;
	    }
	    break;

	  AIT_LOGICAL:
	    /* may use one or two accumulators, may set zero or one of
	       them. */

	    if (reg_a_read)
	    {
		if (AS_DATA != accum_state[REG_A][1])
		{
		    assert (AS_DATA == accum_state[REG_A][0]);

		    emit_insn_before (move_a0_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		    accum_state[REG_A][0] = AS_ZERO;
		}
	    }

	    if (reg_b_read)
	    {
		if (AS_DATA != accum_state[REG_B][1])
		{
		    assert (AS_DATA == accum_state[REG_B][0]);

		    emit_insn_before (move_b0_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		    accum_state[REG_B][1] = AS_DATA;
		    accum_state[REG_B][0] = AS_ZERO;
		}
	    }

	    if (reg_a_written || reg_b_written)
	    {
		enum which_accum set_accum = (reg_a_written) ? REG_A : REG_B;

		if ((reg_a_read && AS_ZERO != accum_state[REG_A][2]) ||
		    (reg_b_read && AS_ZERO != accum_state[REG_B][2]))
		{
		    accum_state[set_accum][2] = AS_UNKNOWN;
		}
		else
		{
		    accum_state[set_accum][2] = AS_ZERO;
		}
		accum_state[set_accum][1] = AS_DATA;
	    }
	    break;

	  AIT_MAC:
	    /* may use one accumulator, may set one. */

	    if (reg_a_read)
	    {
		if (AS_DATA != accum_state[REG_A][0])
		{
		    assert (AS_DATA == accum_state[REG_A][1]);

		    emit_insn_before (move_a1_a0_rtx, insn);

		    accum_state[REG_A][2] = AS_UNKNOWN;
		    accum_state[REG_A][1] = AS_UNKNOWN;
		    accum_state[REG_A][0] = AS_DATA;
		}
	    }

	    if (reg_b_read)
	    {
		if (AS_DATA != accum_state[REG_B][0])
		{
		    assert (AS_DATA == accum_state[REG_B][1]);

		    emit_insn_before (move_b1_b0_rtx, insn);

		    accum_state[REG_B][2] = AS_UNKNOWN;
		    accum_state[REG_B][1] = AS_UNKNOWN;
		    accum_state[REG_B][0] = AS_DATA;
		}
	    }

	    if (reg_a_written || reg_b_written)
	    {
		enum which_accum set_accum = (reg_a_written) ? REG_A : REG_B;

		accum_state[set_accum][2] = AS_UNKNOWN;
		accum_state[set_accum][1] = AS_UNKNOWN;
		accum_state[set_accum][0] = AS_DATA;
	    }
	    break;

	  AIT_ASR:
	    /* will use one accumulator, will set one. */

	    if (reg_a_read)
	    {
		if (AS_DATA != accum_state[REG_A][1])
		{
		    assert (AS_DATA == accum_state[REG_A][0]);

		    emit_insn_before (move_a0_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		}
		if (AS_POFLOW != accum_state[REG_A][2])
		{
		    assert (AS_DATA == accum_state[REG_A][1]);

		    emit_insn_before (move_a1_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		}
		accum_state[REG_A][1] = AS_DATA;
		accum_state[REG_A][0] = AS_UNKNOWN;
	    }

	    if (reg_b_read)
	    {
		if (AS_DATA != accum_state[REG_B][1])
		{
		    assert (AS_DATA == accum_state[REG_B][0]);

		    emit_insn_before (move_b0_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		    accum_state[REG_B][1] = AS_DATA;
		}
		if (AS_POFLOW != accum_state[REG_B][2])
		{
		    assert (AS_DATA == accum_state[REG_B][1]);

		    emit_insn_before (move_b1_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		}
		accum_state[REG_B][1] = AS_DATA;
		accum_state[REG_B][0] = AS_UNKNOWN;
	    }
	    break;

	  AIT_ALU_ASR:
	    /* may use one or two accumulators, may set zero or one of
	       them. */

	    if (reg_a_read)
	    {
		if (AS_DATA != accum_state[REG_A][1])
		{
		    assert (AS_DATA == accum_state[REG_A][0]);

		    emit_insn_before (move_a0_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		    accum_state[REG_A][0] = AS_ZERO;
		}
		if (AS_ZERO != accum_state[REG_A][0])
		{
		    assert (AS_DATA == accum_state[REG_A][1]);

		    emit_insn_before (move_a1_a_rtx, insn);

		    accum_state[REG_A][2] = AS_POFLOW;
		    accum_state[REG_A][1] = AS_DATA;
		    accum_state[REG_A][0] = AS_ZERO;
		}
	    }

	    if (reg_b_read)
	    {
		if (AS_DATA != accum_state[REG_B][1])
		{
		    assert (AS_DATA == accum_state[REG_B][0]);

		    emit_insn_before (move_b0_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		    accum_state[REG_B][1] = AS_DATA;
		    accum_state[REG_B][0] = AS_ZERO;
		}
		if (AS_ZERO != accum_state[REG_B][0])
		{
		    assert (AS_DATA == accum_state[REG_B][1]);

		    emit_insn_before (move_b1_b_rtx, insn);

		    accum_state[REG_B][2] = AS_POFLOW;
		    accum_state[REG_B][1] = AS_DATA;
		    accum_state[REG_B][0] = AS_ZERO;
		}
	    }

	    if (reg_a_written || reg_b_written)
	    {
		enum which_accum set_accum = (reg_a_written) ? REG_A : REG_B;

		if ((reg_a_read && AS_POFLOW != accum_state[REG_A][2]) ||
		    (reg_b_read && AS_POFLOW != accum_state[REG_B][2]))
		{
		    accum_state[set_accum][2] = AS_UNKNOWN;
		}
		else
		{
		    accum_state[set_accum][2] = AS_POFLOW;
		}
		accum_state[set_accum][1] = AS_DATA;
		accum_state[set_accum][0] = AS_UNKNOWN;
	    }
	    break;

	}
    }
    else if (CALL_INSN == GET_CODE (insn))
    {
	accum_state[REG_A][2] = AS_POFLOW;
	accum_state[REG_A][1] = AS_DATA;
	accum_state[REG_A][0] = AS_ZERO;
    }
}

#endif
